package com.blackledge.david.html;

import java.awt.Color;
import java.util.ArrayList;
import java.util.List;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

public class HTMLTableCell extends HTMLChunk {

	private static final Pattern WIDTH_PATTERN = Pattern.compile("(\\d+)[^\\d]*");
	public List<HTMLRow> rows = new ArrayList<HTMLRow>();
	public int rowSpan = 1;
	public int colSpan = 1;
	public Color bgcolor;
	public boolean header = false;
	private boolean widthIsPercent = false;
	private int widthNum = -1;
	private int calculatedWidth = -1;
	public int align = HTMLRow.ROW_ALIGN_LEFT;
	public int bgposition;
	public int bgrepeat;
	public String bg;
	
	@Override
	public String toString() {
		StringBuffer sb = new StringBuffer();
		for(HTMLRow row : rows) {
			sb.append("|"+row+"|\n");
		}
		return sb.toString();
	}
	
	public HTMLTableCell(boolean header, int rowspan, int colspan, String width, Color bgcolor, int bgposition, int bgrepeat, String bg) {
		chunkType = ChunkType.TABLECELL;
		this.header = header;
		this.rowSpan = rowspan;
		this.colSpan = colspan;
		this.bgcolor = bgcolor;
		this.bgposition = bgposition;
		this.bgrepeat = bgrepeat;
		this.bg = bg;
		
		this.rows.add(new HTMLRow());
		
		if (width != null) {
			width = width.trim();
			Matcher m = WIDTH_PATTERN.matcher(width);
			if (m.matches()) {
				if (width.endsWith("%")) {
					widthIsPercent = true;
				}
				if (width.endsWith("*")) {
					//FIXME support proportional somehow
				}
				widthNum = Integer.parseInt(m.group(1));
			}
		}
	}

	/**
	 * get cell width in pixels (if stored as a percent, returns that percenet
	 * of argument)
	 * @param width the parent block's total width
	 */
	public int getWidth(int width) {
		// allows calculated width to override if it's larger than specified... but within some limit.
		// currently limit is 1/4 of width
		if(widthIsPercent) {
			int returnWidth = (int) (width*(((double)this.widthNum)/100.0));
			if(this.calculatedWidth > 0 && this.calculatedWidth > returnWidth
					&& this.calculatedWidth < width/4
					) {
//				WebBrowserApp.debugOut("getWidth percent "+this.widthNum+"% of "+width+" but using calc'd width "+this.calculatedWidth);
				return this.calculatedWidth;
			} else {
//				WebBrowserApp.debugOut("getWidth percent "+this.widthNum+"% of "+width+" is "+returnWidth);
				return returnWidth;
			}
		} else if(this.widthNum > 0){
			int returnWidth = this.widthNum;
			if(this.calculatedWidth > 0 && this.calculatedWidth > returnWidth
					&& this.calculatedWidth < width/4
					) {
//				WebBrowserApp.debugOut("getWidth specified "+returnWidth+" but using calc'd width "+this.calculatedWidth);
				return this.calculatedWidth;
			} else {
//				WebBrowserApp.debugOut("getWidth specified "+returnWidth);
				return returnWidth;
			}
		} else {
//			WebBrowserApp.debugOut("getWidth had "+this.widthNum+"so using calc'd width "+this.calculatedWidth);
			return this.calculatedWidth;
		}
	}
	
	public void setCalculatedWidth(int width) {
		this.calculatedWidth = width;
	}

	public boolean isWidthSpecified() {
		return (this.widthIsPercent || this.widthNum > 0);
	}

}
