package com.blackledge.david.html.hme;

import java.awt.Color;
import java.awt.Rectangle;
import java.net.URL;
import java.text.MessageFormat;
import java.util.List;
import java.util.Vector;
import java.util.Map.Entry;

import com.blackledge.david.html.UnicodeHelper;
import com.blackledge.david.tivo.V49BApplication;
import com.blackledge.david.tivo.V49IHmeProtocol;
import com.blackledge.david.tivo.widgets.typo.TyPo;
import com.blackledge.david.tivo.widgets.typo.TyPo.Keyboard;

import com.tivo.hme.bananas.BApplication;
import com.tivo.hme.bananas.BButton;
import com.tivo.hme.bananas.BEvent;
import com.tivo.hme.bananas.BList;
import com.tivo.hme.bananas.BScreen;
import com.tivo.hme.bananas.BView;
import com.tivo.hme.sdk.StreamResource;
import com.tivo.hme.sdk.HmeEvent.FontInfo;

public class BookmarkFolderScreen extends BScreen implements VideoBackgroundScreen {

	private FolderList list;
	private List<String> folders;
	private int folderStartIndex;
	private Entry<String, URL[]> bookmark;
	private TyPo typo;
	private String folder;
	private int indexInFolder;

	public BookmarkFolderScreen(BApplication app, String folder, int indexInFolder) {
		/* BEGIN HME 1.4.1e only */
		super(app, app.getReceiverResolutionInfo().getPreferredRenderingResolution());
		/* END HME 1.4.1e only */
		this.folder = folder;
		this.indexInFolder = indexInFolder;
		// this.getBelow().setResource(Color.yellow);
		WebBrowserApp wba = (WebBrowserApp)getApp();
		Rectangle titleArea = wba.getTitleRegion();
		BView banner = new BView(this.getNormal(), titleArea.x, titleArea.y, titleArea.width, titleArea.height);
		BView title = new BView(banner, 0, 0, banner.getWidth(), 20);
//debugOut("notice"+WebBrowserApp.NOTICE+":"+WebBrowserApp.BROWSER_NAME);
		title.setResource(
				title.createText("system-"+(getRowFontSize(this)+1)+".font", Color.white,  //$NON-NLS-1$ //$NON-NLS-2$
						"Select group for the "+BookmarkScreen.BOOKMARK_NAME
						), RSRC_HALIGN_CENTER
						| RSRC_VALIGN_BOTTOM );
//		BView hints = new BView(banner, 0, title.getHeight(), banner.getWidth(),
//				banner.getHeight() - title.getHeight());
//		hints.setResource(
//				hints.createText(
//						"default-"+(getRowFontSize(this)-2)+".font", //$NON-NLS-1$ //$NON-NLS-2$
//						Color.LIGHT_GRAY,
//						MessageFormat.format(Messages.getString("BookmarkScreen.hints"), BookmarkScreen.BOOKMARK_NAME)), //$NON-NLS-1$
//				RSRC_HALIGN_CENTER | RSRC_VALIGN_CENTER | RSRC_TEXT_WRAP);

		Rectangle scrollArea = wba.getScrollRegion();
		BView listArea = new BView(this.getNormal(), scrollArea.x, scrollArea.y, scrollArea.width, scrollArea.height);

		int buttonHeight = getBApp().getSkin().get(H_BAR).getHeight();
		list = new FolderList(this, listArea, 0, 0, listArea.getWidth(),
				listArea.getHeight(), buttonHeight/2);
		list.setBarAndArrows(BAR_HANG, BAR_HANG, "back", "FOLDER"); //$NON-NLS-1$ //$NON-NLS-2$

		setFocusDefault(list);
		setupList();
	}
	
	private static final String NEW_FOLDER = Messages.getString("BookmarkFolderScreen.New_Folder"); //$NON-NLS-1$
	
	private void setupList() {
		int oldFocus = list.getFocus();
		if(oldFocus < 0) oldFocus = 0;
		list.clear();
		// have to set this value BEFORE adding list entries.  creatRow is called immediately on add.
		if(this.bookmark != null) {
			this.folderStartIndex = 4;
		} else {
			this.folderStartIndex = 1;
		}
		
		if(this.bookmark != null) {
			String folder = ((WebBrowserApp)getBApp()).getFolder(this.bookmark.getKey());
			if(folder == null) folder = "";
			String absfolder = WebBrowserApp.folderToReadablePath(this.folder)+"/"+WebBrowserApp.folderToReadablePath( folder );
			list.add(MessageFormat.format(Messages.getString("BookmarkFolderScreen.Edit_Name"),  //$NON-NLS-1$
					this.bookmark.getKey().substring(folder.length())));
			if(absfolder.length() == 0) absfolder = "(\"/\" - top level)";
			list.add(MessageFormat.format(Messages.getString("BookmarkFolderScreen.Current_Folder"),absfolder)); //$NON-NLS-1$
			list.add(Messages.getString("BookmarkFolderScreen.Change_Folder")); //$NON-NLS-1$
		}

		list.add(NEW_FOLDER);
		
		WebBrowserApp.debugOut("before folders "+this.folderStartIndex);

		if (folders != null) {
			list.add(new Vector<String>(folders));
		}
		if(oldFocus >= list.size()) {
			oldFocus = list.size()-1;
		}
		list.setFocus(oldFocus, false);
	}

	@Override
	public boolean handleAction(BView view, Object action) {
		if (action.equals("back")) { //$NON-NLS-1$
				getBApp().pop();
			return true;
		} else if (action.equals("FOLDER")) {
			int row = list.getFocus();
			if(row < folderStartIndex) {
				if(row == folderStartIndex - 1) {
					// create new folder and set folder
					addFolder(this.folder);
					return true;
				}
				else
				if(this.bookmark != null) {
					switch(row) {
					case 0: // edit name
						editTitle();
						return true;
					case 1: // current folder (cancel, I guess)
						getBApp().pop();
						return true;
					case 2: // top (set with no folder)
						setFolder("");
						getBApp().pop();
						return true;
					}
				}
			} else {
				// set folder
				int index = row - folderStartIndex;
				if (index < folders.size()) {
					String folder = folders.get(index);
					setFolder(folder);
					getBApp().pop();
					return true;
				}
			}
			return false;
		}
		return super.handleAction(view, action);
	}

	private void addFolder(String parent) {
		setUpKeyboard();
		
		typo.setKeyboard(Keyboard.DEFAULT);
		typo.setActions("newFolder", "newFolderCancel"); //$NON-NLS-1$ //$NON-NLS-2$
		if (parent != null && parent.length() > 0) {
			typo.setValue(WebBrowserApp.folderToReadablePath(parent) + "/");
		} else {
			typo.setValue("");
		}
	
	}

	private void editTitle() {
		Entry<String, URL[]> bookmark = this.bookmark;

		setUpKeyboard();
//			//					this.getAbove().setResource(Color.darkGray);
//			this.getNormal().setTransparency(0.5f);
//			typo = new TyPo(this.getAbove(), getApp().getSafeTitleHorizontal(), 200, this
//					.getAbove().getWidth() - (2 * getApp().getSafeTitleHorizontal())-75, 15,
//					true);
//			typo.setExitOptions(TyPo.EXIT_CANCEL, TyPo.EXIT_FINISHED,
//					true, true, false, false);
//			typo.setActivateOnFocus(true);
//			this.setFocus(typo);
//			typo.setFocus();
//			// prevent NPE when typo's list clears and set focus to the default
//			setFocusDefault(typo);
			
			typo.setKeyboard(Keyboard.DEFAULT);
			typo.setActions("editTitle", "editCancel"); //$NON-NLS-1$ //$NON-NLS-2$
			typo.setValue(bookmark.getKey());

	}

	private void setFolder(String folder) {
		URL urlBase = this.bookmark.getValue()[0];
		URL icon = null;
		if(this.bookmark.getValue().length > 1)
			icon = this.bookmark.getValue()[1];
		String title = this.bookmark.getKey();
		
		// delete old bookmark
		((WebBrowserApp)getBApp()).removeBookmark(urlBase, title, icon);		
		
		// save new bookmark with new folder
		((WebBrowserApp)getBApp()).saveBookmark(urlBase, title, folder, icon);		
	}

	@Override
	public boolean handleEnter(Object arg, boolean isReturn) {
		showBackground();
		if(isReturn) {
			if(handleReturnArg(arg)) {
				return true;
			}
		}
		this.folders = ((WebBrowserApp) getBApp()).getBookmarkFolders();
		if(arg instanceof Entry<?, ?>) {
			// this bookmark doesn't include the parent folder in its key.
			// That's in this.folder
			this.bookmark = (Entry<String, URL[]>)arg;
		}
		setupList();
		return super.handleEnter(arg, isReturn);
	}
	
	@Override
	public boolean handleExit() {
		disposeBackground();
		return super.handleExit();
	}

	private boolean handleReturnArg(Object arg) {
		if(arg == null) return false;
		String action = arg.toString();
		WebBrowserApp web = (WebBrowserApp) getBApp();
		if (action.equals("newFolder")) { //$NON-NLS-1$
			String newFolder = null;
			if(typo != null)
				newFolder = typo.getValue();
			
			newFolder = WebBrowserApp.readablePathToFolder(newFolder);

			setFolder(newFolder);
			getBApp().pop();
			
			shutDownKeyboard();
			return true;
		} else if (action.equals("newFolderCancel")) { //$NON-NLS-1$
			this.setFocus(list);
			this.getNormal().setTransparency(0.0f);
			((WebBrowserApp)getBApp()).closeTyPo();
			typo = null;

			return true;
		} else if (action.equals("editTitle")) { //$NON-NLS-1$
			String newTitle = null;
				if(typo != null)
					newTitle = typo.getValue();
//				} else {
//				newTitle = bkey.getValue();
			Entry<String, URL[]> bookmark = this.bookmark;
			if (newTitle != null && newTitle.trim().length() > 0) {
				((WebBrowserApp)getBApp()).moveBookmark(this.folder, this.indexInFolder, this.indexInFolder, newTitle);
//				//FIXME create .renameBookmark instead so it keeps its spot.
//				((WebBrowserApp) getBApp()).removeBookmark(
//						bookmark.getValue()[0], bookmark.getKey(), bookmark.getValue()[1]);
//				//FIXME problem is file: getting used in folder name when saving a bookmark!
////((WebBrowserApp)getBApp()).debugOut("saving bookmark as "+this.folder+" / "+newTitle +" key was "+bookmark.getKey());
//				((WebBrowserApp) getBApp()).saveBookmark(
//						bookmark.getValue()[0], newTitle, this.folder, bookmark.getValue()[1]);

				// return to bookmark screen.
				//FIXME *** return with renamed bookmark selected?? WBA.renamebookmark will fix that.
				getBApp().pop();
//				// OR just keep this screen and update 
//				list.set(0, "Edit Name: "+ this.bookmark.getKey().substring(folder.length()));

			}//TODO return false (after cleaning up) if we had no title.


			shutDownKeyboard();
			//		this.setFocus(list);
			//		if(isTypo) {
			//			// change default back
			//			setFocusDefault(list);
			////			this.getAbove().setResource(null);
			//			this.getNormal().setTransparency(0.0f);
			//			typo.remove();
			//			typo = null;
			////		} else {
			////			this.getNormal().setVisible(true);
			////			bkey.remove();
			////			cancel.remove();
			////			ok.remove();
			////			bkey = null;
			//		}

			return true;
		} else if (action.equals("editCancel")) { //$NON-NLS-1$
			this.setFocus(list);
			// this.getAbove().setResource(null);
			this.getNormal().setTransparency(0.0f);
			((WebBrowserApp)getBApp()).closeTyPo();
			//			typo.setVisible(false);
			//			 typo.remove(); // now it's a app-wide resource
			typo = null;

			return true;
		}
		
		
		return false;
	}
	
	
	private static int getRowFontSize(BScreen screen) {
		float ratio = ((float) screen.getHeight()) / 480f;
		return (int) (15*(1.0+(ratio-1.0)/2));
	}

	private static class FolderList extends BList {

		private BookmarkFolderScreen screen;

		public FolderList(BookmarkFolderScreen screen, BView parent, int x, int y,
				int width, int height, int rowHeight) {
			super(parent, x, y, width, height, rowHeight);
			this.screen = screen;
		}

		/**
		 * Handle focus movement.
		 */
		public boolean handleFocus(boolean isGained, BView gained, BView lost) {
			// work around a bug in BList when gained is null.
			if (!isGained && gained == null) {
				refresh();
				return true;
			} else {
				return super.handleFocus(isGained, gained, lost);
			}
		}
		
		@Override
		protected void createRow(BView parent, int index) {
//			debugOut("create row " + index);
			if (screen.folders == null || index < size()) {
				if (index < screen.folderStartIndex) {
					BButton button = new BButton(parent, 0, 0,
							parent.getWidth(), parent.getHeight());
					button.setTranslation(10, 0);
					String text = String.valueOf(get(index));
WebBrowserApp.debugOut("spot "+index+": "+text);
					int fontSize = this.getRowHeight();
					fontSize = getRowFontSize(getScreen())+5;
					String fontString = "default-"+(fontSize)+"-bold.font"; //$NON-NLS-1$ //$NON-NLS-2$
					restrictWidth(text, fontString, button.getWidth());
					button.setResource(button.createText(
							fontString, Color.white,
							UnicodeHelper.fixTextForRendering(text)), RSRC_HALIGN_LEFT|RSRC_VALIGN_CENTER);
//debugOut("action:" + get(index)+" "+screen.folderStartIndex);
				} else if (screen.folders != null) {
//					debugOut("bkmrk " + get(index));
					BButton button = new BButton(parent, 0, 0,
							parent.getWidth(), parent.getHeight());
					String title = screen.folders.get(index
							- screen.folderStartIndex);
					
					title = title.replace(WebBrowserApp.BOOKMARK_FOLDER_SEPARATOR.charAt(0), '/');
WebBrowserApp.debugOut("folder "+index+": "+title);
					
					int fontSize = this.getRowHeight();
					String fontString = "default-"+fontSize+".font"; //$NON-NLS-1$ //$NON-NLS-2$
					title = restrictWidth(title, fontString, button.getWidth());
					button.setResource(button.createText(
							fontString, Color.yellow, 
							MessageFormat.format(Messages.getString("BookmarkFolderScreen.Folder_Entry"),  //$NON-NLS-1$
									UnicodeHelper.fixTextForRendering(title))), RSRC_HALIGN_LEFT|RSRC_VALIGN_CENTER);
				}
			} 
//			else debugOut("huh?");
		}
		private String restrictWidth(String title, String fontString, int width) {
			if(width <= 10 || title == null) return title;
			FontInfo fontInfo = ((WebBrowserApp)getApp()).getFontInfos().get(fontString);
			if(fontInfo != null) {
				while(fontInfo.measureTextWidth(title)>width) {
					title = title.substring(0, title.length()-1);
				}
				return title;
			} else {
				int pxPerChar = 12;
				if(title.length() > width/pxPerChar) {
					return title.substring(0, width/pxPerChar);
				}
			}
			return title;
		}

		@Override
		public boolean handleKeyPress(int code, long rawcode) {
			switch (code) {
			case V49IHmeProtocol.KEY_OPT_ACTION_B:
				postEvent(new BEvent.Action(this, "back")); //$NON-NLS-1$
				return true;
			case KEY_ADVANCE:
				if(size() <= 0)
					return false;
				if(getFocus() >= size()-1) {
					setFocus(0, true);
				} else {
					setFocus(size()-1, true);
				}
				play(V49BApplication.SOUND_SELECT);
				return true;
			case KEY_SELECT:
			case KEY_FORWARD:
				postEvent(new BEvent.Action(this, "FOLDER")); //$NON-NLS-1$
				return true;
			}
			return super.handleKeyPress(code, rawcode);
		}
	}

	public void showBackground() {
//		if(((WebBrowserApp)getBApp()).isStreamTypeVideo()) {
			Object resource = new Color(0, 0, 80);
			try {
				resource = getBApp().getSkin().get("setup").getResource();
			} catch (Exception e) {}
			getBelow().setResource(resource, RSRC_IMAGE_BESTFIT);
//		} else {
//			StreamResource videoLoop = createStream(WebBrowserApp.STREAM_LOOP_BLUE_SETUP, "video/mpeg", true); //$NON-NLS-1$
//			getBelow().setResource(videoLoop);
//			flush();
//		}
	}
	public void disposeBackground() {
		if(getBelow().getResource() != null && getBelow().getResource() instanceof StreamResource 
				&& ((StreamResource)getBelow().getResource()).getURI().equals(WebBrowserApp.STREAM_LOOP_BLUE_SETUP)) {
			getBelow().getResource().remove();
			getBelow().clearResource();
			flush();
		}
	}

	private void setUpKeyboard() {
			//					this.getAbove().setResource(Color.darkGray);
			this.getNormal().setTransparency(0.5f);
			if(typo == null) {
				typo = ((WebBrowserApp)getBApp()).showTyPo();
//				TyPo.setUserWordListAllowed(false);
//				typo = new TyPo(this.getAbove(), ((WebBrowserApp)getApp()).getSafeTitleHorizontal(), 200, this
//						.getAbove().getWidth() - (2 * ((WebBrowserApp)getApp()).getSafeTitleHorizontal())-75, 15,
//						true);
			}
//			typo.setExitOptions(TyPo.EXIT_CANCEL, TyPo.EXIT_FINISHED,
//					true, true, false, false);
			
//			typo.setActivateOnFocus(true);
//			this.setFocus(typo);
//			typo.setFocus();
//			// prevent NPE when typo's list clears and set focus to the default
//			setFocusDefault(typo);

			//					getBApp().push(new StringGetterScreen(null, getBApp()),
			//							TRANSITION_FADE,
			//							new String[] { "Enter URL", this.typedURL });

	}

	private void shutDownKeyboard() {
		this.setFocus(list);
			// change default back
			setFocusDefault(list);
//				this.getAbove().setResource(null);
			this.getNormal().setTransparency(0.0f);
			((WebBrowserApp)getBApp()).closeTyPo();
//			typo.remove();
			typo = null;
	}
	
}
