package com.blackledge.david.html.hme;

import java.awt.Color;
import java.awt.Dimension;
import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.net.HttpURLConnection;
import java.net.MalformedURLException;
import java.net.URL;
import java.nio.charset.Charset;
import java.nio.charset.CharsetEncoder;
import java.util.ArrayList;
import java.util.List;
import java.util.Map;
import java.util.StringTokenizer;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

// JSOUP
import org.jsoup.Connection.Response;
import org.jsoup.Jsoup;
import org.jsoup.nodes.Document;
import org.jsoup.nodes.Element;
import org.jsoup.select.Elements;
import org.jsoup.select.NodeTraversor;
import org.jsoup.select.Selector;
// /JSOUP

import com.blackledge.david.html.HTMLChunk;
import com.blackledge.david.html.HTMLChunk.ChunkType;
import com.blackledge.david.html.hme.WebBrowserApp.MEDIA;
import com.blackledge.david.html.hme.WebBrowserApp.STATUS;
import com.blackledge.david.html.CSSSpecificity;
import com.blackledge.david.html.HTMLFormField;
import com.blackledge.david.html.HTMLFormInput;
import com.blackledge.david.html.HTMLFormSelect;
import com.blackledge.david.html.HTMLFormTextarea;
import com.blackledge.david.html.HTMLHRChunk;
import com.blackledge.david.html.HTMLImageChunk;
import com.blackledge.david.html.HTMLParserCallback;
import com.blackledge.david.html.HTMLRow;
import com.blackledge.david.html.HTMLStyleChunk;
import com.blackledge.david.html.HTMLTableCell;
import com.blackledge.david.html.HTMLTableRow;
import com.blackledge.david.html.HTMLTextChunk;
import com.blackledge.david.html.UnicodeHelper;

import com.tivo.core.ds.TeDict;
import com.tivo.hme.bananas.BApplication;
import com.tivo.hme.bananas.BView;
import com.tivo.hme.sdk.HmeEvent.FontInfo;
import com.tivo.hme.sdk.HmeEvent.ResourceInfo;
import com.tivo.hme.sdk.HmeEvent.FontInfo.GlyphInfo;
import com.tivo.hme.sdk.HmeEvent;
import com.tivo.hme.sdk.IHmeEventHandler;
import com.tivo.hme.sdk.IHmeProtocol;
import com.tivo.hme.sdk.Resource;
import com.tivo.hme.sdk.StreamResource;
import com.tivo.hme.sdk.View;

public class HTMLBreakdown 
{

	// pieces: text block with font and style and color.
	//  sub-blocks broken down by row, given a width (and "scale"?) to work with
	// image with size, etc.
	// any block might be a link (additional style will be added)
	// form field (textfield, button, others in future)
	// later:
	// HR?
	// tables

	private static final int DEFAULT_IMAGE_WIDTH = 50;
	private static final int DEFAULT_IMAGE_HEIGHT = 50;
	
	private static void debugOut(String output) {
		WebBrowserApp.debugOut(output);
	}

	public class RenderProgress {
		private URL urlBase;
		private Map<String, FontInfo> fontInfos;
		private BView parent;
		private List<HTMLRow> rows;
		int rowIndex;
		private int maxX;
		private boolean ignoreWrap;
		public boolean isComplete;
		public BRowView rowView;
		public int nextRowY;
		public int restartChunk = 0;
		public int innerX = 0;
//		public BView chunkView;
		private int align;
		public RenderProgress tableCellProgress;
		public int tableCellHeight;
//		public int width;

		public RenderProgress(URL urlBase, Map<String, FontInfo>fontInfos, BView parent, List<HTMLRow> rows, int rowIndex, int maxX, boolean ignoreWrap, int align) {
			this.urlBase = urlBase;
			this.fontInfos = fontInfos;
			this.parent = parent;
			this.rows = rows;
			this.rowIndex = rowIndex;
			this.maxX = maxX;
			this.ignoreWrap = ignoreWrap;
			this.align = align;
		}
		
		@Override
		public String toString() {
			return "row:"+rowIndex+",col:"+restartChunk+"-x:"+innerX+",y:"+nextRowY+",h:"+tableCellHeight+"--"+tableCellProgress;
		}

	}

	private WebBrowserApp bapp;
	public Color fgColor = Color.black;
	public Color linkColor = Color.blue;
	public Color bodyColor = Color.white;
	public String bodyImg;
	public int bodyBgrepeat = -1;
	public int bodyBgposition = -1;
	public String title;
	public String refresh;
	public String embeddedStream;
	public int jumpToRow;
	
	URL icon;
	private Resource waitingOnImageResource;
	public URL baseURL;
	private RSSBreakdown rss;
	
	/** if multiple one-link-lines in a row, combine them on a single line with a separator*/
	static boolean groupIndividualLinks = false;
	static String linkSeparator = " || ";
	static String bulletImage = " @ "; // default - replaced using UnicodeHelper
	static String circleImage = " O ";
	static String squareImage = " [] ";
	
	public HTMLBreakdown(WebBrowserApp bapp) {
		this.bapp = bapp;
		try {
			this.rss = new RSSBreakdown(bapp);
			linkSeparator = bapp.getContext().getBaseURI().toString()+"link-separator.png";
			bulletImage = 
//					bapp.getContext().getBaseURI().toString()+"bullet.png";
					UnicodeHelper.getDiscBulletCharacter();
			circleImage = UnicodeHelper.getCircleBulletCharacter();
			squareImage = UnicodeHelper.getSquareBulletCharacter();
		} catch(Throwable t) {
			// just use the defaults
		}
		
		try {
			textTestISOEncoder = Charset.forName("ISO-8859-1").newEncoder();
		} catch (Throwable t) {
			// just leave it null.  will skip the test
		}
		try {
			textTestUTFEncoder = Charset.forName("UTF-8").newEncoder();
		} catch (Throwable t) {
			// just leave it null.  will skip the test
		}
//debugOut(bulletImage);
//debugOut(linkSeparator);
	}
	
	private Object getResourceDetail(String src) {
		return this.bapp.getResourceCache(src);
	}
	
	private void putResourceDetail(String src, Object obj) {
		this.bapp.putResourceCache(src, obj);
	}

	// encoders used to test that CSS file isn't binary.
	private CharsetEncoder textTestISOEncoder;  
	private CharsetEncoder textTestUTFEncoder;  

	/**
	 * Credit for starting point to: Grekz - http://stackoverflow.com/questions/4521557/automatically-convert-style-sheets-to-inline-style
	 * @param doc
	 * @param urlBase 
	 */
	private void inlineCSS(Document doc, URL urlBase) {
		// style tags with no media attribute, or the media attribute has "tv" or "all" in it.
		String styleSelector = "style:not([media]), ";
		if(bapp.isMediaStyleActive(WebBrowserApp.MEDIA.TV))
			styleSelector += "style[media*=tv], ";
		if(bapp.isMediaStyleActive(WebBrowserApp.MEDIA.SCREEN))
			styleSelector += "style[media*=screen], ";
		if(bapp.isMediaStyleActive(WebBrowserApp.MEDIA.HANDHELD))
			styleSelector += "style[media*=handheld], ";
		styleSelector += "style[media*=all]";
		String linkstyle = "link[rel=stylesheet]";
		Elements els = doc.select(styleSelector+", "+linkstyle);// to get all the style elements
		int totalStyles = els.size();
		int i = 0;
		for (Element e : els) {
			++i;
			bapp.setLoadingStatus(i,totalStyles);
			if(e.tagName().equalsIgnoreCase("style")) {
				String styleRules = e.getAllElements().get(0).data();
				int inlinedStyles = inlineCSS(urlBase, doc, styleRules);
//				debugOut("Inlined "+inlinedStyles+" styles");
				e.remove();
			} else {
//		}
//		Elements links = doc.select(link);
//		for (Element link : links) {
				Element link = e;
				String href = link.attr("href");
				try {
					URL url = WebBrowserApp.getURL(urlBase, href);
					String styleRules = bapp.getRawText(url);
					if(styleRules != null) {
						// protect against binary data (seen: a link stylesheet pointed to a .otf font file)
						if ((textTestISOEncoder == null || textTestISOEncoder.canEncode(styleRules))
								&& (textTestUTFEncoder == null || textTestUTFEncoder.canEncode(styleRules))) {
							int inlinedStyles = inlineCSS(urlBase, doc, styleRules);
//							debugOut("Inlined "+inlinedStyles+" styles from CSS link "+href);
						} else {
							debugOut("Skipping binary 'stylesheet' "+url);
						}
					}
				} catch (Exception e1) {
					debugOut("can't use CSS link: "+e1.getLocalizedMessage());
				}
				link.remove();
			}
		}
	}

	private int inlineCSS(URL urlBase, Document doc, String styleRules) {
		styleRules = preprocessCSS(urlBase, styleRules);
		
		int inlinedStyles = processCSS(doc, styleRules, true);
		return inlinedStyles;
	}

	public static int processCSS(Document doc, String styleRules, boolean includeSpecificity) {
		// StringTokenizer skips contiguous delimiters, but we need them all to
		// keep our selector/properties sequence right. So we make sure none are
		// next to each other.
		styleRules = styleRules.replace("{}","{ }").replace("}{", "} {");
		StringTokenizer st = new StringTokenizer(styleRules, "{}");
		int inlinedStyles = 0;
		while (st.countTokens() > 1) {
			String selector = st.nextToken(), properties = st.nextToken();
//			if(selector.toLowerCase().endsWith("li") && properties.indexOf("disc")>-1) {
//				debugOut("push on "+selector+" with "+properties);
//			}
			if(selector == null || selector.trim().length() == 0) continue;
			// to enforce precedence, we have to break multi-selectors into single selectors.
			try {
				if(includeSpecificity) {
					String selectors[] = selector.split(",");
					for(String selectorEntry : selectors) {
						CSSSpecificity specificity = new CSSSpecificity(selectorEntry);
						
						Elements selectedElements = doc.select(selectorEntry);
						for (Element selElem : selectedElements) {
							String oldProperties = selElem.attr("style");
							selElem.attr(
									"style",
									specificity.concatenateProperties(
											oldProperties, properties));
							++inlinedStyles;
						}
					}
				} else {
					// just prepend the new values
					Elements selectedElements = doc.select(selector);
					for (Element selElem : selectedElements) {
						String oldProperties = selElem.attr("style");
						if (oldProperties == null) {
							oldProperties = "";
						} else if (!oldProperties.trim().startsWith(";")
								&& !properties.trim().endsWith(";")) {
							properties += ";";
						}
						selElem.attr(
								"style",
								properties + oldProperties);
						++inlinedStyles;
					}
				}
			} catch (Selector.SelectorParseException ex) {
				// just assume this is an unsupported selector type... skip it.
				// e.g. a:link or a:active
				//TODO this skips the entire selector SET, even if only one of the comma-separated entries was bad.
				debugOut("Skipping selector "+ex.getLocalizedMessage());
				continue;
			} catch (Exception ex) {
				// just assume this is an unsupported selector type... skip it.
				// e.g. a:link or a:active
				//TODO this skips the entire selector SET, even if only one of the comma-separated entries was bad.
				debugOut("Skipping selector "+ex.getLocalizedMessage());
				continue;
			}
		}
		return inlinedStyles;
	}

	private String preprocessCSS(URL urlBase, String styleRules) {
		styleRules = styleRules.trim().replace('\n',' ');
		// clean out comments
		String htmlComment1 = "<!--";
		if(styleRules.startsWith(htmlComment1)) styleRules = styleRules.substring(htmlComment1.length()).trim();
		String htmlComment2 = "-->";
		if(styleRules.endsWith(htmlComment2)) styleRules = styleRules.substring(0, styleRules.length()-htmlComment2.length()).trim();
		
		// add a safety brace... missing end brace can cause Pattern to loop forever.
		// (the actual Matcher.find() call... a bug in Java)
		/*
		 * Regular Expression	@media\b([^{}]*(tv|screen|handheld|all)[^{}]*)\{([^{}]*((\{[^{}]*\})*[^{}]*)*)\}
		 * as a Java string	"@media\\b([^{}]*(tv|screen|handheld|all)[^{}]*)\\{([^{}]*((\\{[^{}]*\\})*[^{}]*)*)\\}"
		 * freezes on (missing final brace):
		 * @media screen and (max-width:480px){  #topsection{ font-size: 12px; width: 100%; }  #bannerarea, footer{ display:none; } }
		 */
		styleRules = styleRules + " } ";

		// get rid of comments - might have commented-out @rules
		String rep1 = removeCSSComments(styleRules);		
		styleRules = rep1;
		
		String at1 = processAtRules(urlBase, styleRules);
		styleRules = at1;
				
		String sel1 = processUnknownSelectors(styleRules);
		styleRules = sel1;
		return styleRules;
	}

	private Pattern cssCommentPattern = Pattern.compile(
			"(?s)"+ // DOTALL flag.
					Pattern.quote("/*")+".*?"+Pattern.quote("*/")
			);
	private String removeCSSComments(String styleRules) {
		return cssCommentPattern.matcher(styleRules).replaceAll(" ");
	}
	
	private Pattern aLinkPattern = Pattern.compile("(A|a)"+Pattern.quote(":")+"link");
	private Pattern linkSelPattern = Pattern.compile(Pattern.quote(":")+"link");
	// includes :link to catch any possible cases not caught above
	private Pattern unsupPseudoSelPattern = Pattern.compile(Pattern.quote(":")+
			"(active|visited|hover|link|focus" +
	// jsoup 1.7.2 supports:
			// root|empty|first-child|last-child|only-child|nth-child|nth-last-child|nth-of-type|nth-last-of-type|first-of-type|last-of-type|only-of-type|
	// common pseudo selectors that jsoup may support in the future
	"|before|after|first-letter|first-line" +
	// seen, but not part of the original standard:
	"|disabled|target" +
	// seen, but not worth listing here, I think
//	"|first-child"+Pattern.quote("+")+"html|last-child"+Pattern.quote("+")+"html"+
			")");
	private String userAgentStyleRules;
	String[] accessNumLinks;

	private String processUnknownSelectors(String styleRules) {
		// currently Jsoup doesn't handle :before :after :first-child :last-child, etc.
		// ... but maybe it will in the future since they are calculable.
		
		// replace A:link with A and :link with A so we can still use them.
		styleRules = aLinkPattern.matcher(styleRules).replaceAll("$1");
		styleRules = linkSelPattern.matcher(styleRules).replaceAll(" a");
		
		// :hover :active :visited - change : to a space
		// we'll never handle those here... jsoup can't detect those cases, but
		// we don't want to break other selectors in the query, so just replace
		// with a "nonsense" tag reference
		styleRules = unsupPseudoSelPattern.matcher(styleRules).replaceAll(" $1");

		return styleRules;
	}
	
	private static final String lb = Pattern.quote("{");
	private static final String rb = Pattern.quote("}");
	private static final String nobrace = "[^" + lb + rb + "]*";
	private static final Pattern ourMediaAtGroup = Pattern.compile( Pattern.quote("@")+"media"+"\\b"+
			"(" + // group (1) for atrule's media types
			nobrace + "(?<media>tv|screen|handheld|all)" + nobrace + // group 2
			")" + // end group for atrule's media types
			lb + 
			"(" + // group (3) for atrule's block content
			nobrace +
			"((" + lb + nobrace + rb + ")*" + // (group 4,5)
			nobrace + ")*" +
			")" + // end group for atrule's block content
			rb);
//	private static final Pattern mediaAtGroup = Pattern.compile(Pattern.quote("@")+"media"+"\\b"+
//			"(" + // group (1) for atrule's media types
//			nobrace +
//			")" + // end group for atrule's media types
//			lb + 
//			"(" + // group (2) for atrule's block content
//			nobrace +
//			"((" + lb + nobrace + rb + ")*" + // (group 3,4)
//			nobrace + ")*" +
//			")" + // end group for atrule's block content
//			rb);
	private static final Pattern atGroup = Pattern.compile(Pattern.quote("@")+"(?:\\w|-)+"+"\\s*"+
			"(" + // group (1) for atrule's media types
			nobrace +
			")" + // end group for atrule's media types
			lb + 
			"(" + // group (2) for atrule's block content
			nobrace +
			"((" + lb + nobrace + rb + ")*" + // (group 3,4)
			nobrace + ")*" +
			")" + // end group for atrule's block content
			rb);
//	private static final Pattern pageAtGroup = Pattern.compile( Pattern.quote("@")+"page"+"\\b"+
//			"(" + // group (1) for atrule's media types
//			nobrace +
//			")" + // end group for atrule's media types
//			lb + 
//			"(" + // group (2) for atrule's block content
//			nobrace +
//			"((" + lb + nobrace + rb + ")*" + // (group 3,4)
//			nobrace + ")*" +
//			")" + // end group for atrule's block content
//			rb);
//	private static final Pattern fontFaceAtGroup = Pattern.compile( Pattern.quote("@")+"font-face"+"\\b"+
//			"(" + // group (1) for atrule's media types
//			nobrace +
//			")" + // end group for atrule's media types
//			lb + 
//			"(" + // group (2) for atrule's block content
//			nobrace +
//			"((" + lb + nobrace + rb + ")*" + // (group 3,4)
//			nobrace + ")*" +
//			")" + // end group for atrule's block content
//			rb);
	private static final Pattern atRule = Pattern.compile(Pattern
			.quote("@") + "[^" + lb + rb + ";]+");
//	private static final Pattern charsetAtRule = Pattern.compile(Pattern
//			.quote("@") + "charset" + "\\s+" + "[^" + lb + rb + ";]*");
//	private static final Pattern namespaceAtRule = Pattern.compile(Pattern
//			.quote("@") + "namespace" + "\\s+" + "[^" + lb + rb + ";]*");
	private static final Pattern importAtRule = Pattern.compile(Pattern
			.quote("@") + "import" + "\\s+" + "[^" + lb + rb + ";]*");
	private static final Pattern ourImportAtRule = Pattern.compile(Pattern
			.quote("@") + "import" + "\\s+" + 
					"(url\\s*\\()?" + // may or may not be a url
					"\\s*[\"\']([^\"\']*)[\"\']\\s*" + // but definitely is in quotes
					"(\\))?" + // end of url
					"([^;]*(?<media>tv|screen|handheld|all)[^;]*|\\s*)" + // no media, or media contains "all" or "tv"
					";");

	private String processAtRules(URL urlBase, String styleRules) {
		//		@import url("/css/main.css") tv, screen, projection;
		//		@import "local.css";
		//		@import must be first, so just process them all before handing content over for pushdown. (repeat, but watch for loops, or just don't repeat) - if media, only "all" or "tv"
		//		Also, preprocess all @media that match "all" or "tv" (ONLY) by breaking them out of their block or eliminating them if they aren't a matching media
		//		 
		//		eat
		//		@charset "encoding"
		//		@page {}
		//		@font-face {}
		//		@namespace [name] "URI"
				
		//		@media type(, type)* { content no braces ({content no braces}content no braces)*}
				// handle @ rules which can cause troublesome nested {}'s that tokenizer couldn't handle.
		
		// also handle media types like" @media screen and (-webkit-min-device-pixel-ratio:0)"

		// process media-compatible imports (recurses one more level of imports)
		String imports1 = processImports(styleRules, urlBase, null);
		styleRules = imports1;
		// eat all other import rules. (including any 3rd set inserted by processing the above!)
		String imports2 = importAtRule.matcher(styleRules).replaceAll("");
		styleRules = imports2;
		
		// expose rules for our media type!
		String media1 = processMediaGroups(styleRules);
		styleRules = media1;
		
		// eat other block rules
		// general rule - seems to work
		String at1 = atGroup.matcher(styleRules).replaceAll("");
		styleRules = at1;
		
//		// eat other media rules.
//		styleRules = mediaAtGroup.matcher(styleRules).replaceAll("");
//		// eat page rules.
//		styleRules = pageAtGroup.matcher(styleRules).replaceAll("");
//		// eat font-face rules.
//		styleRules = fontFaceAtGroup.matcher(styleRules).replaceAll("");
		
		// eat other at (non-block) rules.
		String at2 = atRule.matcher(styleRules).replaceAll("");
		styleRules = at2;
//		// eat charset rules.
//		styleRules = charsetAtRule.matcher(styleRules).replaceAll("");
//		// eat namespace rules.
//		styleRules = namespaceAtRule.matcher(styleRules).replaceAll("");
		
//		if(styleRules.indexOf("@")>-1) {
//			debugOut("OOPS!!! "+styleRules);
//		}
		return styleRules;
	}

	private String processMediaGroups(String styleRules) {
		Matcher mediaMatcher = ourMediaAtGroup.matcher(styleRules);
		// restrict based on bapp.isMediaStyleActive
		StringBuffer media1 = new StringBuffer();
		while(mediaMatcher.find()) {
			boolean keep = isActiveMedia(mediaMatcher);
			if(keep) {
				mediaMatcher.appendReplacement(media1, " $3 ");
			} else {
				mediaMatcher.appendReplacement(media1, " ");
			}
		}
		mediaMatcher.appendTail(media1);
		styleRules = media1.toString();
		return styleRules;
	}

	private boolean isActiveMedia(Matcher mediaMatcher) {
		boolean keep = true;
		String media = mediaMatcher.group("media");
		if (media != null) {
			if (media.equals("handheld")) {
				if (bapp.isMediaStyleActive(MEDIA.HANDHELD))
					keep = true;
				else
					keep = false;
			}
			else if (media.equals("tv")) {
				if (bapp.isMediaStyleActive(MEDIA.TV))
					keep = true;
				else
					keep = false;
			}
			else if (media.equals("screen")) {
				if (bapp.isMediaStyleActive(MEDIA.SCREEN))
					keep = true;
				else
					keep = false;
			}
			else if(media.equals("all")){
				keep = true;
			} else {
				keep = false;
			}
		}
		return keep;
	}

	private String processImports(String styleRules, URL baseURL2, URL parentURL) {
		StringBuffer updatedRules = new StringBuffer();
		Matcher matcher = ourImportAtRule.matcher(styleRules);
		while(matcher.find()) {
			// restrict based on bapp.isMediaStyleActive
			boolean keep = isActiveMedia(matcher);
			if(!keep) {
				matcher.appendReplacement(updatedRules, "");
			} else {
				// should have 5 or 3 groups... 4 means something matched weird. 
				// (last groups are media, and then either tv or all)
				int contentGroup;
				if(matcher.groupCount() == 3) {
					contentGroup = 1;
				} else if(matcher.groupCount() == 5) {
					contentGroup = 2;
				} else {
					debugOut("BAD CSS import! "+matcher.groupCount()+" groups not right "+matcher.group());
					for(int i = 0 ; i <= matcher.groupCount() ; ++i) {
						debugOut("group "+i+": "+matcher.group(i));
					}
					matcher.appendReplacement(updatedRules, "");
					continue;
				}
				String importURLString = matcher.group(contentGroup);			
				// load import file, append to updatedRules
				matcher.appendReplacement(updatedRules, "");
				try {
					URL importURL = WebBrowserApp.getURL(baseURL2, importURLString);
					String importContent = bapp.getRawText(importURL);
					if (importContent != null) {
						// recurse just one layer down.
	//					if(baseURL2.equals(parentURL)) {
						if(parentURL == null) {
							importContent = processImports(importContent, importURL, baseURL2);
						}
						updatedRules.append(importContent);
						// add a safety brace... missing end brace can cause Pattern to loop forever.
						updatedRules.append(" } ");
	//					debugOut("CSS IMPORT! "+importURL);
					} else {
						debugOut("BAD CSS import! - download empty for "+importURLString+" ("+importURL+" - "+matcher.group()+")");
					}
				} catch(Exception ex) {
					debugOut("BAD CSS import! for "+matcher.group()+" - "+ex);
				}
			}
		}
		matcher.appendTail(updatedRules);
		
		// get rid of comments again - might have @imported content with comments
		String updated = removeCSSComments(updatedRules.toString());		
		
		return updated;
	}

	public List<HTMLRow> getRows(URL urlBase, Response soupResp) throws IOException {
		Document doc;
		String contentType = soupResp.contentType().toLowerCase();
//		application/rss+xml, application/rdf+xml, application/atom+xml, application/xml, text/xml
		if(contentType.contains("text/xml") || 
				(contentType.contains("application/") && contentType.contains("xml"))) {
			bapp.setLoadingStatus(STATUS.PARSING_XML);
			doc = convertXMLToHTML(urlBase, soupResp);
		} else {
			//TODO handle empty comment - parse() will blow up.
//			soupResp.bodyAsBytes().length == 0
			doc = soupResp.parse();
		}
		return getRows(urlBase, doc, soupResp.header("Refresh"));
	}
	
	public List<HTMLRow> getRows(URL urlBase, Document doc, String refreshHeader) throws IOException {
		HTMLParserCallback pc = new HTMLParserCallback(urlBase, groupIndividualLinks, linkSeparator, 
				bulletImage, circleImage, squareImage,
				this.bapp.isLayingOutTables());
		
		bapp.setLoadingStatus(STATUS.PARSING_CSS);
		inlineCSS(doc, urlBase);
		inlineUserAgentCSS(doc, urlBase);
//java.io.FileWriter writer = new java.io.FileWriter("C:\\Users\\David\\page.html");writer.write(doc.toString());writer.close();
		NodeTraversor traversor = new NodeTraversor(pc);
		bapp.setLoadingStatus(STATUS.PARSING_HTML);
		traversor.traverse(doc); // walk the DOM, and call .head() and .tail() for each node
		
		if (pc.fgColor != null) {
			fgColor = pc.fgColor;
		}
		if (pc.linkColor != null) {
			linkColor = pc.linkColor;
		}
		if (pc.bodyColor != null) {
			bodyColor = pc.bodyColor;
		}
		if (pc.baseURL != null) {
			baseURL = pc.baseURL;
		}
		bodyImg = pc.bodyImg;
		bodyBgposition = pc.bodyBgposition;
		bodyBgrepeat = pc.bodyBgrepeat;
		title = pc.title;
		refresh = pc.refresh;
		jumpToRow = pc.mainContentRow;
		accessNumLinks = pc.accessNumLinks;
		embeddedStream = pc.embeddedStream;
		if(refresh == null || refresh.trim().length() == 0) {
			refresh = refreshHeader;
		}
		icon = pc.icon;
		return pc.rows;
	}
	
	private void inlineUserAgentCSS(Document doc, URL urlBase) {
		if(userAgentStyleRules == null) {
			userAgentStyleRules = getFileText("/com/blackledge/david/html/user-agent.css");
		}
		if(userAgentStyleRules != null) {
			String styleRules = preprocessCSS(urlBase, userAgentStyleRules);
			
			// inline, but don't use specificity
			//TODO technically not quite right, but I think it's close enough.
			int inlinedStyles = processCSS(doc, styleRules, false);
//			debugOut("Inlined "+inlinedStyles+" User Agent styles");
		}
	}

	private String getFileText(String resourceFile) {
		try {
			InputStream stream = HTMLBreakdown.class
					.getResourceAsStream(resourceFile);

			BufferedReader reader = new BufferedReader(new InputStreamReader(
					stream));
			StringBuffer result = new StringBuffer();
			for (String line = reader.readLine(); line != null; line = reader.readLine()) {
				result.append(line).append("\n");
			}
			stream.close();
			return result.toString();
		} catch (IOException e) {
			debugOut("GETTING FILE " + resourceFile + ": " + e);
			return null;
		}
	}

	private Document convertXMLToHTML(URL urlBase, Response soupResp)
			throws IOException {
		// text/xsl text/xml
		
		String body = soupResp.body();
		Document result = convertXMLToHTML(urlBase, body);
		if(result == null) {
			return soupResp.parse();
		}
		return result;
	}
	
	Document convertXMLToHTML(URL urlBase, String body)
			throws IOException {
		body = rss.convertXMLToHTML(urlBase, body);
		if(body != null) {
			return Jsoup.parse(body);
		} else {
			return null;
		}
	}

	public List<HTMLRow> getRows(URL urlBase, InputStream htmlStream) throws IOException {
		HTMLParserCallback pc = new HTMLParserCallback(urlBase, groupIndividualLinks, linkSeparator,
				bulletImage, circleImage, squareImage,
				this.bapp.isLayingOutTables());
////			HTMLEditorKit kit = new HTMLEditorKit(); 
////			HTMLDocument doc = (HTMLDocument) kit.createDefaultDocument(); 
////			doc.putProperty("IgnoreCharsetDirective", Boolean.TRUE);
////			Reader HTMLReader = new InputStreamReader(urlBase.openConnection().getInputStream()); 
////			kit.read(HTMLReader, doc, 0); 
////
////	    //  Get an iterator for all HTML tags.
////	    ElementIterator it = new ElementIterator(doc); 
////	    Element elem; 
////	    
////	    while( elem = it.next() != null  )
////	    { 
////	      if( elem.getName().equals(  "img") )
////	      { 
////	        String s = (String) elem.getAttributes().getAttribute(HTML.Attribute.SRC);		
//			
//			HTMLDocument doc = (HTMLDocument) new HTMLEditorKit().createDefaultDocument();
//			//TODO BufferedReader didn't help.
//			doc.getParser().parse(new BufferedReader(new InputStreamReader(htmlStream)), pc, true);

		// JSOUP
			Document doc = Jsoup.parse(htmlStream, null, urlBase.toString());
//			Document doc = Jsoup.connect(urlBase.toString()).get();
			NodeTraversor traversor = new NodeTraversor(pc);
			traversor.traverse(doc); // walk the DOM, and call .head() and .tail() for each node
		// /JSOUP
		
		
		if (pc.fgColor != null) {
			fgColor = pc.fgColor;
		}
		if (pc.linkColor != null) {
			linkColor = pc.linkColor;
		}
		if (pc.bodyColor != null) {
			bodyColor = pc.bodyColor;
		}
		if (pc.baseURL != null) {
			baseURL = pc.baseURL;
		}
		bodyImg = pc.bodyImg;
		bodyBgposition = pc.bodyBgposition;
		bodyBgrepeat = pc.bodyBgrepeat;
		title = pc.title;
		icon = pc.icon;
//for(int i = 0 ; i < pc.rows.size() ; ++i) {
//	debugOut(i+"~~"+pc.rows.get(i));
//}
		return pc.rows;
	}
	
	public RenderProgress renderRow(URL urlBase, Map<String, FontInfo>fontInfos, BView parent, List<HTMLRow> rows, int rowIndex, int maxX, boolean ignoreWrap, int align) {
		RenderProgress progress = new RenderProgress( urlBase, fontInfos,  parent,  rows,  rowIndex,  maxX,  ignoreWrap, align);
		progress.isComplete = true;
		return renderRow(progress);
	}

	//FIXME copied from WebScreen
	private void removeAllChildren(View parent) {
		if(parent == null) return;
		// collect the views before removing them, otherwise the getChild call
		// gets thrown by the remove calls.
		ArrayList<View> children = new ArrayList<View>();
		for(int i = 0 ; i < parent.getChildCount() ; ++i) {
			 children.add( parent.getChild(i));
		}
		
		for(View child : children) {
			removeAllChildren(child);
			if(child.getResource() != null) {
				if(child.getResource().getClass().getName().contains("Text")) {
					child.getResource().remove();
				}
			}
			child.clearResource(); // resource's own finalize will remove itself from tivo
//			// make sure focus gets cleared if the one we're removing has it.
//			if(child.hasFocus()) {
//				this.setFocus(null);
//			}
			child.remove();
		}
	}

	public RenderProgress renderRow(RenderProgress prevProgress) {
		URL urlBase = prevProgress.urlBase;
		// FIXME experimental - maybe should check for null, first? are these even equivalent? this is weird
		this.baseURL = urlBase;
		Map<String, FontInfo> fontInfos = prevProgress.fontInfos;
		BView parent = prevProgress.parent;
		List<HTMLRow> rows = prevProgress.rows;
		int rowIndex = prevProgress.rowIndex;
		int maxX = prevProgress.maxX;
		boolean ignoreWrap = prevProgress.ignoreWrap;
		int align = prevProgress.align;
		
		RenderProgress progress = new RenderProgress( urlBase, fontInfos,  parent,  rows,  rowIndex,  maxX,  ignoreWrap, align);

		//id when continuing from prevProgress, id current chunk in returned incomplete progress
		BRowView rowBView;
		int innerX;
		
		//FIXME !!!!!!!!!!! just brute force make it work for now... totally rebuild row
		//FIXME need to do proper restart that doesn't screw up on first pass like code currently does
		if(!prevProgress.isComplete) {
			if(prevProgress.rowView != null) {
				removeAllChildren(prevProgress.rowView);
				prevProgress.rowView.remove();
			}
			prevProgress.rowView = null;
			prevProgress.restartChunk = 0;
			prevProgress.isComplete = true;
			prevProgress.tableCellProgress = null;
		}
		
		if (prevProgress.isComplete) {
			// TODO init with proper sizes?
			rowBView = new BRowView(parent, 0, 0, parent.getWidth(), 5);
			rowBView.rowIndex = rowIndex;
			innerX = 0;
		} else {
			rowBView = prevProgress.rowView;
			innerX = prevProgress.innerX;
// now we don't create the view until the image is figured out
//			// do final adjustment of the asynch completion of the chunk view.
//			innerX = adjustChunkInrow(rowBView, innerX, prevProgress.chunkView, prevProgress.width);
		}
		progress.isComplete = false;
		progress.rowView = rowBView;
		
//debugOut(rowIndex+"--");
		HTMLRow nextRow = null;
		synchronized(rows) {
		HTMLRow row = rows.get(rowIndex);
		if(row instanceof HTMLTableRow) {
			setTableWidths(((HTMLTableRow)row), rows, rowIndex, rowBView.getWidth(), fontInfos);
		}
//		setRowBG(rowBView, row);
		int startChunk = prevProgress.restartChunk;
		for(int chunkIndex = startChunk ; chunkIndex < row.getHTMLChunks().size() ; ++chunkIndex) {
			HTMLChunk chunk = row.getHTMLChunks().get(chunkIndex);
			// except tables, if we're past our max or we've moved one thing to next row, move the rest there.
			if((!(row instanceof HTMLTableRow)) && (innerX > maxX || nextRow != null)) {
//debugOut("{"+chunk.toString()+"}");
				nextRow = addChunk(nextRow, chunk, row);
				continue;
			}
			if(nextRow != null) {
				debugOut("WRAPPING TABLE ROW ... PROBLEM!!!");
			}
			RenderResult chunkResult = null;
//debugOut((row instanceof HTMLTableRow?"TABLE ":"")+"render "+chunkIndex+":"+chunk.getType()+" ("+innerX+"/"+maxX+")");
			switch(chunk.getType()) {
			case TEXT:
				if(!(row instanceof HTMLTableRow))
				chunkResult = renderTextChunk(fontInfos, (HTMLTextChunk)chunk, rowBView, maxX-innerX, maxX);
				break;
			case IMAGE:
				if(!(row instanceof HTMLTableRow))
				chunkResult = renderImageChunk((HTMLImageChunk)chunk, rowBView, innerX, maxX-innerX, urlBase);
				break;
			case FORM:
				if(!(row instanceof HTMLTableRow))
				chunkResult = renderFormChunk(fontInfos, (HTMLFormField)chunk, rowBView, innerX, maxX-innerX, urlBase);
				break;
			case TABLECELL:
//				debugOut("TABLECELL renderTableCellChunk(...idx/size="+chunkIndex+"/"+row.getHTMLChunks().size()+", innerX="+innerX+")");
				chunkResult = renderTableCellChunk(fontInfos, (HTMLTableCell)chunk, chunkIndex, row.getHTMLChunks().size(), innerX, rowBView, urlBase, prevProgress.tableCellProgress);
//				debugOut("TABLECELL renderTableCellChunk() chunkResult..."+(chunkResult.view==null?"NULL view":"view.x="+chunkResult.view.getX()+" view.width="+chunkResult.view.getWidth())+" width="+chunkResult.width);
				break;
			case HR:
				if(!(row instanceof HTMLTableRow))
				chunkResult = renderHRChunk((HTMLHRChunk)chunk, rowBView, innerX, maxX-innerX );
				break;
			}
//			debugOut("chunkRendered");
			if(chunkResult != null) {
				if(chunkResult.isWaitforResource()) {
					if(chunk.getType().equals(ChunkType.TABLECELL)) {
						progress.restartChunk = chunkIndex;
						progress.tableCellProgress = chunkResult.tableCellProgress;
					} else {
						progress.restartChunk = chunkIndex;
//						progress.restartChunk = chunkIndex+1;
					}
					progress.innerX = innerX;
//					progress.chunkView = chunkResult.view;
//					progress.width = chunkResult.width;
					//FIXME we created a view... need to consider it might need to wrap.
					return progress;
				}
				innerX = adjustChunkInrow(rowBView, innerX, chunkResult.view, chunkResult.width);
//if(chunk.getType().equals(ChunkType.TABLECELL))
//  debugOut("TABLECELL adjusted...innerX now "+innerX+(chunkResult.view==null?"NULL view":"view.x="+chunkResult.view.getX()+" view.width="+chunkResult.view.getWidth())+" width="+chunkResult.width);
//else if(chunk.getType().equals(ChunkType.IMAGE))				
//	  debugOut("IMAGE adjusted...innerX now "+innerX+(chunkResult.view==null?"NULL view":"view.x="+chunkResult.view.getX()+" view.width="+chunkResult.view.getWidth())+" width="+chunkResult.width);
				if(chunkResult.remainingChunk != null) {
if(row instanceof HTMLTableRow) {
	debugOut("TABLE ROW WITH WRAP??? Not good");
} else {
//					debugOut("add remaining chunk");
//debugOut("{"+chunkResult.remainingChunk.toString()+"}");
					nextRow = addChunk(nextRow, chunkResult.remainingChunk, row);
}
				}
			} else {
				// must be unknown chunk type
				//FIXME do something. or just don't render at all
//if(!(row instanceof HTMLTableRow))
if(!(chunk instanceof HTMLTextChunk && chunk.toString().trim().length() == 0))
debugOut("UNKNOWN CHUNK TYPE??"+chunk.getType()+":"+chunk);
			}
		}

		//FIXME this was moved from below "perform alignment" section to see if it fixes last row cell bg size issue
		setRowBG(rowBView, rows, rowIndex);
		
		// adjust cells in table row
		if(row instanceof HTMLTableRow) {
			// increase row height based on its requested height
			if(((HTMLTableRow) row).height > rowBView.getHeight()) {
				debugOut("ROW increasing height from "+rowBView.getHeight()+" to "+((HTMLTableRow) row).height);
				rowBView.setSize(rowBView.getWidth(), ((HTMLTableRow) row).height);
			}
			// increase cell heights to match
			for(int i = 0 ; i < rowBView.getChildCount() ; ++i) {
				View cell = rowBView.getChild(i);
				if(cell instanceof BView) {
					int heightDiff = rowBView.getHeight() - cell.getHeight();
					if(heightDiff > 0) {
						cell.setSize(cell.getWidth(), rowBView.getHeight());
						//FIXME get valign for cell and adjust cell's translation if center or bottom
					}
				}
			}
		}
		
		//perform alignment
		if(align == HTMLRow.ROW_ALIGN_CENTER) {
//			debugOut("CENTERED");
			//TODO should center relative to page area, not relative to itself which might be wider.
			rowBView.setTranslation((rowBView.getWidth()-innerX)/2, 0);
		} else if(align == HTMLRow.ROW_ALIGN_RIGHT) {
//			debugOut("RIGHT-ALIGNED");
			//TODO should right-align relative to page area, not relative to itself which might be wider.
			rowBView.setTranslation((rowBView.getWidth()-innerX), 0);
		}
		// else LEFT - already that way.
		
//debugOut("rendered row "+rowIndex+"/"+rows.size());
		//TODO unwrap, pulling from next row(s) continuation chunks, allowing for redisplay with a wider window
		if(nextRow != null) {
			// remove chunks from current row that were placed in new row so we don't keep re-wrapping.
			// (new partial chunk cut from end of last row wont' be removed,
			// cutting of that partial chunk altered text of last chunk of
			// previous row)
			row.getHTMLChunks().removeAll(nextRow.getHTMLChunks());
			// we wrapped, we'll render this one next.
			rows.add(rowIndex+1, nextRow);
			
//debugOut(" +1 new wrapped row");
		}
		}
//debugOut();
//debugOut("row: "+rowBView.getWidth()+"x"+rowBView.getHeight());
		progress.isComplete = true;
		progress.rowView = rowBView;
		return progress;
	}

	private void setRowBG(BRowView rowBView, List<HTMLRow> rows, int rowIndex) {
		rowBView.setBG(rows, rowIndex, baseURL);
	}

//	private void setRowBG(BRowView rowBView, HTMLRow row) {
//		//FIXME support color + image
//		if(row.bg != null) {
//			try {
////				debugOut("!!!set "+row.getClass()+" bg to "+new URL(baseURL, row.bg));
//				rowBView.setResource(new URL(baseURL, row.bg));
//				return;
//			} catch (MalformedURLException e) {
//				// TODO Auto-generated catch block
//				e.printStackTrace();
//			}
//		}
//		if(row.bgcolor != null) {
////			debugOut("!!!set "+row.getClass()+" bg to "+row.bgcolor);
//			rowBView.setResource(row.bgcolor);
//		}
////		else debugOut("!!! null bgcolor");
//	}

	private ArrayList<Integer> setTableWidths(HTMLTableRow htmlTableRow, List<HTMLRow> rows,
			int rowIndex, int tableWidth, Map<String, FontInfo> fontInfos) {
		if(htmlTableRow.isWidthProcessed) {
			return null;
		}
		String tableid = htmlTableRow.tableid;
		
		// start at the first row
		int firstRow = findFirstTableRow(tableid, rows, rowIndex);

		int DEFAULT_CELL_WIDTH = -1000;
		ArrayList<Integer> maxCellWidths = calculateTableWidths(tableid, rows,
				tableWidth, fontInfos, firstRow, DEFAULT_CELL_WIDTH);
		
		//FIXME don't set calc'd width on "widthspecified" ones, figure out when to let calc'd override the specified width
		// then set the override for ALL in that column  maybe make two lists and decide afterwards
		// ... or just let calc override specified when we call getWidth on the chunk
		// "If column widths prove to be too narrow for the contents of a particular table cell, user agents may choose to reflow the table."
		
		//FIXME need to be able to re-figure table widths after images loaded... 
		// so have a DEFAULT_CELL_WIDTH_FROM_IMAGE also, and recalc when that's something.
		int distribCellWidth = calculateDistributedCellWidth(tableWidth, DEFAULT_CELL_WIDTH,
				maxCellWidths);

		processTableWidths(rows, firstRow, DEFAULT_CELL_WIDTH, maxCellWidths,
				distribCellWidth, tableid, tableWidth);
		
		return maxCellWidths;
		// FIXME ELSEWHERE clear all the (cell width overrides and) isWidthProcessed when font size changes!
	}

	private int calculateDistributedCellWidth(int tableWidth, int DEFAULT_CELL_WIDTH,
			ArrayList<Integer> maxCellWidths) {
		// assign width of smallest calculated columns until avail width is 2*currentcolwidth,
		//    then divide remaining space equally
		// NAH... let all calc'd cells get their requested width.
		int availWidth = tableWidth;
		int setWidths = 0;
		int unsetCols = 0;
		for(Integer cw : maxCellWidths) {
			if(cw != DEFAULT_CELL_WIDTH && cw > 0) {
				setWidths += cw;
			} else {
				++unsetCols;
			}
		}
		availWidth -= setWidths;
		
		// make sure we have a reasonable minimum for the remaining ones.
		//TODO get a better min-col-width handling
		if(availWidth < unsetCols) {
			availWidth = unsetCols*20;
		}
		int distribCellWidth = 0;
		if(unsetCols > 0) distribCellWidth = availWidth/unsetCols;
		return distribCellWidth;
	}

	private void processTableWidths(List<HTMLRow> rows, int firstRow,
			int DEFAULT_CELL_WIDTH, ArrayList<Integer> maxCellWidths,
			int distribCellWidth, String tableid, int tableWidth) {
		debugOut(tableid+":"+ maxCellWidths+" : "+distribCellWidth);

		// explicitly set calc'd cell widths based on calculations (including cells with colspans) and set row.isWidthProcessed
		for(int row = firstRow ; row < rows.size() ; ++row) {
			if(!(rows.get(row) instanceof HTMLTableRow)) {
				break;
			}
			HTMLTableRow tableRow = (HTMLTableRow) rows.get(row);
			if (!tableIdsMatch(tableid, tableRow.tableid)) {
				break;
			}
			List<HTMLChunk> chunks = tableRow.getHTMLChunks();
			int colIndex = 0;
			for(int cell = 0 ; cell < chunks.size() ; ++cell) {
				HTMLChunk ch = chunks.get(cell);
				// not every chunk is a table cell, unfortunately.
				if(ch instanceof HTMLTableCell) {
					HTMLTableCell chunk = ((HTMLTableCell)ch);
					int w = 0;
					if(chunk.colSpan == maxCellWidths.size()) {
						// handle degenerate case - we definitely should have a
						// full-table width in this case
						w = tableWidth;
					} else {
						for(int i = 0 ; i < chunk.colSpan ; ++i) {
							int colWid = maxCellWidths.get(colIndex);
							if(colWid == DEFAULT_CELL_WIDTH || colWid < 0) {
								colWid = distribCellWidth;
							}
							w += colWid;
						}
					}
					
					chunk.setCalculatedWidth(w);
					
					colIndex += chunk.colSpan;
				}
			}
			
			tableRow.isWidthProcessed = true;
		}
	}

	private ArrayList<Integer> calculateTableWidths(String tableid,
			List<HTMLRow> rows, int tableWidth,
			Map<String, FontInfo> fontInfos, int firstRow,
			int DEFAULT_CELL_WIDTH) {
		ArrayList<Integer> maxCellWidths = new ArrayList<Integer>();
		// for every cell, figure out its max width
		for(int row = firstRow ; row < rows.size() ; ++row) {
			if(!(rows.get(row) instanceof HTMLTableRow)) {
				break;
			}
			HTMLTableRow tableRow = (HTMLTableRow) rows.get(row);
			if (!tableIdsMatch(tableid, tableRow.tableid)) {
				break;
			}
			List<HTMLChunk> chunks = tableRow.getHTMLChunks();
			int cellsInRow = 0;
			maxCellWidths.ensureCapacity(chunks.size()); // doesn't set size
			for(int cell = 0 ; cell < chunks.size() ; ++cell) {
				HTMLChunk ch = chunks.get(cell);
				// not every chunk is a table cell, unfortunately.
				if(ch instanceof HTMLTableCell) {
					HTMLTableCell chunk = (HTMLTableCell) ch;
					int colIndex = cellsInRow;
					cellsInRow += chunk.colSpan;
					while(maxCellWidths.size() < cellsInRow) {
						maxCellWidths.add(-1);
					}
					//skip this cell if colspan or column already maxed out
					if(
							chunk.colSpan > 1 || //FIXME comment out to activate colspan calcs
							maxCellWidths.get(colIndex) == DEFAULT_CELL_WIDTH)
						continue;
					
					if(chunk.isWidthSpecified()) {
		//   if it has explicit size/percent, use that.
						int cellWidth = chunk.getWidth(tableWidth);
//debugOut("width specified: "+cellWidth+" / "+chunk.colSpan);
						// apply width evenly across spanned columns
						cellWidth = cellWidth/chunk.colSpan;
						for(int i = 0 ; i < chunk.colSpan ; ++i) {
							maxCellWidths.set(colIndex+i, Math.max(maxCellWidths.get(colIndex+i), cellWidth));
						}
					} else {
						// maxWidth needs to be small enough to allow other cells some space within tableWidth still
						// maxWidth also needs to take into account fixed widths that are committed to
						// estimate: divide total width by estimated number of cells 
						// (estimated number of cells = chunks less known non-cell chunks so far)
						int maxWidth = tableWidth/(chunks.size()-(cellsInRow-cell-chunk.colSpan));
						int cellWidth = calculateCellWidth(chunk, 200, maxWidth, DEFAULT_CELL_WIDTH, fontInfos);
//debugOut("width calculated: "+cellWidth+" / "+chunk.colSpan+" from max "+maxWidth);
						if(cellWidth == DEFAULT_CELL_WIDTH) {
							// TODO experimental - don't apply default when we're covering multiple columns
							if(chunk.colSpan <= 1)
								maxCellWidths.set(colIndex, DEFAULT_CELL_WIDTH);
						} else {
							// apply width evenly across spanned columns
							cellWidth = cellWidth/chunk.colSpan;
							for(int i = 0 ; i < chunk.colSpan ; ++i) {
								maxCellWidths.set(colIndex+i, Math.max(maxCellWidths.get(colIndex+i), cellWidth));
							}
						}
					}
				}
			}
		}
		return maxCellWidths;
	}

	private boolean tableIdsMatch(String tableid, String rowtableid) {
		return (tableid == null && rowtableid == null)
				|| (tableid != null && tableid.equals(rowtableid));
	}


	private int findFirstTableRow(String tableid, List<HTMLRow> rows,
			int rowIndex) {
		int firstRow = rowIndex;
		while(firstRow > 0) {
			HTMLRow prevRow = rows.get(firstRow-1);
			if(prevRow instanceof HTMLTableRow) {
				if (tableIdsMatch(tableid, ((HTMLTableRow) prevRow).tableid)) {
					firstRow = firstRow-1;
				} else {
					break;
				}
			} else {
				break;
			}
		}
		return firstRow;
	}

	private int calculateCellWidth(HTMLTableCell cell, int maxTextChars, int maxWidth,
			int returnValueForMax, Map<String, FontInfo> fontInfos) {
		int width = -1;
		int indeterminateImageWidth = 0;
		for(int rowIndex = 0 ; rowIndex < cell.rows.size() ; ++rowIndex) {
			HTMLRow row = cell.rows.get(rowIndex);
			if(row instanceof HTMLTableRow) {
				// table in a table cell... typically same as long text or wide image
				//TODO could handle this in small cases... 
				// will be calculating it on render anyhow, but this could be out of view... defer it until really needed.
				ArrayList<Integer> widths = setTableWidths((HTMLTableRow) row, cell.rows, rowIndex, maxWidth, fontInfos);
				if(widths == null) {
					return returnValueForMax;
				} else {
					int tableWidth = 0;
					int otherCols = 0;
					for(Integer colWidth : widths) {
						if(colWidth != null && colWidth != returnValueForMax) {
							tableWidth += colWidth;
						} else {
							++otherCols;
						}
					}
					if(otherCols > 0) {
						if(tableWidth > maxWidth-otherCols) {
							// we're oversize, just give each remaining col 5 pixels
							tableWidth += otherCols*5;
						} else {
							//this is silly.. just returning maxWidth if there are any remaining cols... is that OK?
//							tableWidth += (maxWidth-tableWidth)/otherCols * otherCols;
							tableWidth = maxWidth;
						}
					}
					return tableWidth;
				}
			}
			int rowWidth = 0;
			int cumulativeTextLength = 0;
		//   add up widths of text chunks (use default if longer than maxTextChars characters)
		//    and image chunks (give up if size unknown)
			for(HTMLChunk chunk : row.getHTMLChunks()) {
				switch(chunk.getType()) {
				case FORM:
					rowWidth += getFormFieldWidth((HTMLFormField)chunk, fontInfos);
					break;
				case TEXT:
					String text = ((HTMLTextChunk)chunk).text;
					if(text != null) {
						cumulativeTextLength += text.length();
						if(cumulativeTextLength > maxTextChars) {
							return returnValueForMax;
						}
						String fontString = bapp.getFontString(((HTMLTextChunk)chunk), getFontScale());
						FontInfo fontInfo = getFontInfo(fontInfos, fontString);
						int textWidth = measureTextWidth(fontInfo, text);
//int otherTextWidth = -1;
//FontInfo fontInfo2 = bapp.localFontInfos.get(fontString);
//if(fontInfo2 != null)
//	otherTextWidth = measureTextWidth(fontInfo2,text);
//debugOut("!!!!!!!!!!! textWidth "+fontString+" TiVo: "+textWidth+" localFont: "+ otherTextWidth);
						rowWidth += textWidth;
					}
					break;
				case IMAGE:
					int iwidth = ((HTMLImageChunk)chunk).width;
					if(iwidth <= 0) {
						++indeterminateImageWidth;
					} else {
						rowWidth += iwidth;
					}
					break;
				}
			}
			if(indeterminateImageWidth > 0) {
				//TODO this didn't work out... not sure why. Even failed after images preloaded??
//				if(rowWidth == 0) {
					// if all we have are unknown images, we can't assume anything.
					return returnValueForMax;
//				} else {
//					// If there is text width, too, images are probably not that big.
//					// Account for some at least some width per image
//					rowWidth += indeterminateImageWidth*20;
//				}
//				indeterminateImageWidth = 0;
			}
			width = Math.max(width, rowWidth);
			if(width > maxWidth) {
				return returnValueForMax;
			}
		}
		
		//   if width greater than maxWidth, use default
		if(width > maxWidth) {
			return returnValueForMax;
		} else {
			return width;
		}
	}

	private RenderResult renderTableCellChunk(Map<String, FontInfo> fontInfos,
			HTMLTableCell chunk, int chunkIndex, int size, int innerX, BView rowParent, URL urlBase, RenderProgress tableCellProgress) {
		int cellWidth = chunk.getWidth(rowParent.getWidth());
		if(cellWidth < 0) {
			int width = (rowParent.getWidth()-innerX)/(size-chunkIndex);
			cellWidth = width*chunk.colSpan;
//TODO this should never be happening now that we pre-calculate widths, right?
	debugOut("!!!! default cell width "+cellWidth);
		}
//		else debugOut(" cell width "+cellWidth);
		int align = chunk.align;
		
		BView cellBase;
		int rowStart = 0;
		RenderProgress firstProgress = null;
		int height = 0;
		if(tableCellProgress != null) {
			cellBase = tableCellProgress.parent;
			rowStart = tableCellProgress.rowIndex;
			firstProgress = tableCellProgress;
			height = tableCellProgress.tableCellHeight;
		} else {
			cellBase = BRowViewParent.newInstance(rowParent, innerX, 0, cellWidth, rowParent.getHeight());
			((BRowViewParent)cellBase).setRowParentBG(baseURL, chunk);
		}
		for(int rowIndex = rowStart ; rowIndex < chunk.rows.size() ; ++rowIndex) {
			//TODO pass true if cell had "nowrap" set.
			RenderProgress rowResult;
			if(firstProgress != null) {
				rowResult = renderRow(firstProgress);
				firstProgress = null;
			} else {
				rowResult = renderRow(urlBase, fontInfos, cellBase,
						chunk.rows, rowIndex, cellBase.getWidth(), false, align);
			}
//			while
			if
			(!rowResult.isComplete) {
				RenderResult result = new RenderResult(cellBase, null, true);
				rowResult.tableCellHeight = height;//FIXME this was supposed to fix it :[
				//FIXME add more context to rowResult?
				result.tableCellProgress = rowResult;
				return result;
//debugOut("quick completion for cell");
//				rowResult = renderRow(rowResult);
			}
			if(rowResult.rowView != null) {
				// place the row, vertically, in the cell
				rowResult.rowView.setLocation(rowResult.rowView.getX(), height);
//				// and its background
//				if(rowResult.rowView.isBGTopOwner())
//					rowResult.rowView.bgView.setLocation(rowResult.rowView.bgView.getX(), height);
				height += rowResult.rowView.getHeight();
				//TODO add equivalent of the following for situations with giant table rows
//				if(nextRowY > pageArea.getHeight()) {
//					break;
//				}
			} 
			else debugOut("No view for cell!");
		}
		cellBase.setSize(cellBase.getWidth(), height);
		return new RenderResult(cellBase, null, false);
	}

	private int adjustChunkInrow(BRowView rowBView, int innerX,
			BView view, int width) {
		// recorded a width, but no view needed.
		if(view == null) return (width>0?innerX+width:innerX);
//debugOut(chunk.toString());
		// increase row height as needed
		if(view.getHeight() > rowBView.getHeight()) {
			rowBView.setSize(rowBView.getWidth(), view.getHeight());
		}
//					debugOut("position chunk");
		innerX += positionChunkBView(view, rowBView, innerX);
		
		// adjust row width to include the entire last chunk
		if (view != null) {
			int finalWidth = view.getX()
					+ view.getWidth();
			if (rowBView != null
					&& finalWidth > rowBView.getWidth()) {
				rowBView.setSize(finalWidth,
						rowBView.getHeight());
			}
		}
		return innerX;
	}
	
	private RenderResult renderFormChunk( Map<String, FontInfo> fontInfos, HTMLFormField chunk,
			BRowView rowBView, int innerX, int availWidth, URL urlBase) {
		BView fieldView = null;
		String type = chunk.type;
		if(type == null) type = "";
		else type = type.toLowerCase();
		if(type.equals("hidden")) {
			// just leave it be.
		} else if((type.equals("select-one") || type.equals("select-multiple")) && chunk instanceof HTMLFormSelect) {
			HTMLFormSelect htmlSelect = (HTMLFormSelect)chunk;
			Color labelColor = chunk.color;
			String fontString = bapp.getFontString(htmlSelect, getFontScale());
			
			Dimension dim = getSelectDimensions(htmlSelect, fontInfos);
			int textWidth = dim.width;
			int textHeight = dim.height;
			if(innerX > 0 && FormSelect.getWidth(textWidth, bapp.getTheme()) > availWidth) {
				return new RenderResult(null, chunk, false);
			}
			
			FormSelect button = new FormSelect(htmlSelect.getFormActionSpec(), htmlSelect.name, rowBView, 0, 0, textWidth, textHeight);
			button.setAction(htmlSelect);
			button.updateSelect(htmlSelect.valuesDisplays, htmlSelect.selectedIndices, labelColor, fontString);
			
			fieldView = button;
		} else if(type.equals("button") || type.equals("submit") || type.equals("reset")) {
			HTMLFormInput htmlButton = (HTMLFormInput)chunk;
			String value = htmlButton.getAttribute("value");
			if(value == null) {
				value = type;
			}
			Color labelColor = chunk.color;
			String fontString = bapp.getFontString(htmlButton, getFontScale());
			
			Dimension dim = getButtonDimensions(htmlButton, fontInfos);
			int textWidth = dim.width;
			int textHeight = dim.height;			
			if(innerX > 0 && FormButton.getWidth(textWidth, bapp.getTheme()) > availWidth) {
				return new RenderResult(null, chunk, false);
			}
			
			FormButton button = new FormButton(htmlButton.getFormActionSpec(), htmlButton.name, rowBView, 0, 0, textWidth, textHeight);
			button.setAction(htmlButton);
			button.updateButton(value, labelColor, fontString);
			
			fieldView = button;
		} else if(type.equals("checkbox") || type.equals("radio")) {
			HTMLFormInput htmlCheck = (HTMLFormInput)chunk;
			boolean checked = false;
			String check = htmlCheck.getAttribute("checked");
			if(check != null) checked = true;

			Color labelColor = chunk.color;
			String fontString = bapp.getFontString(htmlCheck, getFontScale());
			if(innerX > 0 && 20 > availWidth) {
				return new RenderResult(null, chunk, false);
			}

			FormCheck button = new FormCheck(type.equals("checkbox"), htmlCheck.getFormActionSpec(), htmlCheck.name, rowBView, 0, 0, 20, 20);
			button.setAction(htmlCheck);
			button.updateCheck(type, checked, labelColor, fontString);
			
			fieldView = button;
		} else if( type.equals("textarea") && (chunk instanceof HTMLFormTextarea)) {
			HTMLFormTextarea htmlText = (HTMLFormTextarea)chunk;
			String value = htmlText.contents;
			if(value == null) {
				value = "";
			}
			
			Color labelColor = chunk.color;
			String fontString = bapp.getFontString(htmlText, getFontScale());

			Dimension dim = getTextAreaDimensions(htmlText, fontInfos);
			int textWidth = dim.width;
			int textHeight = dim.height;
			if(innerX > 0 && FormInput.getWidth(textWidth, bapp.getTheme()) > availWidth) {
				return new RenderResult(null, chunk, false);
			}
			
			FormInput button = new FormInput(htmlText.getFormActionSpec(), htmlText.name, rowBView, 0, 0, textWidth, textHeight);
			button.setAction(htmlText);
			button.updateTextarea(value, labelColor, fontString);
			
			fieldView = button;
		} else if (type.equals("image")) {
			HTMLFormInput htmlText = (HTMLFormInput)chunk;
			
			Dimension dim = getImageDimensions(htmlText);
			int height = dim.height;
			int width = dim.width;
			if(innerX > 0 && FormImage.getWidth(width, bapp.getTheme()) > availWidth) {
				return new RenderResult(null, chunk, false);
			}

			FormImage button = new FormImage(htmlText.getFormActionSpec(), htmlText.name, rowBView, 0, 0, width, height);
			button.setAction(htmlText);
			button.updateImage(htmlText.getAttribute("value"), urlBase, htmlText.getAttribute("src"));
			
			fieldView = button;
		
		} else { // anything unknown (or text or password) is treated as text
			//if(type.equals("") || type.equals("text") || type.equals("password")) {
			HTMLFormInput htmlText = (HTMLFormInput)chunk;
			String value = htmlText.getAttribute("value");
			if(value == null) {
				value = "";
			}
			
			Color labelColor = chunk.color;
			String fontString = bapp.getFontString(htmlText, getFontScale());
			
			Dimension dim = getInputDimensions(htmlText, fontInfos);
			int textWidth = dim.width;
			int textHeight = dim.height;
			if(innerX > 0 && FormInput.getWidth(textWidth, bapp.getTheme()) > availWidth) {
				return new RenderResult(null, chunk, false);
			}
			
debugOut("render text input "+htmlText.name);			
			FormInput button = new FormInput(htmlText.getFormActionSpec(), htmlText.name, rowBView, 0, 0, textWidth, textHeight);
			button.setAction(htmlText);
			button.updateText(type, value, labelColor, fontString);
			
			fieldView = button;
		
		}
		
		// handled disabled fields
		if(chunk.disabled) {
			fieldView.setFocusable(false);
			fieldView.setTransparency(0.5f);
		}
		
		//TODO wrap it if width is too wide
		return new RenderResult(fieldView, null, false);
	}

	private Dimension getInputDimensions(HTMLFormInput htmlText,
			Map<String, FontInfo> fontInfos) {
		String fontString = bapp.getFontString(htmlText, getFontScale());
		FontInfo fontInfo = fontInfos.get(fontString);
		
		String sizeString = htmlText.getAttribute("size");
		int size = 20;
		if(sizeString != null) {
			try {
				size = Integer.parseInt(sizeString);
			} catch(Exception e) {
			}
		}
		if(fontInfo == null) {
			return new Dimension(15*size,15);
		}
		int textWidth = (int) (fontInfo.getGlyphInfo('X').getAdvance()*size);
		int textHeight = (int) fontInfo.getHeight();
		return new Dimension(textWidth, textHeight);
	}

	private Dimension getImageDimensions(HTMLFormInput htmlText) {
		int width = -1;
		int height = -1;
		try {
			width = Integer.parseInt(htmlText.getAttribute("width"));
		} catch(Exception e) {
			// use default
		}
		try {
			height = Integer.parseInt(htmlText.getAttribute("height"));
		} catch(Exception e) {
			// use default
		}
		if(width < 0) {
			width = getStyleSize(htmlText.getAttribute("style"), "width");
		}
		if(height < 0) {
			height = getStyleSize(htmlText.getAttribute("style"), "height");
		}
		
		if(width < 0) width = 75;
		if(height < 0) height = 50;

		return new Dimension(width, height);
	}
	
	private int getStyleSize(String a, String sizeStyle) {
		String wid = getStyleValue(a, sizeStyle);
		if(wid != null) {
			Integer size = HTMLParserCallback.getCSSLength(wid);
			if(size != null) {
				return size;
			}
		}
		return -1;
	}

	/** extracts a named css item's value from the style attribute's value */
	private String getStyleValue(String astyle, String styleName) {
		return HTMLParserCallback.getStyleValue(astyle, styleName);
	}
	
	private Dimension getSelectDimensions(HTMLFormSelect htmlSelect,
			Map<String, FontInfo> fontInfos) {
		String fontString = bapp.getFontString(htmlSelect, getFontScale());
		FontInfo fontInfo = fontInfos.get(fontString);
		
		String widestText = "";
		for(String display : htmlSelect.valuesDisplays.values()) {
			if(display.length() > widestText.length()) {
				widestText = display;
			}
		}
		int textWidth = measureTextWidth(fontInfo, widestText);
		int textHeight = (int) fontInfo.getHeight();
		return new Dimension(textWidth, textHeight);
	}

	private Dimension getButtonDimensions(HTMLFormInput htmlButton,
			Map<String, FontInfo> fontInfos) {
		String value = htmlButton.getAttribute("value");
		if(value == null) {
			value = htmlButton.type;
		}
		String fontString = bapp.getFontString(htmlButton, getFontScale());
		FontInfo fontInfo = fontInfos.get(fontString);
		int textWidth = measureTextWidth(fontInfo, value);
		int textHeight = (int) fontInfo.getHeight();
		Dimension dim = new Dimension(textWidth, textHeight);
		return dim;
	}

	private Dimension getTextAreaDimensions(HTMLFormTextarea htmlText, Map<String, FontInfo> fontInfos) {
		String fontString = bapp.getFontString(htmlText, getFontScale());
		FontInfo fontInfo = fontInfos.get(fontString);
		
		String sizeString = htmlText.getCols();
		
		int size = 20;
		if(sizeString != null) {
			try {
				size = Integer.parseInt(sizeString);
			} catch(Exception e) {
			}
		}
		int textWidth = (int) (fontInfo.getGlyphInfo('X').getAdvance()*size);
		int textHeight = (int) fontInfo.getHeight();
//		if(type.equals("textarea")) {
			String heightString = htmlText.getRows();
			int height = 20;
			if(heightString != null) {
				try {
					height = Integer.parseInt(heightString);
				} catch(Exception e) {
				}
			}
			textHeight = (int) (fontInfo.getHeight()*height);
//		}
		Dimension dim = new Dimension(textWidth, textHeight);
		return dim;
	}

	private int getFormFieldWidth(HTMLFormField chunk, Map<String, FontInfo> fontInfos) {
		String type = chunk.type;
		if(type == null) type = "";
		if(type.equals("textarea")) {
			return FormInput.getWidth( getTextAreaDimensions((HTMLFormTextarea) chunk, fontInfos).width, bapp.getTheme() );
		} else if(type.equals("checkbox") || type.equals("radio")) {
			return FormCheck.getWidth( 20, bapp.getTheme(), type.equals("checkbox") );
		} else if(type.equals("button") || type.equals("submit") || type.equals("reset")) {
			return FormButton.getWidth( getButtonDimensions((HTMLFormInput) chunk, fontInfos).width, bapp.getTheme() );
		} else if(type.equals("select-one") || type.equals("select-multiple")) {
			return FormSelect.getWidth( getSelectDimensions((HTMLFormSelect) chunk, fontInfos).width, bapp.getTheme() );
		} else if (type.equals("image")) {
			return FormImage.getWidth( getImageDimensions((HTMLFormInput) chunk).width, bapp.getTheme() );
		} else if(type.equals("hidden")) {
			return 0;
		} else {
			return FormInput.getWidth( getInputDimensions((HTMLFormInput) chunk, fontInfos).width, bapp.getTheme() );
		}
	}

	private HTMLRow addChunk(HTMLRow row, HTMLChunk chunk, HTMLRow prevRow) {
		if(row == null) {
			row = new HTMLRow();
			row.align = prevRow.align;
			row.bg = prevRow.bg;
			row.bgcolor = prevRow.bgcolor;
		}
		row.add(chunk);
		return row;
	}
	
	private int positionChunkBView(BView chunkBView, BView parentBView, int rowXPos) {
		chunkBView.setLocation(rowXPos, chunkBView.getY());
		return chunkBView.getWidth();
	}
	
	private RenderResult renderTextChunk(Map<String, FontInfo> fontInfos, HTMLTextChunk chunk, BView parent, int availWidth, int entireWidth) {
		FontInfo fontInfo = getFontInfo(fontInfos, bapp.getFontString(chunk, getFontScale())
//				, chunk.font, chunk.isBold, chunk.isItalic, chunk.fontSize
				);

		//FIXME if we're first chunk in row, figure all remaining rows, and render all but last (up to some limit) as one big view.
		int endIndex = renderableTextEnd(chunk.text, fontInfo, availWidth, entireWidth);
		HTMLStyleChunk returnChunk = null;
		// could have wrapped entire chunk
		if (endIndex <= 0 && chunk.text.length() > 0) {
			returnChunk = chunk;
//debugOut("wrapping all of:"+chunk.text);
			return new RenderResult(null, returnChunk, false);
		} else if (endIndex < chunk.text.length()) {
//debugOut("remainder:"+chunk.text.substring(endIndex));
			returnChunk = getRemainingTextChunk(endIndex, chunk);
		}

		String renderText = getRenderText(chunk.text, endIndex);
		chunk.text = renderText;

		// last second before final measurement - adjust content to display right on TiVo
		String unalteredText = renderText;
		renderText = fixTextForRendering(renderText);
		
		int textWidth = measureTextWidth(fontInfo, renderText);
		// decide on view-or-not based on original text... e.g.
		// non-breaking-space would replaced with space would get trimmed away,
		// but by its nature should get a view to take up space.
		if(unalteredText != null && unalteredText.trim().length() > 0) {
		//TODO note when bgcolor and text/link color are the same and use lighter/darker to fix it.
//debugOut("rendering("+HTMLStyleChunk.getFontString(chunk, )+") "+textWidth+"/"+availWidth+(chunk.link!=null)+":"+renderText);
			BView result;
			int viewHeight = 10;
			if(fontInfo != null) viewHeight = (int) fontInfo.getHeight();
			if (isButtonText(chunk.link)) {
				result = newButtonTextBView(parent, renderText, bapp.getFontString(chunk, getFontScale()),
						chunk.color == null ? fgColor : chunk.color, 
								//FIXME chunk.bgcolor, chunk.bg, 
								textWidth, viewHeight);
			} else {
				result = newTextBView(parent, renderText, bapp.getFontString(chunk, getFontScale()),
						// link color is priority but local color is higher priority
						chunk.link!=null?(chunk.isLocalColor&&chunk.color!=null?chunk.color:linkColor):(chunk.color == null ?fgColor : chunk.color), 
//						//explicit color is the priority
//						chunk.color == null ? (chunk.link!=null?linkColor:fgColor) : chunk.color, 
								 chunk.bgcolor, chunk.bgposition, chunk.bgrepeat, chunk.bg, textWidth, viewHeight);
			}
			if(chunk.link != null) {
				addLinkView(result, chunk.link);
			}
			
			return new RenderResult(result, returnChunk, false);
		} else {
//			debugOut("*************saved a view by skipping whitespace!");
			return new RenderResult(null, returnChunk, false, textWidth);
		}
	}

	/**
	 * Fix any characters that won't display right on the TiVo e.g. the
	 * character used by JSoup for nbsp
	 * 
	 * @param text
	 * @return the improved text - should be an equivalent width as much as
	 *         possible - this is usually called only during render, not during
	 *         other calculations that figure width based on the original text
	 */
	public static String fixTextForRendering(String text) {
		try {
			return UnicodeHelper.fixTextForRendering(text);
		} catch(Throwable t) {
			return text;
		}
	}
	
	private float getFontScale() {
		return this.bapp.getFontScale();
	}

	private BView newButtonTextBView(BView rowView, String renderText,
			String font, Color color, int textWidth, int textHeight) {

		
		BView textBView = new BView(rowView, 0, 0, textWidth+4, textHeight+4);
		//TODO make a better look - bevel with transparent middle, or outer part of reg button border
		// need buttonBG to be a child of the returned view so it gets positioned properly
//		BView buttonBG = new BView(textBView, 0, 0, width, height);
//		buttonBG.setResource(Color.lightGray);
		BView buttonBG2 = new BView(textBView, 2, 2, textWidth+2, textHeight+2);
//		buttonBG2.setResource(new Color(50,50,50));
		buttonBG2.setResource(Color.black);
		buttonBG2.setTransparency(0.85f);
		BView buttonBG1 = new BView(textBView, 0, 0, textWidth+2, textHeight+2);
		buttonBG1.setResource(new Color(200,200,200));
		buttonBG1.setTransparency(0.75f);

		if (font != null) {
			BView textBView2 = new BView(textBView, 2, 2, textWidth, textHeight);
			textBView2.setResource(
					textBView2.createText(font, color, renderText),
					IHmeProtocol.RSRC_HALIGN_LEFT);
		}
		return textBView;
	}

	private boolean isButtonText(String link) {
		return link != null && link.startsWith("BUTTON:");
	}

	private void addLinkView(BView result, String link) {
		BLinkView linkView = new BLinkView(link, result, 0, 0, result.getWidth(),
				result.getHeight());
//			linkView.setResource(Color.lightGray);
	}
	
//	private ArrayList<String> imageQueue = new ArrayList<String>(); 
	
	private RenderResult renderImageChunk(HTMLImageChunk chunk, BView parent, int innerX, int availWidth, URL urlBase) {
		if(chunk.width > availWidth && innerX > 0) {
			// move image to a return chunk if there isn't enough good room.
			return new RenderResult(null, chunk, false);
		}
		Object replacementResource=null;
		if(chunk.width < 0 || chunk.height < 0) {
			Object detail = getResourceDetail(chunk.src);
			//TODO allow for weird src entries?
			if(detail == null) {
//				queueImageDetails(chunk);
				// fall through.. newImageBView will return null in this case
				//TODO should be able to queue rest of row
			} else if(detail instanceof Boolean) {
				// This is the flag that an attempt is/was in progress... don't get caught in a loop.
				// but we have to set size info to prevent other problems.
				// The getResourceDetail will be replaced when it gets details
				// or an error for the NEXT time this gets rendered
				
				// only change size if -1
				if(chunk.width < 0)
					chunk.width = DEFAULT_IMAGE_WIDTH;
				if(chunk.height < 0)
					chunk.height = DEFAULT_IMAGE_HEIGHT;
			} else if(detail instanceof Dimension){
				Dimension d = (Dimension)detail;
//debugOut("CACHED DETAIL "+d+":"+chunk.src);
				if (chunk.width >= 0) {
					chunk.height = d.height * (chunk.width / d.width);
				} else if (chunk.height >= 0) {
					chunk.width = d.width * (chunk.height / d.height);
				} else {
					chunk.width = d.width;
					chunk.height = d.height;
				}
			} else {
				// only change size if -1
				if(chunk.width < 0)
					chunk.width = DEFAULT_IMAGE_WIDTH;
				if(chunk.height < 0)
					chunk.height = DEFAULT_IMAGE_HEIGHT;
				// use alt text if it's available.
				if(chunk.alt != null && chunk.alt.trim().length() > 0) {
					// but don't forget the error info
					if (detail instanceof String) {
						replacementResource = chunk.alt + " ["
								+ detail + "]";
					} else {
						replacementResource = chunk.alt;
					}
				} else {
					replacementResource = detail;
				}
			}
		}
		
//		if (chunk.width <= 0
//				&& chunk.height <= 0
//				&& (chunk.src.indexOf("cleardot.") > 0 || chunk.src
//						.indexOf("blank.") > 0)) {
//			chunk.width = 2;
//			chunk.height = 2;
//		}
//debugOut("image("+chunk.width+"x"+chunk.height+") "+chunk.src);
		BView result = newImageBView(parent, chunk.src, chunk.width, chunk.height, chunk, replacementResource, urlBase);
		if(result != null) {
			if(chunk.link != null) {
				addLinkView(result, chunk.link);
			}
		}
		HTMLChunk returnChunk = null;
		return new RenderResult(result, returnChunk, result==null);
	}

//	private void queueImageDetails(HTMLImageChunk chunk) {
//		imageQueue.add(chunk.src);
//		//FIXME do something... might have to return a view and fix it afterwards
////		new Resource(chunk.src)
//		
//		//FIXME to prevent race condition, don't create resources until queue is filled.
//	}
	
	private IHmeEventHandler createImageStreamHandler(final BView view, final HTMLImageChunk chunk, final URL urlBase) {
		return new IHmeEventHandler() {
			
			private Object IMAGE_ERROR_RESOURCE = //"http://127.0.0.1/images/tivodance.gif";//Color.red;
					view.getBApp().getContext().getBaseURI().toString()+"image-error.png";
			private boolean redirectedOnce = false;
			public void postEvent(HmeEvent event) {
				if(event.getOpCode() == HmeEvent.EVT_APP_INFO) {
	              HmeEvent.ApplicationInfo info = (HmeEvent.ApplicationInfo)event;
	              if (info.getMap().get("error.code") != null) {
//	                  int errorCode = Integer.parseInt((String)info.getMap().get("error.code"));
						String errorText = (String) info.getMap().get(
								"error.text");

						// Http Error 301/302 are redirects. we can try again!
//						handleRedirectError(view, event, errorText, urlBase);
						//	                  return handleApplicationError(errorCode, errorText);
  debugOut("app error on resource - "+errorText);
	              }
				} else if (event.getOpCode() == HmeEvent.EVT_RSRC_INFO) {
					Map map = ((ResourceInfo) event).getMap();
debugOut("RSRC "+ResourceInfo.statusToString(((ResourceInfo) event).getStatus())+":"+map);					
					//If this is a resource-done or resource-error, update the view and chunk appropriately
					// and remove from queue
					// Actually, width/height can appear in different states,
					// and so can errors, we just care if we get them, not what
					// status it is.
//					if (((ResourceInfo) event).getStatus() == ResourceInfo.RSRC_STATUS_COMPLETE
//							|| ((ResourceInfo) event).getStatus() == ResourceInfo.RSRC_STATUS_READY) {
						int width = -1;
						int height = -1;
						try {
							width = Integer.parseInt(String.valueOf(map.get("width")));
							height = Integer.parseInt(String.valueOf(map.get("height")));
						} catch (NumberFormatException ex) {
						}
						if (width >= 0 && height >= 0) {
							//FIXME reduce size of too-large width
							
							//FIXME if height or width is already specified, keep it and stretch the image
							
//							imageQueue.remove(chunk.src);
							chunk.width = width;
							chunk.height = height;
							// update the chunk with the src that works.
							// TODO should update ALL uses of src with new one maybe use resource map to hold a redirect url.
							Resource r = ((ResourceInfo) event).getResource();
							if(r != null && r instanceof StreamResource && ((StreamResource)r).getURI() != null) {
								chunk.src = ((StreamResource)r).getURI();
							}
							putResourceDetail(chunk.src, new Dimension(width, height));
//							setSize(view, width, height);
							finishedWithResource(event);
						} else if (map.get("error.code") != null) {
//					} else if(((ResourceInfo) event).getStatus() >= ResourceInfo.RSRC_STATUS_ERROR) {
//							imageQueue.remove(chunk.src);
							// replace image with error
							String text = (String) map.get("error.text");
							// Http Error 301/302 are redirects. we can try again!
							//FIXME not actually implemented yet
							if(!handleRedirectError(view, event, text, urlBase)) {
								if(!tryHTTPInstead(view, event, text, urlBase, chunk)) {
								Resource errorResource;
								if(text == null) {
									errorResource = view.getResource(IMAGE_ERROR_RESOURCE);
									putResourceDetail(chunk.src, errorResource);
								} else {
									if(view != null && view.getResource() != null && view.getResource() instanceof StreamResource)
										debugOut("ERROR SRC final: "+((StreamResource)view.getResource()).getURI());
									else
										debugOut("ERROR SRC orig:"+chunk.src);
									errorResource = createErrorTextResource(view, text);
									putResourceDetail(chunk.src, text);
								}
//								view.setResource(errorResource,
//										ResourceInfo.RSRC_TEXT_WRAP
//										| ResourceInfo.RSRC_HALIGN_LEFT
//										| ResourceInfo.RSRC_VALIGN_TOP
//										| ResourceInfo.RSRC_IMAGE_BESTFIT);
								finishedWithResource(event);
								}
							}
						} else if(((ResourceInfo) event).getStatus() >= ResourceInfo.RSRC_STATUS_COMPLETE) {
							// looks like we're not getting what we want... finish the page.
//							imageQueue.remove(chunk.src);
							width = DEFAULT_IMAGE_WIDTH;
							height = DEFAULT_IMAGE_HEIGHT;
							chunk.width = width;
							chunk.height = height;
//							setSize(view, width, height);
							finishedWithResource(event);
						}
						
						//FIXME oversize image gives us crappy status information...
						// closed state with no map
//						LOG: event 2575.RESOURCE_INFO(closed, com.tivo.core.ds.MapThreadWrapper@0{})
//						RSRC closed:com.tivo.core.ds.MapThreadWrapper@0{}
						// Application error about bad width, no resource id reference
//						LOG: event 1.APP_INFO(com.tivo.core.ds.MapThreadWrapper@311525890{error.code=2, error.text=max image width exceeded (1600 > 1024)})
//						APP_INFO error 2: max image width exceeded (1600 > 1024)
//						LOG: handleApplicationError(2,max image width exceeded (1600 > 1024))
						// Application error about bad height, no resource id reference
//						LOG: event 1.APP_INFO(com.tivo.core.ds.MapThreadWrapper@1128792516{error.code=2, error.text=max image height exceeded (1200 > 768)})
//						APP_INFO error 2: max image height exceeded (1200 > 768)
//						LOG: handleApplicationError(2,max image height exceeded (1200 > 768))
						// and that's IT... no more status!
//					}
				}
			}


//			private void setSize(BView view, int width, int height) {
//				view.setSize(width, height);
//				for (int i = 0; i < view.getChildCount(); ++i) {
//					if (view.getChild(i) instanceof BLinkView) {
//						BLinkView link = (BLinkView) view.getChild(i);
//						link.setSize(width, height);
//					}
//				}
//			}


			private boolean tryHTTPInstead(BView view, HmeEvent event,
					String text, URL referer, HTMLImageChunk chunk) {
				
//				
//				if(true)return false;
//				
				
				if(chunk.src.startsWith("https://")) {
					String url = WebBrowserApp.getFixedResourceSrc(chunk.src);
//					String url = "http://"+chunk.src.substring("https://".length());
					debugOut("TRYING HTTP: "+url);
					try {
						if (url != null) {
							URL resURL = new URL(url.trim());
							URL newresURL = ((WebBrowserApp)view.getBApp()).resolveRedirects(resURL, referer);
//								Resource newres = view.getResource(newresURL);
								Resource newres = getImageResource(bapp, newresURL.toString(), urlBase);
								if (newres != null) {
									newres.addHandler(this);
//									view.setResource(newres);

									Resource res = view.getBApp().getResource(event.getID());
									if (res != null) {
										res.removeHandler(this);
									}
									return true;
								}
						}
					} catch (MalformedURLException e) {
						// give up and continue with regular error display
					}
				}
				return false;
			}


			private boolean handleRedirectError(BView parentView, HmeEvent rsrcInfoEvent,
					String errorText, URL urlBase) {
				// if this is a 301/302 redirect error, handle the redirect and
				// set up the new stream in the parentView - flag that we did
				// this so we don't get caught in a loop
				if(this.redirectedOnce) return false;
				if(handleRedirectError2(parentView, rsrcInfoEvent, errorText, urlBase)) {
					debugOut("Redirect Error handled");
					this.redirectedOnce = true;
					return true;
				}
				return false;
			}

			public boolean handleRedirectError2(final BView view, HmeEvent event,
					String errorText, URL referer) {
				if (errorText != null) {
					Matcher m = ((WebBrowserApp)view.getBApp()).redirectErrorPattern.matcher(errorText);
					if (m.matches()) {
						debugOut("Matched redirect error... " + m.group(1)); //$NON-NLS-1$
						try {
							// unfortunately, this url isn't from the error, it's the
							// original one
							String url = m.group(1);
							if (url != null) {
								URL resURL = new URL(url.trim());
								URL newresURL = ((WebBrowserApp)view.getBApp()).resolveRedirects(resURL, referer);
								if (!resURL.equals(newresURL)) {
									Resource newres = view.getResource(newresURL);
									if (newres != null) {
										newres.addHandler(this);
//										view.setResource(newres);

										Resource res = view.getBApp().getResource(event.getID());
										if (res != null) {
											res.removeHandler(this);
										}
										return true;
									}
								}
							}
						} catch (MalformedURLException e) {
							// give up and continue with regular error display
						}
					}
				}
				return false;
			}



			private void finishedWithResource(HmeEvent event) {
				debugOut("finished with resource, removing handler of "+event.getID());
				// we're done with status, remove ourselves.
				Resource res = bapp.getResource(event.getID());
				if(res != null) {
					res.removeHandler(this);
				}
//				if(imageQueue.size() == 0) {
					continueRender();
//				}
			}
		};
	}


	private Resource createErrorTextResource(final BView view, String text) {
debugOut("ERROR TEXT:"+text);
		return view.createText("default-10.font",
				Color.red, text);
	}

	protected void continueRender() {
		if (bapp.getCurrentScreen() != null
				&& bapp.getCurrentScreen() instanceof WebScreen) {
			((WebScreen) bapp.getCurrentScreen()).continueRender();
		}
	}

	private RenderResult renderHRChunk(HTMLHRChunk chunk, BView parent, int innerX, int availWidth) {
debugOut("hr("+chunk.percent+"%"+chunk.width+"x"+chunk.height+") ");
		BView result = newHRBView(parent, chunk.percent, chunk.width, chunk.height);
		HTMLChunk returnChunk = null;
		return new RenderResult(result, returnChunk, false);
	}

	private BView newTextBView(BView parent, String renderText, Object font,
			Color color, Color bgcolor, int bgposition, int bgrepeat, String bg, int textWidth, int textHeight) {
		BView textBView = new BView(parent, 0, 0, textWidth,
				textHeight);
		BView result = textBView;
		if(bgcolor != null) {
			textBView.setResource(bgcolor);
			if(bg != null) {
				try {
					BView bgView = new BView(textBView, 0, 0, textWidth, textHeight);
					setupInlineBackground(bgView, baseURL, bg, bgposition,
							bgrepeat);
				} catch (Exception e) {
					debugOut("!!! bg url over color for "+renderText+" : "+e);
				}
			}
			textBView = new BView(textBView, 0, 0, textWidth, textHeight);
		} else
		if(bg != null) {
			try {
				setupInlineBackground(textBView, baseURL, bg, bgposition, bgrepeat);
//				URL bgu = new URL(baseURL, bg);
//				//FIXME this is for images... should handle the same as we do for IMG tag
//				textBView.setResource(bgu);
				textBView = new BView(textBView, 0, 0, textWidth, textHeight);
			} catch (Exception e) {
				debugOut("!!! bg url for "+renderText+" : "+e);
			}
		}
		if (font != null) {
			textBView.setResource(
					textBView.createText(font, color, renderText),
					IHmeProtocol.RSRC_HALIGN_LEFT);
		}
		return result;
	}

	public static void setupInlineBackground(BView fgBView, URL baseURL, String bg, int bgposition, int bgrepeat)
	{
		URL bgu = WebBrowserApp.getURL(baseURL, bg);
		int flags = BRowBGView.getResourceFlags(bg, bgrepeat, bgposition);
		//FIXME this is for images... should handle the same as we do for IMG tag
		String bgu_s = bgu.toString();
		// TiVo can't handle https resources.
		bgu_s = WebBrowserApp.getFixedResourceSrc(bgu_s);
//		if(bgu_s.startsWith("https://"))bgu_s="http://"+bgu_s.substring("https://".length());
		Resource bgresource = fgBView.getResource(bgu_s);
		Object bgdet = ((WebBrowserApp)fgBView.getBApp()).getResourceCache(bgu.toString());
		//Boolean (in progress), Dimension, String (error), or Resource (error)
		if(bgdet == null || bgdet instanceof Boolean) {
			bgresource.addHandler(new InlineBackgroundResourceHandler(fgBView, bgu, bgresource, bgrepeat, bgposition));
		} else {
			if(bgdet instanceof Dimension) {
				Dimension d = (Dimension)bgdet;
				InlineBackgroundResourceHandler.adjustInlineBackgroundResource(fgBView, bgresource, bgrepeat, bgposition, d.width, d.height);
			} else {
				// no good...give up.
				return;
			}
		}
		fgBView.setResource(bgresource, flags);
	}

	private BView newHRBView(BView parent, float percent, int width, int height) {
		if(percent > 0.0f) {
			width = (int) (parent.getWidth() * (percent/100.0f));
		} else if(width <= 0) {
			width = parent.getWidth();
		}
		if(height <=0) {
			height = 5;
		}
		BView hrBView = new BView(parent, 0, 0, width, height);
		hrBView.setResource(Color.lightGray);
		return hrBView;
	}
	
	private BView newImageBView(BView parent, String imgSrc, int width, int height, HTMLImageChunk chunk, Object rsrcReplacement, URL urlBase) {
		boolean needSize = false;
		if (width < 0 || height < 0) {
			needSize = true;
			if (width < 0)
				width = parent.getWidth();
			if (height < 0)
				height = 100;
		}
		Resource rsrc;
		if(rsrcReplacement == null && imgSrc != null) {
			if(waitingOnImageResource != null) {
	debugOut("newImageBView USED SAVED RESOURCE");
//				rsrc = waitingOnImageResource;
				//rsrc = parent.getBApp().getResource(imgSrc);
				rsrc = getImageResource(parent.getBApp(), imgSrc, urlBase);
				waitingOnImageResource = null;
			} else {
				rsrc = getImageResource(parent.getBApp(), imgSrc, urlBase);
				// if width/height empty, figure out size
				if (needSize) {
					if(waitingOnImageResource != null) {
	debugOut("!!!!!PROBLEM! waiting on two resources at once!!!");
					}
	debugOut("newImageBView SAVE RESOURCE "+rsrc.getID()+":"+ imgSrc);
					waitingOnImageResource = rsrc;
					//Flag that we're trying... so if it fails, don't get stuck in a loop.
					putResourceDetail(chunk.src, Boolean.TRUE);
					rsrc.addHandler(createImageStreamHandler(parent, chunk, urlBase));
					return null;
				}
			}
		} else {
			// null here not an issue - only used in above case.
			rsrc = null;
		}
		//FIXME reduce size of too-large width		
		BView imageBView = new BView(parent, 0, 0, width, height);
//		textBView.setResource(Color.lightGray);// FIXME placeholder for bad images
		if(rsrcReplacement != null) {
			if(!(rsrcReplacement instanceof Resource)) {
				if (rsrcReplacement instanceof String) {
					rsrcReplacement = createErrorTextResource(imageBView,
							(String) rsrcReplacement);
				}
			}
			imageBView.setResource(rsrcReplacement, ResourceInfo.RSRC_TEXT_WRAP
					| ResourceInfo.RSRC_HALIGN_LEFT
					| ResourceInfo.RSRC_VALIGN_TOP
					| ResourceInfo.RSRC_IMAGE_BESTFIT);
		} else if(imgSrc != null) {
//			Resource rsrc = getImageResource(parent.getBApp(), imgSrc);
//			// if width/height empty, figure out size
//			if (needSize) {
//				rsrc.addHandler(createImageStreamHandler(imageBView, chunk, urlBase));
//			}
//FIXME looks like TiVo can't handle https for these
debugOut("newImageBView ("+width+"x"+height+") SET IMAGE RESOURCE should be "+imgSrc);
			imageBView.setResource(rsrc 
// FIXME no, stretch the image as specified	(have to set scale, and make other adjustments)	
					,IHmeProtocol.RSRC_IMAGE_BESTFIT
					); // FIXME add alignment to chunk and here.
		}
		return imageBView;
	}

	/**
	 * do getResource for something that is definitively an image.
	 * @param pageUrl URL of the page requesting the image
	 */
	private Resource getImageResource(BApplication imageBView, String imgSrc, URL pageUrl) {
		// FIXME verify this is really true:
		// TiVo can't handle https resources... just use http and hope the server will comply
		// lots of error handling code later on resources, but it's not as reliable like this is
		imgSrc = WebBrowserApp.getFixedResourceSrc(imgSrc);
		
		// handle special image urls
		if(imgSrc != null) {
			if(imgSrc.toUpperCase().equals("@VIDEO")) {
//				imgSrc = bapp.getTheme().get
				imgSrc = bapp.getContext().getBaseURI().toString()+"film.png";
			} else if(imgSrc.toUpperCase().equals("@AUDIO")) {
				imgSrc = bapp.getContext().getBaseURI().toString()+"headphone.png";
			} else if(imgSrc.startsWith("@")) {
				if(imgSrc.toLowerCase().contains("youtu")) {
					imgSrc = bapp.getContext().getBaseURI().toString()+"youtube.png";
				} else if(imgSrc.toLowerCase().contains("vimeo")) {
					imgSrc = bapp.getContext().getBaseURI().toString()+"vimeo.png";
				} else {
					imgSrc = bapp.getContext().getBaseURI().toString()+"film.png";
				}
			}
		}
		
		// first check if it's already in the map... if it is, just use the
		// usual technique
		if(imageBView.getApp().getResources().get(imgSrc) != null)
			return imageBView.getResource(imgSrc);
		// if it's not, it will be doing createStream on this "://" string, so
		// we do it ourselves, but include the fact that it's an image.
		
		// peel off non-"file" parts of url.
		String imgFile = imgSrc;
		int end = imgFile.indexOf('?');
		if(end >0) imgFile = imgFile.substring(0, end);
		int start = imgFile.lastIndexOf("/");
		if(start >0) imgFile = imgFile.substring(start);
		String type = HttpURLConnection.guessContentTypeFromName(imgFile);

		// ensure we're sending an "image" type...
		if(type == null || !type.startsWith("image/"))
			type = "image/png";
		// operating on the theory that TiVo will just put all "image/" types
		// through the same pipeline and not make any other assumptions from the
		// type we're sending it.
		
		TeDict dict = null;
		try {
			URL imgUrl = WebBrowserApp.getURL(pageUrl, imgSrc);
			dict = ((WebBrowserApp) imageBView.getApp()).getCookieTeDict(
					imgUrl, pageUrl);
		} catch (Exception e) {
			// just give up on tedict.
		}
		return imageBView.createStream(imgSrc, type, true, dict);
	}

	
	private FontInfo getFontInfo(Map<String, FontInfo> fontInfos, String fontResourceString) {
		FontInfo fontInfo = fontInfos.get(fontResourceString);
		return fontInfo;
	}

	private String getRenderText(String text, int endIndex) {
		return text.substring(0, endIndex);
	}
	
	private HTMLStyleChunk getRemainingTextChunk(int wrapIndex, HTMLTextChunk originalChunk) {
		return new HTMLTextChunk(originalChunk.text.substring(wrapIndex),
				originalChunk.font, originalChunk.isBold,
				originalChunk.isItalic, originalChunk.fontSize,
				originalChunk.color, originalChunk.bgcolor, 
				originalChunk.bgposition, originalChunk.bgrepeat, originalChunk.bg, 
				originalChunk.link, true/* chunkContinued */, originalChunk.isLocalColor);
	}
	
	private int renderableTextEnd(String text, FontInfo fontInfo, int maxWidth, int entireWidth) {
		//enforce sanity
		if(maxWidth < 10) {
			maxWidth = 20;
		}
//// test that proves space's boundingwidth is negative... which gets used in measurement if it's last char
//char t = 'T'; GlyphInfo gi = fontInfo.getGlyphInfo(t); debugOut(t+":"+gi.getAdvance()+":"+gi.getBoundingWidth());
//t=' ';gi = fontInfo.getGlyphInfo(t); debugOut(t+":"+gi.getAdvance()+":"+gi.getBoundingWidth());
//t='\t';gi = fontInfo.getGlyphInfo(t); debugOut(t+":"+gi.getAdvance()+":"+gi.getBoundingWidth());
		
		if(fontInfo == null) return text.length();
		
		int fullTextWidth = measureTextWidth(fontInfo, text);
		if(fullTextWidth <= maxWidth) {
//debugOut("NOWRAP");
			return text.length();
		} else {
//debugOut("wrap");
			// need to wrap
			
			// brute force find the max index
			int maxIndex = text.length() - 1;
			int width = measureTextWidth(fontInfo, text.substring(0, maxIndex));
			while (width > maxWidth) {
				// extremely conservative shortcut for degenerative cases
				if(width > maxWidth*4) {
					maxIndex = maxIndex / 2;
				}
				//TODO allow for possibility that the above shortcut should take us too far.
				maxIndex--;
				width = measureTextWidth(fontInfo, text.substring(0, maxIndex));
			}
			// find break-able character at or before max
			int endIndex = maxIndex;
			//TODO more characters
			while (endIndex > 0 
					&& text.charAt(endIndex - 1) != ' '
					&& text.charAt(endIndex - 1) != '-'
					&& text.charAt(endIndex - 1) != '.'
					&& text.charAt(endIndex - 1) != '/'
					&& endIndex > 1) {
				--endIndex;
			}
			if (endIndex <= 1) {
				// no good wrap
				if(maxWidth < entireWidth) {
//debugOut("wrap all of "+text);
					// must have been near the end of a row, just wrap whole thing.
					endIndex = 0;
				} else {
					// chunk is too big
					// just wrap at max stupidly
					endIndex = maxIndex;
				}
			}
			return endIndex;
		}
	}

	private int measureTextWidth(FontInfo fontInfo, String string) {
		if(string == null || string.length() == 0) return 0;
		if(fontInfo == null) return 10*string.length();
		// The problem: measureTextWidth could have bad widths... 
		//  notably the last char uses its boundingWidth, not it's advance,
		//  and (at least in the simulator) spaces and tabs have large negative boundingWidth values.
//		 GlyphInfo lastWidth = fontInfo.getGlyphInfo(
//				string.charAt(string.length() - 1));
//		if (lastWidth != null && lastWidth.getBoundingWidth() < 0) {
//			return (int) (fontInfo.measureTextWidth(string + ".") - fontInfo
//					.getGlyphInfo('.').getBoundingWidth());
//		}
//		return fontInfo.measureTextWidth(string);
		
		// The other problem: HME v45 measures differently! and wrong!
		// presumably there is more font info sent by the tivo to compensate,
		// but SDK1.4.1e doesn't support it.
		
////		// try some "bonus space", in addition to counteracting the weird ending calculation.
////		// this is because the measures aren't coming out right in the hme 1.4.1e environment... weird.
////		return (int) (fontInfo.measureTextWidth(string + "MMMM") - fontInfo
////				.measureTextWidth("MM"));

		//default seen in code in StreamBaby
        GlyphInfo defaultGlyph = fontInfo.getGlyphInfo((char) 127);

        // do measureTextWidth without the boundingWidth of last char that causes problems.
        float width = 0;
		for (char c : string.toCharArray()) {
			GlyphInfo glyph = fontInfo.getGlyphInfo(c);
			if (glyph == null) {
				if (defaultGlyph == null)
					continue;
				else
					width += defaultGlyph.getAdvance();
			} else {
//				width += glyph.getAdvance();
				width += Math.max( glyph.getAdvance(), glyph.getBoundingWidth());
			}
		}
		return (int) width;
	}


}
