package com.blackledge.david.html.hme;

import java.awt.Color;
import java.awt.Point;
import java.awt.Rectangle;
import java.awt.font.FontRenderContext;
import java.awt.geom.AffineTransform;
import java.awt.geom.Rectangle2D;
import java.io.IOException;
import java.net.MalformedURLException;
import java.net.URL;
import java.net.URLConnection;
import java.text.MessageFormat;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

				// JSOUP
import org.jsoup.Connection.Response;
import org.jsoup.UnsupportedMimeTypeException;
				// /JSOUP

import com.blackledge.david.html.HTMLChunk;
import com.blackledge.david.html.HTMLFormField;
import com.blackledge.david.html.HTMLFormInput;
import com.blackledge.david.html.HTMLFormSelect;
import com.blackledge.david.html.HTMLFormTextarea;
import com.blackledge.david.html.HTMLImageChunk;
import com.blackledge.david.html.HTMLRow;
import com.blackledge.david.html.HTMLStyleChunk;
import com.blackledge.david.html.HTMLTableCell;
import com.blackledge.david.html.HTMLTextChunk;
import com.blackledge.david.html.hme.HTMLBreakdown.RenderProgress;
import com.blackledge.david.html.hme.WebBrowserApp.STATUS;
import com.blackledge.david.tivo.V49BApplication;
import com.blackledge.david.tivo.V49IHmeProtocol;
import com.blackledge.david.tivo.widgets.typo.TyPo;
import com.blackledge.david.tivo.widgets.typo.TyPo.Keyboard;

import com.tivo.core.ds.TeDict;
import com.tivo.hme.bananas.BApplication;
import com.tivo.hme.bananas.BEvent;
import com.tivo.hme.bananas.BFocusMgr;
import com.tivo.hme.bananas.BScreen;
import com.tivo.hme.bananas.BView;
import com.tivo.hme.sdk.HmeEvent;
import com.tivo.hme.sdk.HmeEvent.FontInfo;
import com.tivo.hme.sdk.Resource;
import com.tivo.hme.sdk.Resource.FontResource;
import com.tivo.hme.sdk.StreamResource;
import com.tivo.hme.sdk.View;

public class WebScreen extends BScreen {
	//FIXME crashed on return from an animated gif
	//TODO popups (errors, view/aspect menu (0-9, bright)
	//TODO handle hme: multiple entries as a popup
	//TODO form field display issues
	//TODO proxy icon.png (use jimg4 or whatever it is for .ico)
	//TODO does google use completion lists? support them in general
	
	
	static final int REFRESH_CALLBACK_KEY_CODE = KEY_OPT_GUIDE;
	static final int SCREEN_TRANSITION_CALLBACK_KEY_CODE = KEY_OPT_LIST;
	//TODO bookmarks/quit/size hints
	//TODO figure a way to detect beginning of useful content and jump there?
	//FIXME arrow highlights are a little weird - at least need to be sensitive to left/right exists or not
	private HTMLBreakdown html;
	private List<HTMLRow> pageRows;
	private RenderProgress incompleteRenderProgress;
	private enum RenderAction {NO_ACTION, SCROLL_DOWN, SCROLL_UP};
	private RenderAction incompleteAction = RenderAction.NO_ACTION;
	
	private List<String> fontCacheQueue = new ArrayList<String>();
	
	/** This page's base URL */
	private URL urlBase;
	/** This page's original request URL (urlBase could be set by page content) */
	private URL requestURL;
	/** The "referer" that got us to this page */
	private URL referer;
	/** The displayed title which will be used for a bookmark's title */
	private String titleString;
	/** The displayed icon image which will be used for the bookmark's icon */
	private URL pageIcon;

	private BView titleArea;
	private BView titleIcon;
	private BView titleBookmark;
	private BView titleText;
	private BView pageArea;
	private BView statusArea;
	private BView statusText;

	private BView shade;

	private TyPo typo;
	private HTMLFormField textDestination;
	private FormInput textHolder;

	private Resource scrollAnim;
	private BScrollBar scrollbar;
	/**
	 * if this screen is just doing a raw media display, skips final page entry
	 * steps
	 */
	private boolean isMediaContent = false;
	private BView leftInfo;

	private static void debugOut(String output) {
		WebBrowserApp.debugOut(output);
	}
	
	public static WebScreen newWebScreen(BApplication app) {
		WebScreen webScreen;
		/* BEGIN HME 1.4.1e only */
		if(((WebBrowserApp) app).isResolutionSupported()) {
			webScreen = new WebScreen(app, true);
		} else {
			webScreen = new WebScreen(app);
		}
		/* END HME 1.4.1e only */
		
		// set up custom focus manager
		webScreen.focusMgr = new WebFocusManager(app);
		
		return webScreen;
	}
	private WebScreen(BApplication app, boolean setResolution) {
		/* BEGIN HME 1.4.1e only */
		super(app, app.getReceiverResolutionInfo().getPreferredRenderingResolution());
		/* END HME 1.4.1e only */
	}
	private WebScreen(BApplication app) {
		super(app);
	}
	
//	@Override
//	protected void finalize() throws Throwable {
//		// apparently the only way to deal with getting popped definitively
//		removeAllResourcesAndViews(this);
//		super.finalize();
//	}
//	private void removeAllResourcesAndViews(View parent) {
//		// collect the views before removing them, otherwise the getChild call
//		// gets thrown by the remove calls.
//		ArrayList<View> children = new ArrayList<View>();
//		for(int i = 0 ; i < parent.getChildCount() ; ++i) {
//			 children.add( parent.getChild(i));
//		}
//		
//		for(View child : children) {
//			// remove the not-to-be-saved view and its children
//			removeAllResourcesAndViews(child);
//			if(child.getResource() != null) {
//				child.getResource().remove();
//			}
//			child.remove();
//		}
//	}
	
	public void shade() {
		shade.setTransparency(0.75f);
		shade.setVisible(true);
	}
	
	public void unshade() {
		shade.setVisible(false);
		shade.setTransparency(1.0f);
	}

	public void toggleShade() {
		WebBrowserApp app = (WebBrowserApp) getBApp();
		app.setShaded(!app.isShaded());
		// apply the shading the same as handleEnter does
		if(app.isShaded() && !isMediaContent) {
			shade();
		} else {
			unshade();
		}
	}
	
	@Override
	public boolean handleEnter(Object arg, boolean isReturn) {
		
		WebBrowserApp app = (WebBrowserApp)getBApp();
		if(app.isShaded() && !isMediaContent) {
			shade();
		} else {
			unshade();
		}
		if(isMediaContent && !isReturn) {
			app.hidePleaseWait();
			boolean result = super.handleEnter(arg, isReturn);
			handleEnterFinal(arg, isReturn);
			return result;
		}
		
		// TODO not sure if it's right to restrict this, but I put this call before calling super for some reason and it doesn't make sense to mess with the page's focus on return.
		if(!isReturn) {
		setFocusDefault(pageArea);
		}

		 boolean superResult = super.handleEnter(arg, isReturn);
		 
		 if(isReturn) {
			 handleReturnArg(arg);
			 app.hidePleaseWait();
			 handleEnterFinal(arg, isReturn);
			 return superResult;
		 }
		 
//			PleaseWaitView pw = new PleaseWaitView(getAbove());

		 try {
			 if(html != null) {
	//FIXME ok, stuff that was here is now in prepare, verify it really happened though!
	//FIXME technically, we could move the rest of this there, too, and just have handleEnter wait until font caching is complete.
					
				app.setLastVisited(urlBase, html.title);
				
				// extend the pageArea's bgcolor to the whole screen
				this.getBelow().setResource(html.bodyColor);
				
				setTitle(html.title, urlBase, html.icon);
				setStatus(null);
				
				//FIXME concurrent modification exception on pageRows when doing tables
				cacheFonts(pageRows);
				 
				handleEnterFinal(arg, isReturn);
				return true;
			 } else {
				 handleEnterFinal(arg, isReturn);
				 //FIXME is this the right thing to return here? seems logical, but this is what we get when we launch youtube
				 return false;
			 }
		} finally {
			app.hidePleaseWait();
//			pw.remove();
		}
	}
	
	private void handleEnterFinal(Object arg, boolean isReturn) {
		if(!isReturn) {
			// Have the TiVo let us know when the slide animation should have
			// completed. We won't build page views until it has completed to
			// hopefully prevent crashes. We may be doing a Fade anim (which
			// takes longer) but it shouldn't be as hazardous)
			// NOTE handleEnter fires BEFORE the transition starts, so we wait a moment after the callback.
			this.getApp().sendEvent(new HmeEvent.Key(0, KEY_PRESS, SCREEN_TRANSITION_CALLBACK_KEY_CODE, 0), 
					getResource(BApplication.SLIDE_ANIM));
		}
	}

	@Override
	public boolean handleExit() {
		WebBrowserApp app = (WebBrowserApp)getBApp();
		if(app.isPopping) {
			// clean up everything we created on entering
			removeAllChildren(shade);
			shade.remove();
			removeAllChildren(titleArea);
			titleArea.remove();
			removeAllChildren(statusArea);
			statusArea.remove();
			removeAllChildren(scrollbar);
			scrollbar.remove();
			removeAllChildren(pageArea);
			pageArea.remove();
			// clean up some automatic stuff (especially the focus views)
			//TODO remove actual above/normal/below views themselves?
			//TODO just bite the bullet and removeAllChildren(screen)?
			//TODO make sure this is doing the right thing for resource viewing cases.
			removeAllChildren(getAbove());
			removeAllChildren(getNormal());
			removeAllChildren(getBelow());
			
			return true;
		}
		return super.handleExit();
	}

	private void setUpPageArea() {
		this.scrollAnim = getResource("*250"); //$NON-NLS-1$

		WebBrowserApp wba = (WebBrowserApp)getApp();
		
		Rectangle titleRegion = wba.getWebTitleArea();
		Rectangle renderRegion = wba.getWebRenderArea();
		Rectangle statusRegion = wba.getWebStatusArea();

		BView renderArea = BRowViewParent.newInstance(this.getNormal(), 
				(int)renderRegion.getX(),
				(int)renderRegion.getY(),
				(int)renderRegion.getWidth(),
				(int)renderRegion.getHeight());
		
		//real-world colors are too bright for TV, so we darken the entire page with a transparent overlay
		shade = new BView((BView) renderArea.getParent(),
				renderArea.getParent().getX(), renderArea.getParent().getY(), renderArea.getParent().getWidth(),
				renderArea.getParent().getHeight());
		shade.setResource(Color.black);
		unshade();
		
		if(wba.isMaximized()) {
			// leftInfo not used.
			leftInfo = new BView(this.getAbove(), 0, 0, 1,1, false);
		} else {
			leftInfo = new BView(this.getAbove(),
					wba.getSafeActionHorizontal(), 0,
					(wba.getSafeTitleHorizontal() - wba.getSafeActionHorizontal()),
					getNormal().getHeight());
			// add "back" hint
			leftInfo.setResource(leftInfo.createText("system-15.font", Color.black, "<<"),  //$NON-NLS-1$ //$NON-NLS-2$
					RSRC_HALIGN_RIGHT | RSRC_VALIGN_CENTER);
			//FIXME add font size hint
			//FIXME add mobile hint
			//FIXME add media style hint
		}

		this.titleArea = new BView(this.getNormal(), 
				(int)titleRegion.getX(),
				(int)titleRegion.getY(),
				(int)titleRegion.getWidth(),
				(int)titleRegion.getHeight());
		titleArea.setResource(Color.LIGHT_GRAY);
		int iconWidth = titleArea.getHeight(); // 20x20 for folder
		this.titleText = new BView(titleArea, iconWidth, 0, titleArea.getWidth()-iconWidth*2, titleArea.getHeight());
		this.titleIcon = new BView(titleArea, 0, 0, titleArea.getHeight(), titleArea.getHeight());
		this.titleBookmark = new BView(titleArea, iconWidth+titleText.getWidth(), 0, iconWidth, titleArea.getHeight(),
				false);
		if(!wba.isBookmarksDisabled()) {
			this.titleBookmark.setVisible(true);
			this.titleBookmark.setResource(
					wba.getTheme().getPageNotSaved(), RSRC_IMAGE_BESTFIT | RSRC_HALIGN_CENTER | RSRC_VALIGN_CENTER );
		}

		this.statusArea = new BView(this.getNormal(), 
				(int)statusRegion.getX(),
				(int)statusRegion.getY(),
				(int)statusRegion.getWidth(),
				(int)statusRegion.getHeight());
		statusArea.setResource(Color.LIGHT_GRAY);
		statusArea.setTransparency(0.5f);
		this.statusText = new BView(statusArea, 0, 0, statusArea.getWidth(), statusArea.getHeight());

		pageArea = renderArea;
		
		int scrollbarWidth = wba.getWebScrollbarWidth();
		scrollbar = createScrollBar(pageArea, scrollbarWidth);
	}

	private boolean handleMediaContent(Response soupResp) {
		String contentType = null;
		if(soupResp != null) contentType = soupResp.contentType();
		if(contentType == null) contentType = WebBrowserApp.guessContentType(urlBase);
		URL url = soupResp != null? soupResp.url(): urlBase;
		return handleMediaContent(contentType, url);
	}
	
//	void toggleBGImagesVisible() {
//		((BRowViewParent) pageArea).getBGParent().setVisible(
//				!((BRowViewParent) pageArea).getBGParent().getVisible());
//	}

	/**
	 * Bonks if there was an error, turns off idle timeout if it set up a stream
	 * @param contentType
	 * @param url
	 * @return
	 */
	private boolean handleMediaContent(String contentType, URL url) {
		this.isMediaContent = true;
		
		debugOut("show content "+(contentType!=null?contentType:url)); //$NON-NLS-1$
		// not text, hopefully it's a usable resource
		// finish the display just by showing the resource
		//FIXME looks like TiVo can't handle https for these
		//FIXME (no, not anymore, jsoup throws exc sooner) if needed, could get raw (no exceptions thrown) soupResp.bodyAsBytes
		
//				if(contentType != null && contentType.startsWith("video/")) {
////					getBApp().pop();
//					pageArea.setResource(
//							createText("system-40.font", Color.yellow,
//									"Showing Video.  'RW' from this screen when finished.  'Left' to exit video."),
//									RSRC_TEXT_WRAP);
//					getBApp().push(new VideoScreen(getBApp(), url, contentType), TRANSITION_FORWARD);
//					return true;
//				}
		
//				pageArea.setResource(createStream(url.toString(),
//				contentType, true));
		StreamResource stream = setUpStream(url, this.getBelow(), contentType);
		// if stream == null, then likely no Watcher view was set up, regardless it didn't go well.
		if(stream == null) {
			//setUpStream BONKs if it was really an error (vs. a non-stream solution)
			return false;
		}
		
		pageRows = Collections.synchronizedList(new ArrayList<HTMLRow>());
		if (contentType != null) {
			setStatus(MessageFormat.format(Messages.getString("WebScreen.displaying_contenttype"),contentType)); //$NON-NLS-1$
		} else {
			setStatus(Messages.getString("WebScreen.displaying_unknown_contenttype")); //$NON-NLS-1$
		}
		// don't shade the media
		unshade();
		setTitle(null, url, html.icon);
		// setUpStream set app.setAllowIdleTimeout as needed
		((WebBrowserApp)getBApp()).setLoadingStatus(STATUS.COMPLETE);
		readyToContinueEnter();
		return true;
	}

	private StreamResource setUpStream(URL urlBase, BView resourceParent, String contentType) {
		WebBrowserApp app = (WebBrowserApp)getBApp();
//		app.hideLiveTV();
		
		// strip off extra info
		if (contentType != null) {
			int semicolon = contentType.indexOf(";");
			if (semicolon > 0) {
				contentType = contentType.substring(0, semicolon);
			}
		}
		
		TeDict cookie = getHTTPTeDict(urlBase, this.referer);
		// .asx - xml, "video/x-ms-asf-plugin" (and audio/?)
		// .ram - includes weird realaudio stuff and mp3 files, too "audio/x-pn-realaudio", 
		// .qtl - xml, one entry only "application/x-quicktimeplayer"
		if(contentType != null && contentType.toLowerCase().startsWith("video/")) { //$NON-NLS-1$
			//Now, ALL video goes to video player... if it fails, try to use internal player.
			
			// if videoplayer launched, we don't have our own stream. Return null to cancel internal handling.
			// (but DON'T BONK)
			if(app.playStreamVideo(urlBase, contentType, false)) {
				return null;
			}
			
//			if(contentType.toLowerCase().startsWith("video/mpeg") || contentType.toLowerCase().startsWith("video/x-tivo")) { //$NON-NLS-1$ //$NON-NLS-2$
//				// if videoplayer launched, we don't have our own stream. Return null to cancel internal handling.
//				// (but DON'T BONK)
//				if(app.showVideo(urlBase, contentType)) {
//			        app.setAllowIdleTimeout(false);
//					return null;
//			    }
//			} else {
//				debugOut("INTERNAL video viewer for: "+contentType); //$NON-NLS-1$
//			}
			
//			cookie = new TeDict();
//			cookie.add("recordingId", "tivo:rc.10926");
//			contentType = "video/x-tivo-raw-tts";
			StreamResource stream = createStream(
					getSafeStreamURL(urlBase.toString()),
					
//					"livetv:",
//					"recording:",
//					"recording:download/MAD.TiVo?Container=%2FNowPlaying&id=4515045",
//					"recording:/NowPlaying/1996606/4515045",
//					"recording:/NowPlaying/4515045",
					// json interface: recordingid = tivo:rc.247339
//					 "recording:tivo:rc.10926",
//					 "recording:rc.10926",
//					 "recording:tivo/rc.10926",
//						 "recording:id=tivo:rc.10926",
//					 "recording:id=rc.10926",
//					 "recording:id=tivo/rc.10926",
//					"recording:rc/10926",
//						// http://192.168.0.11/download/Iron%20Man%202.TiVo?Container=%2FNowPlaying&id=10926&Format=video/x-tivo-mpeg
//						// premiere
//						public static final String STREAM_LOOP_PURPLE_SHOWCASES = "recording:rc/10926";
////						http://192.168.0.14/download/Lidia's%20Italy%20in%20America.TiVo?Container=%2FNowPlaying&id=4036331&Format=video/x-tivo-mpeg
//						// tivo hd
////						public static final String STREAM_LIVETV = "recording:rc.4036331";
//					"loopset:Central", // Red (TiVo Central)
//					"loopset:Showcases", // Purple (Showcases, Suggestions)
//					"loopset:Setup", // Blue (Messages & Settings, Music Photos & More, Network Now Playing Lists)
//					"loopset:NowPlaying", // Green (Now Playing, everything not mentioned above)
//					<Url>urn:tivo:image:wishlist-folder</Url>
//					<ContentType>image/*</ContentType>
					contentType, true, cookie);
			VideoWatcher videoWatcher = new VideoWatcher(this.getNormal(), resourceParent, contentType);
			stream.addHandler(videoWatcher);
			setFocusDefault(videoWatcher);
			setFocus(videoWatcher);
			// hide the decorations so the video isn't occluded
			scrollbar.setVisible(false); // just in case.
			leftInfo.setVisible(false);
			// decelerate a 10 second transition
			Resource animation = getResource("*10000,-1"); //$NON-NLS-1$
			titleArea.setLocation(titleArea.getX(), -(titleArea.getHeight()+20), animation);
			statusArea.setLocation(statusArea.getX(), statusArea.getY()+statusArea.getHeight()+20, animation);
			resourceParent.setResource(stream);
			app.setAllowIdleTimeout(false);
			return stream;
		}
		else if(contentType != null && contentType.toLowerCase().startsWith("audio/")) { //$NON-NLS-1$
			// if audioplayer launched, we don't have our own stream. Return null to cancel internal handling.
			// (but DON'T BONK)
			if(app.playStreamAudio(urlBase, contentType)) {
				app.setAllowIdleTimeout(false);
				return null;
			}
			
			// .m3u - 
			debugOut("setUpStream: "+contentType); //$NON-NLS-1$
			List<String> m3uContentTypes = Arrays.asList(new String[] {"audio/x-mpegurl","audio/mpeg-url","application/x-winamp-playlist", //$NON-NLS-1$ //$NON-NLS-2$ //$NON-NLS-3$
					"audio/scm3u","audio/x-scm3u" // "audio/x-scm3u" served by everfreeradio.net and fits the pattern of the pls ones below //$NON-NLS-1$ //$NON-NLS-2$
					});
			List<String> plsContentTypes = Arrays.asList(new String[] {"audio/scpls","audio/x-scpls"}); //$NON-NLS-1$ //$NON-NLS-2$
			if (m3uContentTypes.contains(contentType.toLowerCase())) {
//					// try "text/" here to default as M3U because PLS's start is more
//					// definitively expected than M3U's basic "just a url" format which
//					// is a sensible default anyhow
//					" audio/x-mpegurl audio/mpeg-url application/x-winamp-playlist "//website included these for some reason" audio/scpls audio/x-scpls"
//					.indexOf(" "+contentType.toLowerCase()+" ") >= 0) {
//				#EXTM3U
//				#EXTINF:secs,Title
//				URL
				StreamResource stream = null;
				String file = app.getRawText(urlBase);
//				debugOut("M3U File\n"+file);
				if(file != null) {
					if(file.startsWith("[playlist]")) { //$NON-NLS-1$
						stream = setUpPLSStream(resourceParent, file, cookie);
					} else {
						stream = setUpM3UStream(resourceParent, file, cookie);
					}
				}
				if (stream == null) {
					getBApp().play(V49BApplication.SOUND_BONK);
				} else {
					app.setAllowIdleTimeout(false);
				}
				return stream;
			}
			else if(plsContentTypes.contains(contentType.toLowerCase())) {
//					" audio/scpls audio/x-scpls ".indexOf(" "+contentType.toLowerCase()+" ") >= 0) {
			// .pls - popular, simple, but "text/?" or audio/x-scpls
//			[playlist]
//			File1=URL
//			Title1=Title
//			Length1=secs
//			NumberOfEntries=maxEntryNum
//			Version=2
				StreamResource stream = null;
				String file = app.getRawText(urlBase);
				if (file != null) {
					if (file.startsWith("#EXTM3U")) //$NON-NLS-1$
						stream = setUpM3UStream(resourceParent, file, cookie);
					else {
						stream = setUpPLSStream(resourceParent, file, cookie);
					}
				}
				if (stream == null) {
					getBApp().play(V49BApplication.SOUND_BONK);
				} else {
					app.setAllowIdleTimeout(false);
				}
				return stream;
			}
			StreamResource stream = createStream(getSafeStreamURL(urlBase.toString()),
					contentType, true, cookie);
			//TODO move this view to getBApp().getAbove(); and redirect pause to this stream, tell it to remove itself when complete
			//TODO support playlists
			AudioWatcher audioWatcher = new AudioWatcher(this.getNormal(), resourceParent, contentType);
			stream.addHandler(audioWatcher);
			setFocusDefault(audioWatcher);
			setFocus(audioWatcher);
			resourceParent.setResource(stream);
			app.setAllowIdleTimeout(false);
			return stream;
		} else {
			//FIXME this isn't actually possible currently... text is always just shown as a page... need to get here based on suffix
			StreamResource stream = null;
			if (contentType != null && contentType.startsWith("text/")) { //$NON-NLS-1$
				debugOut(contentType);
					// try "text/" here to default as M3U because PLS's start is more
					// definitively expected than M3U's basic "just a url" format which
					// is a sensible default anyhow
				String file = app.getRawText(urlBase);
				if (file != null) {
					if (file.startsWith("[playlist]")) {
						stream = setUpPLSStream(resourceParent, file, cookie);
					} else {
						stream = setUpM3UStream(resourceParent, file, cookie);
					}
				}
			} else {
				// default stream handler that will do the right thing with e.g. images, too.
				stream = setUpAudioStream(urlBase, resourceParent, contentType, cookie);
			}
			if (stream == null) {
				getBApp().play(V49BApplication.SOUND_BONK);
			} else {
				app.setAllowIdleTimeout(false);
			}
			return stream;
		}
		
	}

	private String getSafeStreamURL(String original) {
		return WebBrowserApp.getFixedResourceSrc(original);
	}

	private StreamResource setUpAudioStream(URL urlBase, BView resourceParent,
			String contentType, TeDict cookie) {
		StreamResource stream = createStream(getSafeStreamURL(urlBase.toString()),
				contentType, true, cookie);
		// need to have a handler still so they can go "back" on other types
		// this does a fine job for images, too.
		VideoWatcher videoWatcher = new VideoWatcher(this.getNormal(), resourceParent, contentType);
		stream.addHandler(videoWatcher);
		setFocusDefault(videoWatcher);
		setFocus(videoWatcher);
		resourceParent.setResource(stream);
		return stream;
	}

	private StreamResource setUpPLSStream(BView resourceParent, String file, TeDict cookie) {
		AudioWatcher audioWatcher = new AudioWatcher(this.getNormal(), resourceParent, null);
		audioWatcher.setPLS(file);
		StreamResource stream = audioWatcher.startNextStream();
		setFocusDefault(audioWatcher);
		setFocus(audioWatcher);
		resourceParent.setResource(stream);
		//FIXME return first stream? not currently used anyhow
		return stream;
	}

	private StreamResource setUpM3UStream(BView resourceParent, String file, TeDict cookie) {
		AudioWatcher audioWatcher = new AudioWatcher(this.getNormal(), resourceParent, null);
		audioWatcher.setM3U(file);
		StreamResource stream = audioWatcher.startNextStream();
		setFocusDefault(audioWatcher);
		setFocus(audioWatcher);
		resourceParent.setResource(stream);
		//FIXME return first stream? not currently used anyhow
		return stream;
	}

	private TeDict getHTTPTeDict(URL url, URL referer) {
		return ((WebBrowserApp)getBApp()).getCookieTeDict(url, referer);
	}

	private Pattern refreshRegex = Pattern.compile("\\s*" + //$NON-NLS-1$
			"(\\d+)" + //1 //$NON-NLS-1$
			"\\s*" + //$NON-NLS-1$
			"(;\\s*[uU][rR][lL]\\s*=\\s*['\"]?" + //2 //$NON-NLS-1$
			"([^'\"]+)" + //3 //$NON-NLS-1$
			"['\"]?\\s*)?"); //$NON-NLS-1$
	private URL refreshCallback;
	/**
	 * 
	 * @param refresh
	 *            string from refresh http-equiv
	 * @return URL if a refresh should occur immediately with this URL, null if
	 *         there is no refresh, an immediate refresh with the same URL as
	 *         current, or it was scheduled to occur later.
	 */
	private URL scheduleRefresh(String refresh, boolean immediateOnly) {
		if(refresh == null || refresh.trim().length() == 0) {
debugOut("empty refresh"); //$NON-NLS-1$
			return null;
		}
		try {
			String ref = refresh.trim();
			// handle badly formatted values
			// content="0; url='http://news.google.com/nwshp?hl=en&amp;tab=wn'" 
			Matcher m = refreshRegex.matcher(ref);
			if(m.matches()) {
debugOut("refresh "+ref+" matched "+m.group(1)+";"+m.group(3)); //$NON-NLS-1$ //$NON-NLS-2$ //$NON-NLS-3$
				Integer seconds = 0;
				String s = m.group(1);
				if (s != null) {
					seconds = Integer.parseInt(s);
				}
				URL refreshUrl = null;
				String u = m.group(3);
				if (u == null || u.length() == 0) {
					refreshUrl = urlBase;
				} else {
					refreshUrl = WebBrowserApp.getURL(urlBase, u);
				}
				if (seconds > 0) {
					if (!immediateOnly) {
						scheduleRefresh(seconds, refreshUrl);
					}
					return null;
				} else {
					//FIXME well, should we skip this or not?
//					if (refreshUrl.equals(urlBase)) {
//						return null;
//					}
					return refreshUrl;
				}
			}
		} catch (Exception e) {
			//FIXME temporary stack trace
			e.printStackTrace();
			// just pretend it wasn't there.
		}
		return null;
	}

	/**
	 * start a thread and push a new screen if this is still the current screen.
	 * @param seconds
	 * @param refreshUrl
	 */
	private void scheduleRefresh(final Integer seconds, final URL refreshUrl) {
debugOut("will refresh "+seconds+"s "+refreshUrl); //$NON-NLS-1$ //$NON-NLS-2$

		this.refreshCallback = refreshUrl;
		this.getApp().sendEvent(new HmeEvent.Key(0, KEY_PRESS, REFRESH_CALLBACK_KEY_CODE, 0), createAnimation(seconds*1000));
	}

	private BScrollBar createScrollBar(BView targetArea, int width) {
		BView parent = (BView) targetArea.getParent();
		BScrollBar result = new BScrollBar(targetArea, parent, width);
		result.setTransparency(0.5f);
		return result;
	}

	private void updateScroller() {
		int size = pageRows.size();
		int minIndex = -1;
		int maxIndex = -1;
		if (size > 0) {
			Rectangle saveArea = pageArea.getBounds();
			// move rectangle into same coordinate space as the children
			saveArea.translate(-pageArea.getX(), -pageArea.getY());
			for (int i = 0; i < pageArea.getChildCount(); ++i) {
				View child = pageArea.getChild(i);
				if (child.getBounds().intersects(saveArea)) {
					// track highest and lowest rowIndex among saved views,
					// scroll them, and skip the rest of this
					if (child instanceof BRowView) {
//++visibleRowCount;
						// track row to position scroller
						int rowIndex = ((BRowView) child).rowIndex;
						if ((minIndex < 0 || minIndex > rowIndex)
								&& rowIndex >= 0) {
							minIndex = rowIndex;
						}
						if ((maxIndex < 0 || maxIndex < rowIndex)
								&& rowIndex >= 0) {
							maxIndex = rowIndex;
						}
					}
				}
			}
		}
		
//boolean firstShowing = firstRowShowing() != null;
//boolean lastShowing = lastRowShowing() != null;
//int visibleRowCount = 0;
//if(visibleRowCount != (maxIndex-minIndex+1))
//	debugOut("Scrollbar ASSUMPTIoN IS WRONG:"+visibleRowCount+" != "+maxIndex+"+1-"+minIndex);
//if(firstShowing != (minIndex == 0))
//	debugOut("Scrollbar ASSUMPTIoN IS WRONG: firstShowing "+firstShowing+" != "+minIndex+" == 0");
//if(lastShowing != (maxIndex == size-1))
//	debugOut("Scrollbar ASSUMPTIoN IS WRONG: lastShowing "+lastShowing+" != "+maxIndex+" == (-1)"+size);

		scrollbar.updateScroller(size, minIndex, maxIndex, scrollAnim);
	}

	@Override
	public boolean handleFocus(boolean isGained, BView gained, BView lost) {
		if(gained instanceof BLinkView) {
			setStatus(MessageFormat.format(Messages.getString("WebScreen.status_focused_link"),((BLinkView)gained).getLink())); //$NON-NLS-1$
		} else {
			setStatus(null);
		}
		return super.handleFocus(isGained, gained, lost);
	}
	
	private void setBackground() {
		debugOut("SET BACKGROUND "+html.bodyColor+" "+html.bodyImg+" "+html.bodyBgposition+" "+html.bodyBgrepeat);
		((BRowViewParent)pageArea).setRowParentBG(urlBase, html.bodyColor, html.bodyImg, html.bodyBgposition, html.bodyBgrepeat);
	}

	private void setTitle(String title, URL url, URL icon) {
		if(title == null) {
			title = url.getAuthority()+"/"+url.getPath(); //$NON-NLS-1$
		}
		this.pageIcon = icon;
		this.titleString = title;
		this.titleText.setResource(
				getBApp().createText("system-15.font", Color.black, title), //$NON-NLS-1$
				RSRC_HALIGN_LEFT | RSRC_VALIGN_BOTTOM );//| RSRC_TEXT_WRAP);
	}

	private void setStatus(String status) {
		if(status == null) {
			status = ""; //$NON-NLS-1$
		}
		if(this.statusText.getResource() != null) {
			this.statusText.getResource().remove();
		}
		this.statusText.setResource(
				getBApp().createText("system-10.font", Color.black, status+" "+WebBrowserApp.NOTICE), //$NON-NLS-1$ //$NON-NLS-2$
				RSRC_HALIGN_RIGHT | RSRC_VALIGN_TOP | RSRC_TEXT_WRAP);
	}

	private void continueEnter() {
		// do the detail work of the background (page area bg color set already)
		setBackground();
		// do the image part of setTitle:
		debugOut("continueEnter: set icon:"+this.pageIcon);//icon.getProtocol() //$NON-NLS-1$
		//FIXME looks like TiVo can't handle https for these
		this.titleIcon.setResource(this.pageIcon, RSRC_IMAGE_BESTFIT | RSRC_HALIGN_RIGHT | RSRC_VALIGN_CENTER);
		updateBookmarkNote();
//		debugOut("continueEnter");
		nextPage(true);
	}
	
	private BView loadingRowPlaceholder;
	private String media;
	
	private void showLoadingRowPlaceholder(int nextRowY, boolean scrollingDown) {
		if(this.loadingRowPlaceholder == null) {
////			this.loadingRowPlaceholder = new BView(pageArea, 0, 0, pageArea.getWidth(), 30);
//			this.loadingRowPlaceholder = new BView(getBelow(), pageArea.getX(), pageArea.getY(), pageArea.getWidth(), 30);
//			//FIXME use an animated gif (maybe a wide one)
//			this.loadingRowPlaceholder
//					.setResource(getResource("com/blackledge/david/tivo/widgets/pleaseWait.png"), 
//							RSRC_HALIGN_CENTER | RSRC_IMAGE_HFIT | RSRC_VALIGN_TOP);
			this.loadingRowPlaceholder = new BView(getAbove(), pageArea.getX(), pageArea.getY(), pageArea.getWidth(), pageArea.getHeight()/2);
			this.loadingRowPlaceholder.setResource(Color.LIGHT_GRAY);
			BView inner = new BView(loadingRowPlaceholder, 0,0, loadingRowPlaceholder.getWidth(), loadingRowPlaceholder.getHeight());
			inner
					.setResource(createText("system-40-italic.font", Color.black, Messages.getString("WebScreen.loading_inline_resource")),  //$NON-NLS-1$ //$NON-NLS-2$
							RSRC_HALIGN_CENTER | RSRC_VALIGN_CENTER);
		}
		if(scrollingDown) {
//			this.loadingRowPlaceholder.setLocation(this.loadingRowPlaceholder.getX(), loadingRowPlaceholder.getParent().getHeight()+20);
			this.loadingRowPlaceholder.setLocation(this.loadingRowPlaceholder.getX(), nextRowY, scrollAnim);
//			this.loadingRowPlaceholder.setLocation(this.loadingRowPlaceholder.getX(), nextRowY);
		} else {
//			this.loadingRowPlaceholder.setLocation(this.loadingRowPlaceholder.getX(), -(loadingRowPlaceholder.getHeight()+20));
			this.loadingRowPlaceholder.setLocation(this.loadingRowPlaceholder.getX(), nextRowY-loadingRowPlaceholder.getHeight(), scrollAnim);
//			this.loadingRowPlaceholder.setLocation(this.loadingRowPlaceholder.getX(), nextRowY-loadingRowPlaceholder.getHeight());
		}
//		this.loadingRowPlaceholder.setTransparency(1.0f);
		this.loadingRowPlaceholder.setVisible(true);
//		this.loadingRowPlaceholder.setTransparency(0.0f, getResource("*250"));
	}

	private void hideLoadingRowPlaceholder() {
		if(this.loadingRowPlaceholder != null) {
			this.loadingRowPlaceholder.setVisible(false);
		}	
	}

	private int findNamedAnchorIndex(String namedAnchor, List<HTMLRow> pageRows2) {
		if(namedAnchor == null || pageRows2 == null) return -1;

		int result = -1;
		
		for(int i = 0 ; i < pageRows2.size() ; ++i) {
			HTMLRow row = pageRows2.get(i);
			for(HTMLChunk chunk : row.getHTMLChunks()) {
				if(chunk instanceof HTMLImageChunk) {
					String chunkAnchor = ((HTMLImageChunk)chunk).anchor;
					if(namedAnchor.equalsIgnoreCase(chunkAnchor)) {
						debugOut("anchor "+namedAnchor+" row "+i); //$NON-NLS-1$ //$NON-NLS-2$
						return i;
					}
				} else if(chunk instanceof HTMLTextChunk) {
					String chunkAnchor = ((HTMLTextChunk)chunk).anchor;
					if(namedAnchor.equalsIgnoreCase(chunkAnchor)) {
						debugOut("anchor "+namedAnchor+" row "+i); //$NON-NLS-1$ //$NON-NLS-2$
						return i;
					}
//else if(chunkAnchor != null) debugOut("anchor OPTION:"+chunkAnchor);
				} else if(chunk instanceof HTMLTableCell) {
					int cellresult = findNamedAnchorIndex(namedAnchor, ((HTMLTableCell)chunk).rows);
					if(cellresult >= 0) {
						debugOut("anchor "+namedAnchor+" (cell) row "+i); //$NON-NLS-1$ //$NON-NLS-2$
						return i;
					}
				}
			}
		}
		
		return result;
	}
	
	void continueRender() {
debugOut("WebScreen.ContinueRender"); //$NON-NLS-1$
		RenderAction action = incompleteAction;
		incompleteAction = RenderAction.NO_ACTION;
		if(action == RenderAction.SCROLL_DOWN) {
			scrollDown(0, false);
		} else if(action == RenderAction.SCROLL_UP) {
			scrollUp(0);
		}
	}
	
	private void scrollToRow(int rowDest) {
		synchronized(this) {
			// remove all shown rows.
			removeAllChildren(pageArea);
			
			int renderedRow = rowDest-1;
			int nextRowY = 0;

			BView prevFocus = getFocus();
			boolean handlePainting = !this.getPainting();
			if (handlePainting) {
				this.setPainting(false);
			}
			try {
			for(int i = renderedRow+1 ; i < pageRows.size() ; ++i) {
					//debugOut(i+"/"+continueRows.size());
					boolean ignoreWrap = false;
					RenderAction incompleteActionType = RenderAction.SCROLL_DOWN;
					
					int align = pageRows.get(i).align;
					
					RenderProgress renderResult; 
					renderResult = html.renderRow(urlBase, getFontInfos(), pageArea, pageRows, i, pageArea.getWidth(), ignoreWrap, align);
					
					BRowView rowView = renderResult.rowView;
					
					boolean continuing = cleanupRow(nextRowY, 600, incompleteActionType,
							renderResult);
					if(!continuing) {
						break;
					}
//					rowView.setLocation(rowView.getX(), nextRowY+600);
//					if(!renderResult.isComplete) {
//						renderResult.nextRowY = nextRowY;
//						showLoadingRowPlaceholder(nextRowY);
//						incompleteRenderProgress = renderResult;
//	debugOut("pause    "+incompleteRenderProgress);
//						incompleteAction = 1;
//						break;
//					}
//					if(this.getFocus() == null) {
//						attemptFocus(rowView);
//					} 
					//else debugOut(this.getFocus());
					if (handlePainting) {
						this.setPainting(true);
					}
					rowView.setLocation(rowView.getX(), nextRowY, scrollAnim);
					nextRowY += rowView.getHeight();
					if(nextRowY > pageArea.getHeight()) {
						break;
					}
			}
			} finally {
				if (handlePainting) {
					this.setPainting(true);
				}
			}
			if(prevFocus != null) {
				prevFocus.getHighlights().refresh(scrollAnim);
			}
			if(getFocus() != null) {
				getFocus().getHighlights().refresh(scrollAnim);
			}
			
			updateScroller();
		}
	}
	
	private void scrollToTop() {
		synchronized(this) {
			scrollToRow(0);
	//		scrollUp(pageArea.getHeight());
			focusToTop();
		}
	}

	private void scrollToBottom() {
		synchronized(this) {
			// make a guess for first row to show, and do page down if needed.
			scrollToRow(pageRows.size()-12);
			int safety = 10;// max pagedown tries.
			while(lastRowShowing() == null && (--safety > 0)) {
				scrollDown(pageArea.getHeight(), false);
			}
			focusToBottom();
		}
	}

	private boolean nextPage(boolean firstLoad) {
		int amount = pageArea.getHeight() * 3 / 4;
		amount = amount - amount % 30;
		return scrollDown(amount, firstLoad);
	}

	private boolean scrollDown(int amount, boolean firstLoad) {
			synchronized(this) {
				int renderedRow;
				int nextRowY;
				if(incompleteRenderProgress == null) {
					if(lastRowShowing() != null) {
						return false;
					}
					if (!firstLoad) {
						// link selection played its own sound, don't play this one
						getBApp().play(V49BApplication.SOUND_PAGEDOWN);
					}
					BRowView renderedRowView = removeFarChildrenAndScroll(-amount, pageArea);
					renderedRow = renderedRowView == null ? -1
							: renderedRowView.rowIndex;
					nextRowY = renderedRowView == null ? 0
							: (renderedRowView.getY() + renderedRowView.getHeight());
					// jump to named anchor from URL if possible
					if (firstLoad) {
						String namedAnchor = urlBase.getRef();
						if (namedAnchor != null && namedAnchor.length() > 0) {
							int anchorIndex = findNamedAnchorIndex(namedAnchor, pageRows);
							if(anchorIndex >= 0) {
								renderedRow = anchorIndex - 1;
							} 
							else debugOut("anchor "+namedAnchor+" NOT FOUND "); //$NON-NLS-1$ //$NON-NLS-2$

//	debugOut(namedAnchor+":"+anchorIndex);
						} else if (html.jumpToRow > 0) {
							renderedRow = html.jumpToRow-1;//FIXME guessing this be -1, didn't used to be
						}
					}
				} else {
					renderedRow = incompleteRenderProgress.rowIndex-1;
					nextRowY = incompleteRenderProgress.nextRowY;
				}
				BView prevFocus = getFocus();
				for(int i = renderedRow+1 ; i < pageRows.size() ; ++i) {
					// allow for nested rows
					boolean handlePainting = !this.getPainting();
					if (handlePainting) {
						this.setPainting(false);
					}
					try {
						//debugOut(i+"/"+continueRows.size());
						boolean ignoreWrap = false;
						RenderAction incompleteActionType = RenderAction.SCROLL_DOWN;
						
						RenderProgress renderResult = renderRow(i, ignoreWrap);
	//					RenderProgress renderResult; 
	//					if(incompleteRenderProgress == null) {
	//						renderResult = html.renderRow(urlBase, fontInfos, pageArea, pageRows, i, pageArea.getWidth(), false);
	//					} else {
	//	debugOut("continue "+incompleteRenderProgress);
	//						hideLoadingRowPlaceholder();
	//						renderResult = html.renderRow(incompleteRenderProgress);
	//						incompleteRenderProgress = null;
	//					}
						BRowView rowView = renderResult.rowView;
	
						boolean continuing = cleanupRow(nextRowY, 600, incompleteActionType,
								renderResult);
						if(!continuing) {
							break;
						}
	//					rowView.setLocation(rowView.getX(), nextRowY+600);
	//					if(!renderResult.isComplete) {
	//						renderResult.nextRowY = nextRowY;
	//						showLoadingRowPlaceholder(nextRowY);
	//						incompleteRenderProgress = renderResult;
	//	debugOut("pause    "+incompleteRenderProgress);
	//						incompleteAction = 1;
	//						break;
	//					}
	//					if(this.getFocus() == null) {
	//						attemptFocus(rowView);
	//					} 
						//else debugOut(this.getFocus());
						if(handlePainting) {
							// turn painting back on for the animation in this case.
							this.setPainting(true);
						}
						rowView.setLocation(rowView.getX(), nextRowY, scrollAnim);
						nextRowY += rowView.getHeight();
						if(nextRowY > pageArea.getHeight()) {
							break;
						}
					} finally {
						if(handlePainting) {
							this.setPainting(true);
						}
					}
				}
				if(prevFocus != null) {
					prevFocus.getHighlights().refresh(scrollAnim);
				}
				if(getFocus() != null) {
					getFocus().getHighlights().refresh(scrollAnim);
				}
				
				updateScroller();
			}
			return true;
		}

	private boolean prevPage() {
		int amount = pageArea.getHeight() * 3 / 4;
		amount = amount - amount % 30;
		return scrollUp(amount);
	}
	
	private boolean scrollUp(int amount) {
		synchronized(this) {
			int renderedRow;
			int nextRowY;
			if(incompleteRenderProgress == null) {
				if(firstRowShowing() != null) {
					return false;
				}
				getBApp().play(V49BApplication.SOUND_PAGEUP);
				BRowView renderedRowView = removeFarChildrenAndScroll(amount, pageArea);
				//FIXME handle null renderedRowView better
				renderedRow = renderedRowView==null?0 : renderedRowView.rowIndex;
				nextRowY = renderedRowView==null?0 : renderedRowView.getY();
			} else {
				renderedRow = incompleteRenderProgress.rowIndex+1;
				nextRowY = incompleteRenderProgress.nextRowY;
			}
			BView prevFocus = getFocus();
			for(int i = renderedRow-1 ; i >=0 ; --i) {
				boolean handlePainting = !this.getPainting();
				if (handlePainting) {
					this.setPainting(false);
				}
				try {
					//debugOut(i+"/"+continueRows.size());
					int preSize = pageRows.size();
					
					// pass "true" for ignoreWrap since we're going up... wrapping was
					// added to rows before, we're just going through it again
					boolean ignoreWrap = true;
					RenderAction incompleteActionType = RenderAction.SCROLL_UP;
					
					RenderProgress renderResult = renderRow(i, ignoreWrap);
					
					BRowView rowView = renderResult.rowView;
					
					boolean continuing = cleanupRow(nextRowY, -600, incompleteActionType,
							renderResult);
					if(!continuing) {
						break;
					}
					
	if(pageRows.size() != preSize)
		debugOut((pageRows.size()-preSize)+" rows added"); //$NON-NLS-1$
	
					nextRowY -= rowView.getHeight();
					if (handlePainting) {
						this.setPainting(true);
					}
					rowView.setLocation(rowView.getX(), nextRowY, scrollAnim);
					if(nextRowY <= 0) {
						break;
					}
				} finally {
					if (handlePainting) {
						this.setPainting(true);
					}
				}
			}
			
			// if we just put the first row in view, make sure it's at the top, not below.
			Point firstRowLoc = firstRowShowing();
			if(firstRowLoc != null) {
				removeFarChildrenAndScroll(-firstRowLoc.y, pageArea);
				// then re-render the removed rows that should have been kept!
				//TODO should instead just re-render the missing ones
				// If a displayed row is at or below just near the bottom of the pageArea, that's good enough,
				// otherwise, scrollToRow(0)
				int bottom = getBottomOfDisplayedRows();
				if (!(bottom > pageArea.getHeight() - 20 || lastRowShowing() != null)) {
					scrollToRow(0);
				}
			}

			// make sure link focus moves, too.
			if(prevFocus != null) {
				prevFocus.getHighlights().refresh(scrollAnim);
			}
			if(getFocus() != null) {
				getFocus().getHighlights().refresh(scrollAnim);
			}
			
			updateScroller();
		}
		return true;
	}

	private int getBottomOfDisplayedRows() {
		int maxY = -10000;
		for (int i = 0; i < pageArea.getChildCount(); ++i) {
			View child = pageArea.getChild(i);
			if (child instanceof BView) {
				int childBottom = child.getY() + child.getHeight();
				if (childBottom > maxY) {
					maxY = childBottom;
				}
			}
		}
		if (maxY > 0)
			return maxY;
		else
			return 0;
	}

	private RenderProgress renderRow(int i, boolean ignoreWrap) {
		RenderProgress renderResult; 
		if(incompleteRenderProgress == null) {
			int align = pageRows.get(i).align;
			renderResult = html.renderRow(urlBase, getFontInfos(), pageArea, pageRows, i, pageArea.getWidth(), ignoreWrap, align);
		} else {
debugOut("continue "+incompleteRenderProgress); //$NON-NLS-1$
			hideLoadingRowPlaceholder();
			RenderProgress progress = incompleteRenderProgress;
			incompleteRenderProgress = null;
			renderResult = html.renderRow(progress);
		}
		return renderResult;
	}
	
	private boolean cleanupRow(int nextRowY, int locationYOffset,
			RenderAction incompleteActionType, RenderProgress renderResult) {
		BRowView rowView = renderResult.rowView;
		rowView.setLocation(rowView.getX(), nextRowY+locationYOffset);
		if(!renderResult.isComplete) {
			renderResult.nextRowY = nextRowY;
			showLoadingRowPlaceholder(nextRowY, (locationYOffset > 0));
			incompleteRenderProgress = renderResult;
debugOut("pause    "+incompleteRenderProgress); //$NON-NLS-1$
			incompleteAction = incompleteActionType;
		} else {
			if(this.getFocus() == null || this.getFocus() == pageArea) {
				attemptFocus(rowView);
			}
		}
		return renderResult.isComplete;
	}

	private boolean focusToTop() {
		BView prevFocus = getFocus();
		BView newFocus = null;
		int lastChildY = -100000;
		while(newFocus == null) {
			// find the next candidate - highest view below the last attempt
			// (pageArea children aren't in any particular order)
			BView focusChild = null;
			int minY = 100000;
			for(int i = 0 ; i < pageArea.getChildCount() ; ++i) {
				View child = pageArea.getChild(i);
				if(child instanceof BView) {
					if(child.getY() > lastChildY && child.getY() < minY) {
						minY = child.getY();
						focusChild = (BView) child;
					}
				}
			}
			if(focusChild == null) {
				// no candidates left
				break;
			}
			lastChildY = focusChild.getY();
			if(attemptFocus(focusChild)) {
				newFocus = getFocus();
				break;
			}
		}
		// return true if the focus changed
		return newFocus != null && prevFocus != newFocus;
	}

	private boolean focusToBottom() {
		BView prevFocus = getFocus();
		BView newFocus = null;
		int lastChildY = 100000;
		while(newFocus == null) {
			// find the next candidate - lowest view above the last attempt
			// (pageArea children aren't in any particular order)
			BView focusChild = null;
			int maxY = -100000;
			for(int i = 0 ; i < pageArea.getChildCount() ; ++i) {
				View child = pageArea.getChild(i);
				if(child instanceof BView) {
					if(child.getY() < lastChildY && child.getY() > maxY) {
						maxY = child.getY();
						focusChild = (BView) child;
					}
				}
			}
			if(focusChild == null) {
				// no candidates left
				break;
			}
			lastChildY = focusChild.getY();
			if(attemptFocus(focusChild)) {
				newFocus = getFocus();
				break;
			}
		}
		// return true if the focus changed
		return newFocus != null && prevFocus != newFocus;
	}

	/** like Focus Manager's "walkTree" */
	private boolean attemptFocus(BView rowView) {
		// find a focus among the immediate children
		for (int i = 0; i < rowView.getChildCount(); ++i) {
			View child = rowView.getChild(i);
			if (child instanceof BView && child.getVisible()
					&& child.getTransparency() != 1f) {
				BView bchild = (BView) child;
				if (bchild.isFocusable()) {
					setFocus(bchild);
//debugOut("focus set");
					return true;
				}
			}
		}
		// recurse down the tree
		for (int i = 0; i < rowView.getChildCount(); ++i) {
			View child = rowView.getChild(i);
			if (child instanceof BView && child.getVisible()
					&& child.getTransparency() != 1f) {
				BView bchild = (BView) child;
				if (attemptFocus(bchild))
					return true;
			}
		}
		return false;
	}

	private Point lastRowShowing() {
		if(pageRows == null) return null;
		return isRowShowing(pageRows.size()-1);
	}

	private Point firstRowShowing() {
		return isRowShowing(0);
	}

	private Point isRowShowing(int targetIndex) {
		for(int i = 0 ; i < pageArea.getChildCount() ; ++i) {
			if(pageArea.getChild(i) instanceof BRowView) {
				BRowView child = ((BRowView)pageArea.getChild(i));
				int index = child.rowIndex;
				if(index == targetIndex) {
					Rectangle childBounds = child.getBounds();
					// account for the fact that we're not comparing peers
					childBounds.translate(pageArea.getX(), pageArea.getY());
					if( pageArea.getBounds().intersects(childBounds))
						return child.getLocation();
					else
						return null;
				}
			}
		}
		return null;
	}

	private BRowView removeFarChildrenAndScroll(int ySlideAmount, View parent) {
		// collect the views before removing them, otherwise the getChild call
		// gets thrown by the remove calls.
		ArrayList<View> children = new ArrayList<View>();
		for(int i = 0 ; i < parent.getChildCount() ; ++i) {
			 children.add( parent.getChild(i));
		}
		
		//don't remove views that are plus or minus a half screen from here
//		Rectangle saveArea = new Rectangle(0, -parent.getHeight() / 2,
//				parent.getWidth(), parent.getHeight() * 2);
		// Retain based on ySlideAmount instead
		Rectangle saveArea = new Rectangle(0, -ySlideAmount,
				parent.getWidth(), parent.getHeight());

		BRowView maxIndex = null;
		BRowView minIndex = null;
		for(View child : children) {
			if(child.getBounds().intersects(saveArea)) {
				// track highest and lowest rowIndex among saved views, scroll them, and skip the rest of this
				if(child instanceof BRowView) {
					BRowView childRow = (BRowView) child;
					if(maxIndex == null || maxIndex.rowIndex < childRow.rowIndex) {
						maxIndex = childRow;
					}
					if(minIndex == null || minIndex.rowIndex > childRow.rowIndex) {
						minIndex = childRow;
					}
				}
				child.setLocation(child.getX(), child.getY() + ySlideAmount, scrollAnim);
				continue;
			}
			boolean handlePainting = !this.getPainting();
			if (handlePainting) {
				this.setPainting(false);
			}
			try {
				// remove the not-to-be-saved view and its children
				removeAllChildren(child);
				if(child.getResource() != null) {
					if(child.getResource().getClass().getName().contains("Text")) { //$NON-NLS-1$
						child.getResource().remove();
					}
				}
				child.clearResource(); // resource's own finalize will remove itself from tivo
				// make sure focus gets cleared if the one we're removing has it.
				if(child.hasFocus()) {
					this.setFocus(null);
				}
				child.remove();
			} finally {
				if (handlePainting) {
					this.setPainting(true);
				}
			}
		}
		if(ySlideAmount < 0) {
			return maxIndex;
		} else {
			return minIndex;
		}
	}
	
	private void removeAllChildren(View parent) {
		// collect the views before removing them, otherwise the getChild call
		// gets thrown by the remove calls.
		ArrayList<View> children = new ArrayList<View>();
		for(int i = 0 ; i < parent.getChildCount() ; ++i) {
			 children.add( parent.getChild(i));
		}
		
		for(View child : children) {
			removeAllChildren(child);
			if(child.getResource() != null) {
				if(child.getResource().getClass().getName().contains("Text")) { //$NON-NLS-1$
					child.getResource().remove();
				}
			}
			child.clearResource(); // resource's own finalize will remove itself from tivo
			// make sure focus gets cleared if the one we're removing has it.
			if(child.hasFocus()) {
				this.setFocus(null);
			}
			child.remove();
		}
	}

	
	private static class WebFocusManager extends BFocusMgr {

		public WebFocusManager(BApplication app) {
			super(app);
		}
		
		@Override
		public BView followArrow(BView src, int code) {
			if(code == KEY_UP) {
				//FIXME do custom handling instead.
				// get current focus's BRowView (or BRowViewParent??), 
				// find next BRowView (above/below) and find closest link in that BRowView.  
				// If there aren't any, repeat until siblings run out
				// parent up until another BRowView, and continue check... if no parent BRowView, give up
				// buuuut... if sibling is a table row, have to get... the closest cell's bottom link?
				return super.followArrow(src, code);
			} else if(code == KEY_DOWN) {
				//FIXME do custom handling instead.
				return super.followArrow(src, code);
			} else {
				return super.followArrow(src, code);
			}
		}
	}
	
	private BFocusMgr focusMgr;
	
	@Override
	public boolean handleAction(BView view, Object action) {
     synchronized (lock) {
	    	
		 // replace BScreen arrow handling with our custom focusMgr
		 boolean actionHandled;
		 List<String> arrowList = Arrays.asList(ARROW_NAMES);
		 int i = arrowList.indexOf(action);
		 if(i >= 0) {
			 BView newFocus = focusMgr.followArrow(view, KEY_UP + i);
			 if (newFocus != null) {
				 setFocus(newFocus);
				 return true;
			 }
			 actionHandled = false;
		 } else {
			 actionHandled = super.handleAction(view, action);
		 }
		
		WebBrowserApp app = (WebBrowserApp)getBApp();
//debugOut(action);
		if(! actionHandled ) {
			// leftover up/down means scroll the page
			if(action.equals(H_UP)) {
				scrollUp(30);
//				prevPage();
			} else if(action.equals(H_DOWN)) {
				scrollDown(30, false);
//				nextPage(false);
				
				//FIXME first, work when there is no focus to throw these actions
				//FIXME second, everything is cut off at pageArea's width currently.
			} else if(action.equals(H_RIGHT)) {
				if(scrollRight(100)) {
					getBApp().play(V49BApplication.SOUND_PAGEDOWN);
					return true;
				} else {
					getBApp().play(V49BApplication.SOUND_BONK);
					return false;
				}
			} else {
				if(action.equals(H_LEFT)) {
					if(scrollLeft(100)) {
						getBApp().play(V49BApplication.SOUND_PAGEDOWN);
						return true;
					} else {
						if(! isRepeating) { // TODO this probably isn't ever set under this circumstance
							if(app.isLeftSameAsBack()) {
								back();
								return true;
							}
						}
						getBApp().play(V49BApplication.SOUND_BONK);
						return false;
					}
					
				} else if(action.equals("LINK")) { //$NON-NLS-1$
					this.refreshCallback = null; // cancel any outstanding refresh.
					if(view instanceof BLinkView) {
						if(((BLinkView)view).getLink().startsWith("BUTTON:")) { //$NON-NLS-1$
							// only submit if button is submit or blank type
							String link = ((BLinkView)view).getLink();
							int formSep = link.indexOf("\n"); //$NON-NLS-1$
							String form = null;
							if(formSep > 0) {
								try {
									getBApp().play(V49BApplication.SOUND_SELECT);
									int valueSep = link.indexOf("\n",formSep+1); //$NON-NLS-1$
									if(valueSep < 0) valueSep = link.length();
									form = link.substring(formSep+1, valueSep);
									final String name = link.substring("BUTTON:".length(), formSep); //$NON-NLS-1$
									String val = null;
									if(valueSep<link.length()) val = link.substring(valueSep+1);
									final String value = val;
									HTMLFormField field = new HTMLFormField(form, name, "submit", null, false, false, -1) { //$NON-NLS-1$
										@Override
										public List<? extends String[]> getSubmissions() {
											if(value == null) return null;
											ArrayList<String[]> result = new ArrayList<String[]>();
											result.add(new String[] {name, value});
											return result;
										}
									};
									submitForm(field);
								} catch (Exception e) {
									debugOut("failure for "+((BLinkView)view).getLink()+"\n"+e);
									getBApp().play(V49BApplication.SOUND_BONK);
								}
							} else {
								getBApp().play(V49BApplication.SOUND_BONK);
							}

							return true;
						}
						app.showPleaseWait();
//					new PleaseWaitView(this, false).setVisible(true, getResource("*10"));
						getBApp().flush();
						loadNewScreen(((BLinkView)view).getLink());
						return true;
					}
				} else if(view instanceof FormButton || view instanceof FormImage) {
					this.refreshCallback = null; // cancel any outstanding refresh.
debugOut("Button with action: " + action); //$NON-NLS-1$
					if (action instanceof HTMLFormField) {
						try {
//					//actually, type button doesn't automatically submit!
//					if (((HTMLFormField) action).type.equals("button"))
//						return true;
							getBApp().play(V49BApplication.SOUND_SELECT);
							submitForm((HTMLFormField) action);
						} catch (Exception e) {
							debugOut("failure for image submit\n"+e);
							getBApp().play(V49BApplication.SOUND_BONK);
						}
						
						return true;
					}
				} else if(view instanceof FormInput) {
//debugOut("FormInput with action: "+action);
					String value = ""; //$NON-NLS-1$
					if (action instanceof HTMLFormInput) {
						HTMLFormInput input = (HTMLFormInput) action;
						value = input.getAttribute("value"); //$NON-NLS-1$
					} else if (action instanceof HTMLFormTextarea) {
						HTMLFormTextarea input = (HTMLFormTextarea) action;
						value = input.contents;
					}
					if(action instanceof HTMLFormField) {
						this.textDestination = (HTMLFormField) action;
						this.textHolder = (FormInput) view;
						Keyboard keyboard = Keyboard.DEFAULT;
						String fieldType = ((HTMLFormField)action).type;
						if (fieldType != null) {
							if (fieldType.equals("url")) { //$NON-NLS-1$
								keyboard = Keyboard.URL;
							} else if (fieldType.equals("email")) { //$NON-NLS-1$
								keyboard = Keyboard.EMAIL;
							} else if (fieldType.equals("password")) { //$NON-NLS-1$
								keyboard = Keyboard.PASSWORD;
							}
						}
						//FIXME support maxlength in forminput and set it on typo
//					((HTMLFormInput)action).getAttribute("maxLength");
//					typo.setMaxLength(num);
						
						typo = getTypo(view, keyboard);
						typo.setUseCompletions(true);
						if(fieldType != null) {
							if (Arrays.asList(
									new String[] { "numeric", "number", "date", //$NON-NLS-1$ //$NON-NLS-2$ //$NON-NLS-3$
											"time", "datetime", "datetime-local", //$NON-NLS-1$ //$NON-NLS-2$ //$NON-NLS-3$
											"month", "week", "range", "color", //$NON-NLS-1$ //$NON-NLS-2$ //$NON-NLS-3$ //$NON-NLS-4$
											"tel" }).contains(fieldType)) { //$NON-NLS-1$
								typo.setUseCompletions(false);
							}
						}
//					typo.setVisible(true);
//					setFocusDefault(typo);
						//FIXME actions might not propagate back to us if typo is on a different "plane"
						typo.setActions("text_set", "text_cancel"); //$NON-NLS-1$ //$NON-NLS-2$
						typo.setValue(value);
//					typo.setFocus();
//					setFocus(typo);
						//FIXME takes focus/inactivate/focus to get everything in right positions in typo
						return true;
					}
				} else if(view instanceof FormCheck) {
debugOut("Check with action: "+action); //$NON-NLS-1$
					if(action instanceof HTMLFormInput) {
						HTMLFormInput input = (HTMLFormInput) action;
						String check = input.getAttribute("checked"); //$NON-NLS-1$
						boolean checked = check != null;
						if(input.type.equals("checkbox")) { //$NON-NLS-1$
							checked = !checked;
						} else {
							checked = true;
						}
						if(checked) {
							getBApp().play(V49BApplication.SOUND_SELECT);
							input.putAttribute("checked", String.valueOf(checked)); //$NON-NLS-1$
						} else {
							getBApp().play(V49BApplication.SOUND_DESELECT);
							input.removeAttribute("checked"); //$NON-NLS-1$
						}
						if(input.type.equals("radio")) { //$NON-NLS-1$
							// uncheck other radio HTMLFormInput instances
							// clear all displayed radios of this name
							uncheckOtherRadios(input);
						}
						// update with actual checked value.
						((FormCheck)view).updateCheck(input.type, checked, input.color, app.getFontString(input, getTextScalePercent()));
						
						return true;
					}
				} else if(view instanceof FormSelect) {
debugOut("Select with action: "+action); //$NON-NLS-1$
					if(action instanceof HTMLFormSelect) {
						HTMLFormSelect input = (HTMLFormSelect) action;
						FormSelect select = (FormSelect) view;

//					String defaultval = null;
//					String nextval = null;
//					int i = 0;
//					for(String val :input.valuesDisplays.keySet()) {
//	//debugOut(val+" "+defaultval+" "+nextval+" "+input.selected);
//						if(i == 1) defaultval = val; // null got them index 0, so next should be 1
//						++i;
//						if(defaultval == null)defaultval = val;
//						if(nextval != null) {
//							nextval = val;
//							break;
//						}
//						if(val.equals(input.selected)) {
//							nextval = val; // use current val as a flag to use next one.
//						}
//						
//					}
//	//debugOut(defaultval+" "+nextval);
//					// didn't find a match, that's what defaultval was for.
//					if(nextval == null)
//						nextval = defaultval;
//					// use first val if we found current but there wasn't a value after it
//	//debugOut(defaultval+" "+nextval);
//					if(nextval.equals(input.selected)) {
//						getBApp().play(V49BApplication.SOUND_SLOWDOWN1);
//						nextval = input.valuesDisplays.keySet().iterator().next();
//					} else {
//						getBApp().play(V49BApplication.SOUND_SPEEDUP1);
//					}
//	//debugOut(defaultval+" "+nextval);
//					input.selected = nextval;

						//TODO this doesn't feel right. why do both of them need a copy of this information? I guess it makes sense, though.
						input.selectedIndices = (ArrayList<Integer>)select.selectedIndices.clone();
//					debugOut("update select with "+input.selectedIndices);
						select.updateSelect(input.valuesDisplays, input.selectedIndices, input.color, app.getFontString(input, getTextScalePercent()));
						
						return true;
					}
				}
			}
			return false;
		}
		return true;
     }
	}
	
	private boolean scrollRight(int i) {
//		if(???) {
//		pageArea.setResource(Color.yellow);
			pageArea.setTranslation(pageArea.getTranslationX()-i, 0, scrollAnim);
//		}
			if(getFocus() != null) {
				getFocus().getHighlights().refresh(scrollAnim);
			}
			return true;
	}

	private boolean scrollLeft(int i) {
		// scroll left if already scrolled to right.
		if(pageArea.getTranslationX() < 0) {
			pageArea.setTranslation(Math.min(0, pageArea.getTranslationX()+i), 0, scrollAnim);
		} else {
			return false;
//			pageArea.clearResource();
		}
		if(getFocus() != null) {
			getFocus().getHighlights().refresh(scrollAnim);
		}
		return true;
	}

	private void showWebPageMenu() {
		WebBrowserApp app = (WebBrowserApp)getBApp();
		this.media = app.getStreamSourceCustomURL();
		BScreen menu = new WebMenuScreen(this, this.media);
		// can't use TRANSITION_FORWARD because many menu commands cause a
		// screen replacement which fires during handleEnter BEFORE the pop
		// transition is processed
		getBApp().push(menu, TRANSITION_NONE);
	}

	private void handleReturnArg(Object arg) {
		WebBrowserApp app = (WebBrowserApp)getBApp();
		if(arg == null) return;
		String action = arg.toString();
		if(action.startsWith(WebMenuScreen.MENU_PREFIX)) {
			WebBrowserApp web = (WebBrowserApp) getBApp();
			web.handleMenuAction(action);
			// these actions require the reload of the screen
			if (action.equals(WebMenuScreen.MENU_TABLES)
					|| action.equals(WebMenuScreen.MENU_MOBILE)
					|| action.equals(WebMenuScreen.MENU_MAXIMIZE)
					|| action.startsWith(WebMenuScreen.MENU_FONT_PREFIX)
					|| action.startsWith(WebMenuScreen.MENU_STYLE_PREFIX)) {
				replaceWebScreen( new URL[] {urlBase, referer});
			} else if(action.equals(WebMenuScreen.MENU_BRIGHTNESS)) {
				toggleShade();
			} else if(action.equals(WebMenuScreen.MENU_PIP)) {
				app.setStartupStreamLiveTV();
			} else if(action.equals(WebMenuScreen.MENU_NO_PIP)) {
				app.clearStartupStream();
			} else if(action.equals(WebMenuScreen.MENU_PIP_MEDIA)) {
				app.setStartupStreamCustom(this.media);
			} else if(action.equals(WebMenuScreen.MENU_LINK_THUMBTACK)) {
				bookmarkFocusedLink();
			}
		} else {
			if(action.equals("text_set")) { //$NON-NLS-1$
				HTMLFormField textDest = this.textDestination;
				FormInput textH = this.textHolder;
				String value = typo.getValue();
				app.closeTyPo();
//		typo.setVisible(false);
				setFocusDefault(pageArea);
				this.textHolder.setFocus();
				this.setFocus(textH);
				this.textDestination = null;
				this.textHolder = null;
				if(textDest instanceof HTMLFormTextarea) {
					((HTMLFormTextarea)textDest).contents = value;
					textH.updateTextarea(value, textDest.color, app.getFontString(textDest, getTextScalePercent()));
				} else if(textDest instanceof HTMLFormInput) {
					((HTMLFormInput)textDest).putAttribute("value", value); //$NON-NLS-1$
					textH.updateText(textDest.type, value, textDest.color, app.getFontString(textDest, getTextScalePercent()));
				}
			} else if(action.equals("text_cancel")) { //$NON-NLS-1$
				app.closeTyPo();
//		typo.setVisible(false);
				setFocusDefault(pageArea);
				this.textHolder.setFocus();
				setFocus(this.textHolder);
				this.textDestination = null;
				this.textHolder = null;
			}
		}
	}

	private void bookmarkFocusedLink() {
		try {
			URL linkUrl = getFocusedLinkURL(urlBase);
			if(linkUrl != null) {
				String linkTitle = getFocusedLinkText();
				if(linkTitle != null) {
					if(((WebBrowserApp)getBApp()).saveBookmark(linkUrl, linkTitle, null)) {
						getBApp().play(V49BApplication.SOUND_THUMBSUP);
						updateBookmarkNote();
						return;
						// all other cases fall through to the "bonk"
					}
				}
			}
		} catch (MalformedURLException e) {
		}
		getBApp().play(V49BApplication.SOUND_BONK);
	}
	
	URL getFocusedLinkURL(URL urlBase2)
			throws MalformedURLException {
		BView view = getFocus();
		if(view instanceof BLinkView) {
			String link = ((BLinkView)view).getLink();
			if(!link.startsWith("BUTTON:")) { //$NON-NLS-1$
				return WebBrowserApp.getURL(urlBase2, link);
			}
		}
		return null;
	}
	
	String getFocusedLinkText() {
		BView view = getFocus();
		String linkTitle = null;
		if(view instanceof BLinkView) {
			String link = ((BLinkView)view).getLink();
			if(!link.startsWith("BUTTON:")) { //$NON-NLS-1$
				linkTitle = String.valueOf(view.getParent().getResource());
				// no easy access to the text resource's text... so hacking here to extract it from known toString format
				int txtStart = linkTitle.indexOf("txt="); //$NON-NLS-1$
				if(txtStart >= 0 && linkTitle.endsWith("]")) { //$NON-NLS-1$
					linkTitle = linkTitle.substring(txtStart+"txt=".length(), linkTitle.length()-1); //$NON-NLS-1$
					if(linkTitle.trim().length() == 0)
						linkTitle = link;
				} else {
					linkTitle = link;
				}
			}
		}
		return linkTitle;
	}
	
	private void uncheckOtherRadios(HTMLFormInput input) {
		if(input == null) {// || input.getFormActionSpec() == null) {
			return;
		}
		
		//find other radio's of same name in form and uncheck them.		
		Object formActionSpec = input.getFormActionSpec();
		synchronized(pageRows) {
			for (HTMLRow row : pageRows) {
				for (HTMLChunk chunk : row.getHTMLChunks()) {
					if (chunk instanceof HTMLFormInput) {
						HTMLFormInput field = (HTMLFormInput) chunk;
						if (input != field &&
								((formActionSpec == null && field.getFormActionSpec() == null) || 
										(formActionSpec != null && formActionSpec.equals(field.getFormActionSpec())))
								&& field.type.equals("radio") //$NON-NLS-1$
								&& field.name.equals(input.name)) {
							input.removeAttribute("checked"); //$NON-NLS-1$
						}
					}
				}
			}
		}
		
		// find any displayed ones and update them.
		// actually, just update all matching ones to unchecked, then it'll get updated correctly later.
		for(int i = 0 ; i < pageArea.getChildCount() ; ++i) {
			View child = pageArea.getChild(i);
			if(child instanceof BRowView) {
				BRowView childRow = (BRowView) child;
				for(int j = 0 ; j < childRow.getChildCount() ; ++j) {
					View chunk = childRow.getChild(j);
					if(chunk instanceof FormCheck) {
						FormCheck check = (FormCheck) chunk;
						if (input.name.equals(check.getName()) && 
								((formActionSpec == null && check.getForm() == null) ||
										(formActionSpec != null && formActionSpec.equals(check.getForm())))) {
							check.updateCheck(input.type, false, input.color,
									((WebBrowserApp) getBApp()).getFontString(input, getTextScalePercent()));
						}
					}
				}
			}
		}

	}

	private TyPo getTypo(BView origin, TyPo.Keyboard keyboard) {
		TyPo typo = ((WebBrowserApp)getBApp()).showTyPo();
//		if(this.typo == null) {
//			TyPo.setUserWordListAllowed(false);
//			this.typo = new TyPo(getNormal(), origin.getX(), origin.getY(), 350, 20, true);
//			this.typo.setActivateOnFocus(true);
//			this.typo.setExitOptions(TyPo.EXIT_CANCEL, TyPo.EXIT_FINISHED, true, false, false, false);
//		}
//		typo.setLocation(origin.getX(), origin.getY());
		typo.setKeyboard(keyboard);
		return typo;
	}

	private void submitForm(HTMLFormField action) {
		// get form... if it's blank, do nothing
		//TODO fix this... wants to check spec, not action
		if(action == null || action.getFormActionSpec() == null || 
				action.getFormActionSpec().toString().trim().equals("null:null") || action.getFormActionSpec().toString().trim().equals("null")) return; //$NON-NLS-1$ //$NON-NLS-2$
		if(!(action.getFormActionSpec() instanceof String)) {
			System.err.println("Unexpected form action object type "+action.getFormActionSpec()); //$NON-NLS-1$
			return;
		}
		String formAction = (String) action.getFormActionSpec();
		// break form into type and link
		String submitType = "GET"; //$NON-NLS-1$
		String submitAction = formAction;
		int separatorIndex = formAction.indexOf(":"); //$NON-NLS-1$
		if(separatorIndex >= 0 && formAction.indexOf(":",separatorIndex+1)>0) { //$NON-NLS-1$
			submitType = formAction.substring(0, separatorIndex).toUpperCase();
			// remove the uniqueness chunk
			submitType = submitType.substring(submitType.indexOf("&")+1); //$NON-NLS-1$
			submitAction = formAction.substring(separatorIndex+1);
		} else if(formAction.endsWith(":null")) { //$NON-NLS-1$
			return;
		}
		if(submitType.equals("null")) submitType = "GET"; //$NON-NLS-1$ //$NON-NLS-2$
		if(submitAction.equals("null")) { //$NON-NLS-1$
			return;
		}
		// get all submit name/value pairs with same form
		ArrayList<String[]> formSubmission = new ArrayList<String[]>();
		addFormFieldsFromRows(pageRows, action, formAction, formSubmission);
		
		// clean up the submission - no nulls.
		Iterator<String[]> iter = formSubmission.iterator();
		while(iter.hasNext()) {
			String[] keyVal = iter.next();
			if(keyVal[0] == null)
				iter.remove();
			else if(keyVal[0] == null) {
				keyVal[0] = ""; //$NON-NLS-1$
			}
		}
		
		WebBrowserApp app = (WebBrowserApp)getBApp();
		app.showPleaseWait();
		// submit name/value using type to link
		if(submitType.equals("POST")) { //$NON-NLS-1$
			loadNewScreen(app.doPost(submitAction,  "UTF-8", formSubmission, this.urlBase)); //$NON-NLS-1$
//		} else if(submitType.equals("GET")) {
		} else {
			loadNewScreen(app.doGet(submitAction, "UTF-8", formSubmission, this.urlBase)); //$NON-NLS-1$
//		} else {
//			System.err.println("don't know how to submit form with method "+submitType);
		}
		
	}

	private void addFormFieldsFromRows(List<HTMLRow> pageRows, HTMLFormField action, String formAction,
			ArrayList<String[]> formSubmission) {
		synchronized(pageRows) {
			for (HTMLRow row : pageRows) {
				for (HTMLChunk chunk : row.getHTMLChunks()) {
					if (chunk instanceof HTMLFormField) {
						HTMLFormField field = (HTMLFormField) chunk;
						if(field == action && field instanceof HTMLFormInput) {
							if(field.name != null) {
								String value = ((HTMLFormInput)field).getAttribute("value"); //$NON-NLS-1$
								formSubmission.add(new String[] {field.name, value});
	//debugOut(field.name+"="+value);
							}
						} else
						if (formAction.equals(field.getFormActionSpec())) {
							List<? extends String[]> submissions = field.getSubmissions();
							if(submissions != null) {
								formSubmission.addAll(submissions);
	//debugOut(field.getClass().getName()+":"+ submissions);
							}
						}
//						else debugOut("not including "+field+" because "+formAction+" != "+field.getFormActionSpec());
					} else if(chunk instanceof HTMLTableCell) {
						// recurse for form fields in tables
						addFormFieldsFromRows(((HTMLTableCell)chunk).rows, action, formAction, formSubmission);
					}
				}
			}
		}
	}

	private void loadNewScreen(URLConnection connection) {
		WebBrowserApp app = (WebBrowserApp)getBApp();
		if(connection == null) {
			app.hidePleaseWait();
			app.play(V49BApplication.SOUND_BONK);
			return;
		}
		URL linkUrl = connection.getURL();
		WebScreen webScreen = WebScreen.newWebScreen(getBApp());
		boolean pageSuccess = webScreen.prepare(linkUrl, this.urlBase, connection);
		if(!pageSuccess) {
			app.hidePleaseWait();
			webScreen.remove();
			return;
		}
		getBApp().push(webScreen, TRANSITION_FORWARD, new Object[]{linkUrl, this.urlBase, connection});
	}
	
	// JSOUP
	private void loadNewScreen(Response soupResp) {
		WebBrowserApp app = (WebBrowserApp)getBApp();
		if(soupResp == null) {
			app.hidePleaseWait();
			app.play(V49BApplication.SOUND_BONK);
			return;
		}
		URL linkUrl = soupResp.url();
//		String contentType = soupResp != null ? soupResp.contentType()
//				: WebBrowserApp.guessContentType(linkUrl);
//		
//		if(contentType != null && contentType.startsWith("video/")) {
//			getBApp().push(new VideoScreen(getBApp(), linkUrl, contentType), TRANSITION_FORWARD);
//		} else {
			WebScreen webScreen = WebScreen.newWebScreen(getBApp());
			boolean pageSuccess = webScreen.prepare(linkUrl, this.urlBase, soupResp);
			if(!pageSuccess) {
				app.hidePleaseWait();
				webScreen.remove();
				return;
			}
			getBApp().push(webScreen, TRANSITION_FORWARD,
					new Object[] { linkUrl, this.urlBase, soupResp });
//		}
	}
				// /JSOUP
	
	private void loadNewScreen(String link) {
		WebBrowserApp app = (WebBrowserApp)getBApp();
		if(link == null) {
			app.hidePleaseWait();
			app.play(V49BApplication.SOUND_BONK);
			return;
		}
			URL linkUrlBase;
			try {
				linkUrlBase = new URL(
						link
				);
			} catch (MalformedURLException e1) {
				app.hidePleaseWait();
				app.play(V49BApplication.SOUND_BONK);
				// FIXME Alert user 
				e1.printStackTrace();
				return;
			}
			//FIXME more to figure out here
	//		if(linkUrlBase.getHost().endsWith("youtube.com")) {
	//			try {
	//				linkUrlBase = new URL("hme://204.176.49.116:7288/videoplayer?"+linkUrlBase.getQuery());
	//			} catch (MalformedURLException e) {
	//				// shouldn't happen, but if it does, just fall through to default handling
	//			}
	//		}

			// handle if it's just a # page jump
			boolean refChange = false;
			if (linkUrlBase.getRef() != null) {
				try {
					if (new URL(this.urlBase, "#" + linkUrlBase.getRef()).equals(linkUrlBase)) {
						refChange = true;
					}
				} catch (MalformedURLException e) {
					// oh well, just do normal processing.
				}
			}
			
			if (linkUrlBase.getProtocol().toLowerCase().startsWith("hme")) { //$NON-NLS-1$
				app.loadApp(linkUrlBase);
			} else if (refChange) {
				WebScreen webScreen = WebScreen.newWebScreen(getBApp());
				boolean pageSuccess = webScreen.prepare(this, linkUrlBase);
				if(!pageSuccess) {
					app.hidePleaseWait();
					app.play(V49BApplication.SOUND_BONK);
					webScreen.remove();
					return;
				}
				getBApp().push(webScreen, TRANSITION_FORWARD, //TODO perhaps TRANSITION_NONE would make more sense here, but transition back would look weird.
						new URL[] { linkUrlBase, this.urlBase });
			} else {
	//			String contentType = WebBrowserApp.guessContentType(linkUrlBase);
				
	//			if(contentType != null && contentType.startsWith("video/")) {
	//				getBApp().push(new VideoScreen(getBApp(), linkUrlBase, contentType), TRANSITION_FORWARD);
	//			} else {
	//debugOut("not a video..."+contentType);
				WebScreen webScreen = WebScreen.newWebScreen(getBApp());
				boolean pageSuccess = webScreen.prepare(linkUrlBase, this.urlBase);
				if(!pageSuccess) {
					app.hidePleaseWait();
					webScreen.remove();
					return;
				}

					getBApp().push(webScreen, TRANSITION_FORWARD,
							new URL[] { linkUrlBase, this.urlBase });
	//			}
			}
		}

	private boolean prepare(URL linkUrl, URL urlBase2, Response soupResp) {
		Object arg = new Object[] {linkUrl, urlBase2, soupResp};
		return prepare(arg);
	}

	/**
	 * do the preprocessing while previous page is showing "please wait". After
	 * this, we're ready for a handleEnter(false) call.
	 */
	private boolean prepare(URL linkUrlBase, URL urlBase) {
		Object arg = new URL[] {linkUrlBase, urlBase};
		return prepare(arg);
	}
	private boolean prepare(URL linkUrl, URL urlBase, URLConnection connection) {
		Object arg = new Object[]{linkUrl, urlBase, connection};
		return prepare(arg);
	}
	/**
	 * prepare based on a previous screen and a new URL that is only a ref away.
	 * @param previous
	 * @param ref
	 * @return true if prepare is successful (always true in this case)
	 */
	private boolean prepare(WebScreen previous, URL ref) {
		setUpPageArea();
		html = previous.html;
		urlBase = ref;
		referer = previous.urlBase;
		pageRows =  previous.pageRows;
		requestURL = urlBase;
		((WebBrowserApp)getBApp()).setLoadingStatus(STATUS.COMPLETE);
		return true;	
	}
	boolean prepare(Object arg) {
		
		setUpPageArea();
		
		this.setStatus(Messages.getString("WebScreen.status_loading")); //$NON-NLS-1$
		
		
//		URLConnection connection = null;
				// JSOUP
		Response soupResp = null;
				// /JSOUP
		try {
//			connection =
			soupResp = 
					setupConnectionInformation(arg, soupResp);
		} catch (MalformedURLException e1) {
			// TODO Auto-generated catch block
			e1.printStackTrace();
			getBApp().play(V49BApplication.SOUND_BONK);
			return false;
		}
		
		WebBrowserApp app = (WebBrowserApp)getBApp();
		app.setLoadingStatus(STATUS.CONNECTING);
		
		boolean refreshing = true;
		boolean launchedVideo = false;
		while(refreshing) {
			html = new HTMLBreakdown((WebBrowserApp) this.getBApp());

			launchedVideo = launchVideoForPage(urlBase, launchedVideo);
			// we continue to display the page underneath the launched video if there was one.
			// except in certain cases
			if(launchedVideo && noPageForVideo(urlBase)) {
				return false;
			}
			
			// JSOUP
			org.jsoup.nodes.Document backupDoc = null; //TODO delete when jsoup supports application/*+xml
			if(soupResp == null) {
				try {
					soupResp = app.getResponse(urlBase, referer);
				} catch (UnsupportedMimeTypeException mimeEx) {
					// until jsoup supports other application / *+xml we have to deal with it here
					if (mimeEx.getMimeType() != null
							&& mimeEx.getMimeType().startsWith("application/")
							&& mimeEx.getMimeType().indexOf("+xml") > 0) { // jsoup handles plain application/xml
						debugOut(mimeEx.getMimeType()+" handling manually");
						String body = ((WebBrowserApp)this.getBApp()).getRawText(urlBase);
						try {
							backupDoc = html.convertXMLToHTML(urlBase, body);
						} catch (IOException e) {
							return launchSpecialMimeType(mimeEx);
						}
					} else {
						return launchSpecialMimeType(mimeEx);
					}
				} catch (IOException e1) {
					e1.printStackTrace();//FIXME handle this
					// this can happen for non-web-page content type, in addition to actual breakages
					// getResponse is set to not do this on errors (currently)
					getBApp().play(V49BApplication.SOUND_BONK);
					return false;
				}
			}
			// use the redirect url as the base of things from now on.
			if(soupResp != null && soupResp.url() != null) {
				urlBase = soupResp.url();
				// this may have changed our video recognition
				launchedVideo = launchVideoForPage(urlBase, launchedVideo);
				// we continue to display the page underneath the launched video if there was one.
				// except in certain cases
				if(launchedVideo && noPageForVideo(urlBase)) {
					return false;
				}
			}
			if (soupResp != null
					&& soupResp.contentType() != null
					&& (soupResp.contentType().toLowerCase()
							.indexOf("text") >= 0 || soupResp.contentType() //$NON-NLS-1$
							.toLowerCase().indexOf("html") >= 0)) { //$NON-NLS-1$
				//			htmlFile = connection.getInputStream();
			} else if(backupDoc != null) {
				//TODO ok... umm... pretty sure this isn't really possible except if we've got nulls
				return launchSpecialContentType(soupResp);
			}
			
			// shortcut check for immediate refresh in soupResp.header("Refresh"); BEFORE getRows
			// ( near duplicate of below refresh code)
			if(soupResp != null) {
				String refresh = soupResp.header("Refresh"); //$NON-NLS-1$
				if(refresh != null) {
					// don't actually schedule if it's more than 0 seconds
					URL url = scheduleRefresh(refresh, true);
					if(url != null) {
	debugOut("refresh(replace) "+urlBase+" immediately with "+url); //$NON-NLS-1$ //$NON-NLS-2$
						if(url.equals(urlBase)) {
							debugOut("immediate refresh skipped since urls are the same"); //$NON-NLS-1$
						} else {
							// url for immediate refresh. otherwise, no refresh or a
							// 1s-or-more refresh
							referer = urlBase;
							urlBase = url;
							soupResp = null;
							refreshing = true;
							app.setLoadingStatus(STATUS.REDIRECTING);
							continue;
						}
					}
				}
			}
			
			try {
				synchronized(this) {
					app.setLoadingStatus(STATUS.PARSING);
					if(backupDoc != null) { // TODO just until jsoup lets us get rid of backupDoc
						pageRows = html.getRows(urlBase, backupDoc, null);
					} else
					pageRows =  html.getRows(urlBase, soupResp);
//					debugOut("prepare: GOT "+pageRows.size()+"ROWS");
					requestURL = urlBase;
					if(html.baseURL != null)
						urlBase = html.baseURL;
				}
			} catch (IOException e) {
				debugOut("it blew up (html.getRows) "+e); //$NON-NLS-1$
			}

			refreshing = false;
			if(html.refresh != null) {
				URL url = scheduleRefresh(html.refresh, false);
				if(url != null) {
debugOut("refresh(replace) "+urlBase+" immediately with "+url); //$NON-NLS-1$ //$NON-NLS-2$
					if(url.equals(urlBase)) {
						debugOut("immediate refresh skipped since urls are the same"); //$NON-NLS-1$
					} else {
						// url for immediate refresh. otherwise, no refresh or a
						// 1s-or-more refresh was scheduled.
						referer = urlBase;
						urlBase = url;
						soupResp = null;
						refreshing = true;
						app.setLoadingStatus(STATUS.REDIRECTING);
						continue;
					}
				}
			}
		}
		// /JSOUP

//InputStream htmlFile;
//try {
//	if(connection == null)
//		connection = ((WebBrowserApp)getBApp()).openLink(urlBase, referer);
//	String contentType = connection.getContentType();
//	if (contentType != null && contentType.toLowerCase().indexOf("text") >= 0) {
//		htmlFile = connection.getInputStream();
//	} else {
////FIXME looks like TiVo can't handle https for these
//		// not text, hopefully it's a usable resource
//		// finish the display just by showing the resource
//		pageArea.setResource(urlBase);
//		pageRows = Collections.synch... new ArrayList<HTMLRow>();
//		setStatus("Displaying "+contentType);
//		setTitle(null, urlBase, html.icon);
//		continueEnter();
//		return true;
//	}
//} catch (IOException e1) {
//	// TODO Auto-generated catch block
//	e1.printStackTrace();
//	return false;
//}
//try {
//	synchronized(this) {
//		pageRows = html.getRows(urlBase, htmlFile);
//	}
//} catch (IOException e) {
//	// FIXME Auto-generated catch block
//	e.printStackTrace();
//	debugOut("it blew up");
//}
		app.setLoadingStatus(STATUS.COMPLETE);
		return true;	
	}

	/**
	 * Check if it's a video that doesn't have a page that goes with it and
	 * return true.
	 */
	private boolean noPageForVideo(URL videoUrl) {
		if (videoUrl.getHost().toLowerCase().contains("player.vimeo.com")) {
			return true;
		}
		if (videoUrl.toString().toLowerCase().contains("youtube.com/embed")) {
			return true;
		}
		return false;
	}

	private boolean launchSpecialContentType(Response soupResp) {
		WebBrowserApp app = (WebBrowserApp) getBApp();
		app.setLoadingStatus(STATUS.CONNECTING_STREAM);
		if(! handleMediaContent(soupResp)) {
			//handleMediaContent BONKs if it had an error
			return false;
		} else {
			return true;
		}
	}

	private boolean launchSpecialMimeType(UnsupportedMimeTypeException mimeEx) {
		WebBrowserApp app = (WebBrowserApp) getBApp();
		URL url;
		url = WebBrowserApp.getURL(urlBase, mimeEx.getUrl());
		if(url == null) {
			// this shouldn't happen, the url string came from a
			// full URL, but let's keep our best option in this
			// case.
			url = urlBase;
		}
		app.setLoadingStatus(STATUS.CONNECTING_STREAM);
		if (!handleMediaContent(mimeEx.getMimeType(), url)) {
			return false;
		} else {
			return true;
		}
	}

	private boolean launchVideoForPage(URL urlBase, boolean launchedVideo) {
		WebBrowserApp app = (WebBrowserApp) getBApp();
		// youtube!
		// youtube page isn't really formatted to allow show-the-web-page,-play-video-from-embed
		// youtube embeds used to be object+param (embed) and now are iframe (/embed/v/)
		if(!launchedVideo && YouTubeLauncher.isYouTubeVideo(urlBase)) {
			if(!YouTubeLauncher.launchVideo(urlBase, (WebBrowserApp) getBApp(), false)) {
				getBApp().play(V49BApplication.SOUND_BONK);
			} else {
				launchedVideo = true;
			}
			
			// continue to load youtube page underneath the launching (or failed) video
//				// either way, don't continue with regular page.
//				return false; (from "prepare" call)
		}
		if(!launchedVideo && VimeoHandler.isVimeoVideo(urlBase)) {
			app.setLoadingStatus(STATUS.CONNECTING_STREAM);
			URL vimeoURL = VimeoHandler.getVideoURL(urlBase, (WebBrowserApp)this.getBApp());
			if(vimeoURL != null) {
				//handle redirects, now.
				vimeoURL = app.resolveRedirects(vimeoURL, urlBase);
				if (!handleMediaContent(VimeoHandler.getVideoMimeType(), vimeoURL)) {
					// handleMediaContent BONKS if it was an error
					// continue to load vimeo page underneath failed video
//						return false; (from "prepare" call)
				} else {
					launchedVideo = true;
					// continue to load vimeo page underneath the launching video
//						return true; (from "prepare" call)
				}
				//this PAGE isn't media content (as handleMediaContent set)
				this.isMediaContent = false;
			} else {
				// notify user there was a problem - we TRIED to launch video, though.
				getBApp().play(V49BApplication.SOUND_BONK);
				debugOut("couldn't get media URL for vimeo video: "+urlBase);
			}
			//ALWAYS report that we launched a video... we tried.
			return true;
		}
		return launchedVideo;
	}

	private Response setupConnectionInformation(Object arg, Response soupResp)
				throws MalformedURLException {
			if (arg != null && arg instanceof URL) {
				urlBase = (URL) arg;
				referer = null;
			} else 		if (arg != null && arg instanceof URL[]) {
				URL[] urls = (URL[]) arg;
				if(urls.length > 0)
				urlBase = urls[0];
				if(urls.length > 1)
					referer = urls[1];
	
			} else 		if (arg != null && arg instanceof Object[]) {
				Object[] urls = (Object[]) arg;
				if(urls.length > 0)
				urlBase = (URL) urls[0];
				if(urls.length > 1)
					referer = (URL) urls[1];
				if(urls.length > 2) {
	//				if(urls[2] instanceof URLConnection)
	//				connection  = (URLConnection)urls[2];
	//				else 
					// JSOUP
						if(urls[2] instanceof Response) {
						soupResp = (Response) urls[2];
					}
					// /JSOUP
				}
	
			} else {
					urlBase = new URL(
	//						"http://david.blackledge.com"
	//						"http://mobile.smashingmagazine.com/2010/11/03/how-to-build-a-mobile-website/"
					// "http://www.google.com/"
							"http://news.google.com" //$NON-NLS-1$
					);
					referer = null;
			}
			// return connection;
			return soupResp;
		}

	@Override
	public boolean handleKeyPress(int code, long rawcode) {
		if(! super.handleKeyPress(code, rawcode)) {
			WebBrowserApp app = (WebBrowserApp)getBApp();
			switch(code) {
			case REFRESH_CALLBACK_KEY_CODE:
				// impossible key, used for sendEvent callback for screen refresh requests
				handleRefreshCallback();
				return true;
			case SCREEN_TRANSITION_CALLBACK_KEY_CODE:
				// callback timing is slightly BEFORE the screen transition is complete
				try { Thread.sleep(50); } catch (InterruptedException e) { }
				synchronized(this) {
					screenTransitionComplete = true;
					// start the embedded video PIP
					if(html.embeddedStream != null) {
						app.playStream(html.embeddedStream, true);
					}
					if (continueEnterOnScreenTransitionComplete) {
						continueEnter();
					}
				}
				return true;
//			case V49IHmeProtocol.KEY_OPT_B_BLUE:
//				toggleBGImagesVisible();
//				return true;
//			case KEY_ENTER:
//				toggleTableLayout();
//				return true;
			case KEY_RECORD:
				bookmarkFocusedLink();
				return true;
			case KEY_OPT_PIP://Zoom/Aspect/Window/PIP
				if(app.isStreamDisplayPIP()) {
					app.zoomPIP();
					return true;
				} else {
					toggleMobileReporting();
					return true;
				}
			case KEY_REPLAY:
				reloadPage();
				return true;
			case KEY_NUM0:
			case KEY_NUM1:
			case KEY_NUM2:
			case KEY_NUM3:
			case KEY_NUM4:
			case KEY_NUM5:
			case KEY_NUM6:
			case KEY_NUM7:
			case KEY_NUM8:
			case KEY_NUM9:
				handleKeyNum(code-KEY_NUM0);
				return true;
			case KEY_INFO:
				showWebPageMenu();
				return true;
			case KEY_ADVANCE:
				if(lastRowShowing() == null) {
					getBApp().play(V49BApplication.SOUND_SPEEDUP3);
					scrollToBottom();
				} else if(firstRowShowing() == null) {
					getBApp().play(V49BApplication.SOUND_SPEEDUP3);
					scrollToTop();
				}
				return true;
			case KEY_CHANNELDOWN:
				// end up on bottom when we don't move, or end up on top when we
				// do move.
				focusToBottom();
				nextPage(false);
				return true;
			case KEY_CHANNELUP:
				// end up on top when we don't move, or end up on bottom when we
				// do move.
				focusToTop();
				prevPage();
				return true;
			case KEY_LEFT:
				if(getFocus() != null) {
				postEvent(new BEvent.Action(getFocus(), H_LEFT));
				} else {
					if(scrollLeft(100)) {
						getBApp().play(V49BApplication.SOUND_PAGEDOWN);
						return true;
					} else {
						if(! isRepeating) {
							if(app.isLeftSameAsBack()) {
								back();
								return true;
							}
						}
						getBApp().play(V49BApplication.SOUND_BONK);
						return false;
					}
				}
				return true;
			case KEY_RIGHT:
				if(getFocus() != null) {
				postEvent(new BEvent.Action(getFocus(), H_RIGHT));
				} else {
					if(scrollRight(100)) {
						getBApp().play(V49BApplication.SOUND_PAGEDOWN);
						return true;
					} else {
						getBApp().play(V49BApplication.SOUND_BONK);
						return false;
					}
				}
				return true;
			case KEY_UP:
				if(getFocus() != null) {
				postEvent(new BEvent.Action(getFocus(), H_UP));
				} else {
					scrollUp(30);
				}
				return true;
			case KEY_DOWN:
				if(getFocus() != null) {
				postEvent(new BEvent.Action(getFocus(), H_DOWN));
				} else {
					scrollDown(30, false);
				}
				return true;
			case KEY_SELECT:
			case KEY_FORWARD:
				if(code == KEY_FORWARD) {
					if(app.isStreamTypeAudio()) {
						app.advanceAudio();
						return true;
					}
				}
				if (getFocus() != null) {
					getBApp().play(V49BApplication.SOUND_SELECT);
					postEvent(new BEvent.Action(getFocus(), "LINK")); //$NON-NLS-1$
					return true;
				}
				break;
			case V49IHmeProtocol.KEY_OPT_ACTION_D:
				toggleShade();
				break;
			case V49IHmeProtocol.KEY_OPT_ACTION_B:
			case KEY_REVERSE:
				back();
				return true;
//				break;
			case KEY_THUMBSUP:
				if(app.saveBookmark(urlBase, titleString, html.icon)) {
					getBApp().play(V49BApplication.SOUND_THUMBSUP);
					updateBookmarkNote();
				} else {
					getBApp().play(V49BApplication.SOUND_BONK);
				}
				break;
			case KEY_THUMBSDOWN:
				if(app.removeBookmark(urlBase, titleString, html.icon)) {
					getBApp().play(V49BApplication.SOUND_THUMBSDOWN);
					updateBookmarkNote();
				} else {
					getBApp().play(V49BApplication.SOUND_BONK);
				}
			}
			return false;
		}
		return true;
	}

	private void handleKeyNum(int num) {
		if (html.accessNumLinks[num] != null) {
			this.refreshCallback = null; // cancel any outstanding refresh.
			//TODO should focus on the link as well
			((WebBrowserApp) getBApp()).showPleaseWait();
			getBApp().flush();
			loadNewScreen(html.accessNumLinks[num]);
		} else {
			if (num == 0) {
				setTextToDefault();
			} else {
				setTemporaryTextKey(num);
			}
		}
	}

	private void handleRefreshCallback() {
		if (this.refreshCallback != null) {
			if (this == getBApp().getCurrentScreen()) {
				replaceWebScreen(new URL[] { this.refreshCallback,
						urlBase });
				// don't forget to flush when you're in a thread!
				getBApp().flush();
			}
			this.refreshCallback = null;
		}
	}

	private void back() {
		if (getBApp().getStackDepth() > 1) {
			getBApp().play(V49BApplication.SOUND_DESELECT);
			getBApp().pop();
		} else {
			((WebBrowserApp)getBApp()).closeBrowser();
		}
	}
	
	private void replaceWebScreen(Object arg) {
		// refresh (with same URL?) should really replace current page.
//			if(refreshUrl.equals(urlBase)) {
//				getBApp().pop();
//			}
			//FIXME not sure if referer should be current page or old referer
//			getBApp().push(
		//FIXME if this was the first screen, we need to use replace, I guess.
		if (getBApp().getStackDepth() > 1) {
			getBApp().pop();
		}
		((WebBrowserApp)getBApp()).showPleaseWait();

		WebScreen webScreen = WebScreen.newWebScreen(getBApp());
		boolean pageSuccess = webScreen.prepare(arg);
		if(!pageSuccess) {
			//FIXME although this is unlikely, it leaves us in a bad state - we lost previous screen and haven't replaced it!
			((WebBrowserApp)getBApp()).hidePleaseWait();
			getBApp().play(V49BApplication.SOUND_BONK);
			webScreen.remove();
			return;
		}

		getBApp().push(
//		getBApp().replaceCurrentScreen(
				webScreen, TRANSITION_FADE,
				arg);
	//	getBApp().replaceCurrentScreen(
	//			webScreen, TRANSITION_FORWARD, new URL[] {urlBase, referer});
	}
	
	private void setTemporaryTextKey(int newKey) {
		// this one doesn't save the setting.
		WebBrowserApp app = (WebBrowserApp) this.getBApp();
		app.setFontScale(app.getFontScale(newKey));
		replaceWebScreen( new URL[] {urlBase, referer});
	}
	private void setTextToDefault() {
		// restore last-saved setting
		WebBrowserApp app = (WebBrowserApp) this.getBApp();
		app.restoreSavedFontScale();
		replaceWebScreen( new URL[] {urlBase, referer});
	}
	private void toggleMobileReporting() {
		WebBrowserApp app = (WebBrowserApp) getBApp();
		app.setAppearingMobile(!app.isAppearingMobile());
		replaceWebScreen( new URL[] {urlBase, referer});
	}

	private void reloadPage() {
		replaceWebScreen( new URL[] {urlBase, referer});
	}
	
	private boolean updateBookmarkNote() {
		WebBrowserApp app = (WebBrowserApp)getBApp();
		if(app.isBookmarksDisabled())
			return false;
		HashMap<String, URL[]> bookmarks = app.getBookmarks();
//		if (bookmarks.containsValue(new URL[] {this.urlBase, this.pageIcon})) { 
		//.containsValue isn't reliable here... we don't want to require the pageIcon to match, just the main URL
		for(URL[] entry : bookmarks.values()) {
			if(entry != null && entry[0] != null) {
				if(entry[0].equals(this.urlBase)) {
					titleArea.setResource(Color.YELLOW);
					this.titleBookmark.setResource(
							app.getTheme().getPageSaved(), RSRC_IMAGE_BESTFIT | RSRC_HALIGN_CENTER | RSRC_VALIGN_CENTER);
					return true;
				}
			}
		}
		titleArea.setResource(Color.LIGHT_GRAY);
		this.titleBookmark.setResource(
				app.getTheme().getPageNotSaved(), RSRC_IMAGE_BESTFIT | RSRC_HALIGN_CENTER | RSRC_VALIGN_CENTER);
		return false;
	}
	
	private boolean isRepeating = false;
	private boolean screenTransitionComplete = false;
	private boolean continueEnterOnScreenTransitionComplete = false;
	@Override
	public boolean handleKeyRepeat(int code, long rawcode) {
		if(! super.handleKeyRepeat(code, rawcode)) {
			switch(code) {
			case KEY_CHANNELDOWN:
			case KEY_CHANNELUP:
			case KEY_UP:
			case KEY_DOWN:
			case KEY_LEFT:
			case KEY_RIGHT:
			case KEY_REVERSE:
				try {
					//TODO not really using this flag yet
					isRepeating = true;
					boolean result = handleKeyPress(code, rawcode);
					return result;
				} finally {
					isRepeating = false;
				}
			}
			return false;
		}
		return true;
	}

	/**
	 * go through rows and chunks to get all font definitions, create all, and
	 * wait for all.
	 */
	private boolean cacheFonts(List<HTMLRow> rows) {
		//FIXME do this at a pseudo-static level so all screens share the cache
		queueFonts(rows);
		if(fontCacheQueue.size() > 0) {
			createNextFontInQueue();
		} else {
			readyToContinueEnter();
			return true;
		}
//		check cache for all fonts needed, if all present, return false
		// otherwise, create missing fonts.
		return false;
	}

	private void readyToContinueEnter() {
		synchronized(this) {
			// if transition callback has fired, do continueEnter.
			// otherwise, set a flag so when transition callback fires, we continue enter.
			if (screenTransitionComplete) {
				continueEnter();
			} else {
				continueEnterOnScreenTransitionComplete = true;
			}
		}
	}

	private void queueFonts(List<HTMLRow> rows) {
		synchronized( this) {
			if(rows != null)
				synchronized(rows) {
					for(HTMLRow row : rows) {
						for (HTMLChunk chunk : row.getHTMLChunks()) {
							if (chunk instanceof HTMLStyleChunk) {
								String fontString = ((WebBrowserApp) getBApp()).getFontString(((HTMLStyleChunk) chunk), getTextScalePercent());
								if (getFontInfos().get(fontString) == null && !fontCacheQueue.contains(fontString)) {
									fontCacheQueue.add(fontString);
								}
							} else if(chunk instanceof HTMLTableCell) {
								queueFonts(((HTMLTableCell)chunk).rows);
							}
						}
					}
				}
		}
	}

	private void createNextFontInQueue() {
		if (fontCacheQueue.size() > 0) {
			String fontString = fontCacheQueue.get(0);
 debugOut("cache "+fontString+" "+(fontCacheQueue.size()-1)+" left"); //$NON-NLS-1$ //$NON-NLS-2$ //$NON-NLS-3$
			FontResource fontResource = getFont(fontString);
			fontResource.addHandler(this);
		} else {
			readyToContinueEnter();
		}
	}
	
	private FontResource getFont(String fontString) {
		String name = fontString;
		// copied from Application.java
        name = name.substring(0, name.length() - ".font".length()); //$NON-NLS-1$
        int dash = name.indexOf('-');
        if (dash == -1) {
            throw new IllegalArgumentException("'" + name + "' is not a valid font."); //$NON-NLS-1$ //$NON-NLS-2$
        }
        int dash2 = name.indexOf('-', dash + 1);
        
        String family = name.substring(0, dash) + ".ttf"; //$NON-NLS-1$
        int size;
        int style = FONT_PLAIN;

        try {
            if (dash2 != -1) {
                size = Integer.parseInt(name.substring(dash + 1, dash2));
                String styles = name.substring(dash2 + 1).toLowerCase();
                if (styles.equals("bold")) { //$NON-NLS-1$
                    style = FONT_BOLD;
                } else if (styles.equals("italic")) { //$NON-NLS-1$
                    style = FONT_ITALIC;
                } else if (styles.equals("bolditalic")) { //$NON-NLS-1$
                    style = FONT_BOLD | FONT_ITALIC;
                } else {
                    throw new IllegalArgumentException("'" + name + "' is not a valid font."); //$NON-NLS-1$ //$NON-NLS-2$
                }
            } else {
                size = Integer.parseInt(name.substring(dash + 1));
            }
        } catch (NumberFormatException e) {
            throw new IllegalArgumentException("'" + name + "' is not a valid font."); //$NON-NLS-1$ //$NON-NLS-2$
        }

        // this was already applied to create the fontString passed in.
//        // apply current scale to the actual font we get
//        size = (int) (size*this.getTextScalePercent());

        return (FontResource) createFont(family, style, size, FONT_METRICS_BASIC | FONT_METRICS_GLYPH);
	}

	public boolean isFontBold(String fontString) {
		String name = fontString;
		// copied from Application.java
		name = name.substring(0, name.length() - ".font".length()); //$NON-NLS-1$
		int dash = name.indexOf('-');
		if (dash == -1) {
			throw new IllegalArgumentException("'" + name //$NON-NLS-1$
					+ "' is not a valid font."); //$NON-NLS-1$
		}
		int dash2 = name.indexOf('-', dash + 1);

		if (dash2 != -1) {
			String styles = name.substring(dash2 + 1).toLowerCase();
			if (styles.equals("bold")) { //$NON-NLS-1$
				return true;
			} else if (styles.equals("italic")) { //$NON-NLS-1$
			} else if (styles.equals("bolditalic")) { //$NON-NLS-1$
				return true;
			} else {
				throw new IllegalArgumentException("'" + name //$NON-NLS-1$
						+ "' is not a valid font."); //$NON-NLS-1$
			}
			return false;
		}
		return false;
	}

	public boolean isFontItalic(String fontString) {
		String name = fontString;
		// copied from Application.java
		name = name.substring(0, name.length() - ".font".length()); //$NON-NLS-1$
		int dash = name.indexOf('-');
		if (dash == -1) {
			throw new IllegalArgumentException("'" + name //$NON-NLS-1$
					+ "' is not a valid font."); //$NON-NLS-1$
		}
		int dash2 = name.indexOf('-', dash + 1);

		if (dash2 != -1) {
			String styles = name.substring(dash2 + 1).toLowerCase();
			if (styles.equals("bold")) { //$NON-NLS-1$
			} else if (styles.equals("italic")) { //$NON-NLS-1$
				return true;
			} else if (styles.equals("bolditalic")) { //$NON-NLS-1$
				return true;
			} else {
				throw new IllegalArgumentException("'" + name //$NON-NLS-1$
						+ "' is not a valid font."); //$NON-NLS-1$
			}
			return false;
		}
		return false;
	}

	public String getFontFilename(String fontString) {
		String name = fontString;
		// copied from Application.java
		name = name.substring(0, name.length() - ".font".length()); //$NON-NLS-1$
		int dash = name.indexOf('-');
		if (dash == -1) {
			throw new IllegalArgumentException("'" + name //$NON-NLS-1$
					+ "' is not a valid font."); //$NON-NLS-1$
		}
		String family = name.substring(0, dash) + ".ttf"; //$NON-NLS-1$
		return family;
	}

	public float getFontSize(String fontString) {
		String name = fontString;
		// copied from Application.java
		name = name.substring(0, name.length() - ".font".length()); //$NON-NLS-1$
		int dash = name.indexOf('-');
		if (dash == -1) {
			throw new IllegalArgumentException("'" + name //$NON-NLS-1$
					+ "' is not a valid font."); //$NON-NLS-1$
		}
		int dash2 = name.indexOf('-', dash + 1);

		float size;

		try {
			if (dash2 != -1) {
				size = Float.parseFloat(name.substring(dash + 1, dash2));
			} else {
				size = Float.parseFloat(name.substring(dash + 1));
			}
		} catch (NumberFormatException e) {
			throw new IllegalArgumentException("'" + name //$NON-NLS-1$
					+ "' is not a valid font."); //$NON-NLS-1$
		}

		// this was already applied to the fontString we're processing
//		// apply current scale to the actual font we get
//		size = (float) (size * this.getTextScalePercent());
		return size;
	}

//	private static Font default_ttf = null;
//	private static Font system_ttf = null;
//	static {
//			 try {
//				default_ttf = Font.createFont(Font.TRUETYPE_FONT, new File("C:\\Users\\David\\Desktop\\Development\\hme_sdk_1.4.1_src\\com\\tivo\\hme\\sim\\default.ttf"));
//			} catch (FontFormatException e) {
//				System.err.println("Unable to create default font for measurement: "+e);
//			} catch (IOException e) {
//				System.err.println("Unable to create default font for measurement: "+e);
//			}
//			 try {
//				system_ttf = Font.createFont(Font.TRUETYPE_FONT, new File("C:\\Users\\David\\Desktop\\Development\\hme_sdk_1.4.1_src\\com\\tivo\\hme\\sim\\system.ttf"));
//			} catch (FontFormatException e) {
//				System.err.println("Unable to create system font for measurement: "+e);
//			} catch (IOException e) {
//				System.err.println("Unable to create system font for measurement: "+e);
//			}
//	}
	@Override
	public boolean handleEvent(HmeEvent event) {
		switch (event.getOpCode()) {
		case EVT_FONT_INFO:
			HmeEvent.FontInfo fi = (HmeEvent.FontInfo) event;
//debugOut("got font "+fi.getID());
			if (fontCacheQueue.size() > 0) {
				String fontName = fontCacheQueue.remove(0);
	debugOut(fontName+"\t h "+fi.getHeight()+"\t lg "+fi.getLineGap()+"\t asc "+fi.getAscent()+"\t desc "+fi.getDescent()); //$NON-NLS-1$ //$NON-NLS-2$ //$NON-NLS-3$ //$NON-NLS-4$
/*	
	//double totalAdvance = 0.0;
	//double totalBounding = 0.0;
	//double totalMax = 0.0;
	FontInfo lfi = createLocalFontInfo(fontName);
//	if(lfi != null) {
////		fi = lfi;
//		debugOut(fontName+"\t h "+lfi.getHeight()+"\t lg "+lfi.getLineGap()+"\t asc "+lfi.getAscent()+"\t desc "+lfi.getDescent());
//		
//		for(char c = ' ' ; c <= '~' ; ++c) {
//			float adv = fi.getGlyphInfo(c).getAdvance() - lfi.getGlyphInfo(c).getAdvance();
//			float bw = fi.getGlyphInfo(c).getBoundingWidth() - lfi.getGlyphInfo(c).getBoundingWidth();
//			debugOut(String.format("%c : %5.3f : %5.3f",c,adv,bw));
//		}
//	}
	//debugOut("");
	//debugOut("all advance: "+totalAdvance+" all bound: "+totalBounding+" all max: "+totalMax);
				((WebBrowserApp)getBApp()).localFontInfos.put(fontName, lfi);
*/			
				
				getFontInfos().put(fontName, fi);
				createNextFontInQueue();
			}
		}	
		return super.handleEvent(event);
	}

//	private FontInfo createLocalFontInfo(String fontSpec) {
//		LocalFontInfo lfi;
//		try {
//		
//		String fontfile = getFontFilename(fontSpec);
//		Font font = default_ttf;
//		if(fontfile.startsWith("system")) {
//			font = system_ttf;
//		}
//		float fontSize = getFontSize(fontSpec);
//		int fontStyle = Font.PLAIN;
//		if (isFontBold(fontSpec))
//			fontStyle |= Font.BOLD;
//		else if (isFontItalic(fontSpec))
//			fontStyle |= Font.ITALIC;
//		font = font.deriveFont(fontStyle, fontSize);
//		 lfi = LocalFontInfo.newFontInfo(font);
//		
////		new FontInfo(fontid, font.getSize2D(), fontStyle, fontSize, fontStyle, getFontInfos())
//		
//		//for(int i = 32 ; i < 127 ; ++i) {
//		//	//int i = 32;
//		//	GlyphInfo gi = fi.getGlyphInfo((char)i);
//		//	if(gi != null) {
//		//	//		totalMax += Math.max(gi.getAdvance(),gi.getBoundingWidth());
//		//	//		totalAdvance += gi.getAdvance();
//		//	//		totalBounding += gi.getBoundingWidth();
//		//		debugOut(gi.getCharacter()+"\t"+gi.getAdvance()+"\t"+gi.getBoundingWidth());
//		//	}
//		//	float bounds = default_ttf.createGlyphVector(fontRenderContext, String.valueOf((char)i)).getGlyphMetrics(0).getAdvance();
//		//	float points = fi.getHeight();
//		//	try {
//		//		if(fontName.indexOf('-') >=0)
//		//	points = Integer.parseInt(fontName.substring(fontName.indexOf('-')+1, fontName.indexOf('-')+1+2));
//		//	} catch(Exception e) {
//		//	}
//		//	debugOut(((char)i)+"\t"+(bounds*points)+" = "+bounds+"*"+points);
//		//}
//				        } catch(Exception e) {
//				        	e.printStackTrace();
//				        	return null;
//				        }
//		return lfi;
//	}
	public static final float getStringWidth(String str, java.awt.Font font) {
	    Rectangle2D rect =
	        font.getStringBounds(str, 0, str.length(),
	                             new FontRenderContext(new AffineTransform(),
	                             true, true));
	    return (float)rect.getWidth();
	}

	private float getTextScalePercent() {
		return ((WebBrowserApp) getBApp()).getFontScale();
	}

	private Map<String, FontInfo> getFontInfos() {
		return ((WebBrowserApp) getBApp()).getFontInfos();
	}

	public int getFontSizeKey() {
		return ((WebBrowserApp)getBApp()).getFontScaleKey();
	}

	public URL getURL() {
		return urlBase;
	}
	
	public String getTitle() {
		return titleString;
	}
	
	public URL getIconURL() {
		return pageIcon;
	}

}
