package com.blackledge.david.html.hme.host;

import java.util.ArrayList;

public class HMEItem {

	public String title;
	public String url;
	public String iconUrl;
	public int topleveldisplayorderhint = -1;
	public boolean hidden = false;
	public String uuid = null;
	public boolean genres = false;
	public static final String GENRE_MUSIC = "music";
	public static final String GENRE_PHOTOS = "photos";
	public static final String GENRE_VIDEO = "video";
	public static final String GENRE_OTHER = "other";
	public ArrayList<String> genreList = new ArrayList<String>();
	

	public HMEItem(String title, String url, String iconUrl, String uuid, String genre) {
		this.title = title;
		this.url = url;
		this.iconUrl = iconUrl;
		this.uuid = uuid;
//		this.uuid = UUID.randomUUID().toString().toUpperCase();
//		this.topleveldisplayorderhint = 15;
		if(genre != null) {
			if(genre.equals(GENRE_VIDEO)) {
				// doesn't work without special permission or something
				// so also add to "other"
				this.genreList.add(GENRE_OTHER);
			}
		this.genreList.add(genre);
		}
		//TODO replace null iconUrl with url/icon.png (that's automatic, though by tivo?)
	}
	public HMEItem(String title, String url, String iconUrl) {
		this(title, url, iconUrl, null, null);
	}

	@Override
	public String toString() {
		StringBuilder stringBuilder = new StringBuilder();
		
		stringBuilder.append("   <Item> ");
		stringBuilder.append("\n");
		stringBuilder.append("    <Details>");
		stringBuilder.append("\n");
		stringBuilder.append("     <ContentType>application/x-hme</ContentType>");
		stringBuilder.append("\n");
		if(hidden) {
			stringBuilder.append("     <Hidden />");
			stringBuilder.append("\n");
		}
		stringBuilder.append("     <SourceFormat>x-container/folder</SourceFormat>");
		stringBuilder.append("\n");
		stringBuilder.append("     <Title>");
		stringBuilder.append(xmlEncode(title));
		stringBuilder.append("</Title>");
		stringBuilder.append("\n");
		if(uuid != null) {
			stringBuilder.append("     <Uuid>");
			stringBuilder.append(xmlEncode(uuid));
	//				"984BBCD1-F705-6F3C-8E1A-BA7F4C26B570" +
			stringBuilder.append("</Uuid>");
			stringBuilder.append("\n");
		}
		if(topleveldisplayorderhint >= 0) { //FIXME verify case of this tag
			stringBuilder.append("     <X-TiVo-TopLevelDisplayOrderHint>");
			stringBuilder.append(topleveldisplayorderhint);
			stringBuilder.append("</X-TiVo-TopLevelDisplayOrderHint>");
			stringBuilder.append("\n");
		}
		stringBuilder.append("    </Details>");
		
		if(genres) {
			stringBuilder.append("\n");
			stringBuilder.append("    <Genres>");
			stringBuilder.append("\n");
			if(genreList == null || genreList.size() == 0) {
				stringBuilder.append("    <Genre>"+GENRE_OTHER+"</Genre>");
				stringBuilder.append("\n");
			} else {
				for(String genre : genreList) {
					stringBuilder.append("    <Genre>"+genre+"</Genre>");
					stringBuilder.append("\n");
				}
			}
			stringBuilder.append("    </Genres>");
			stringBuilder.append("\n");
		}
		
		stringBuilder.append("    <Links>");
		stringBuilder.append("\n");
		stringBuilder.append("     <Content>");
		stringBuilder.append("\n");
		if(genres) { // this kind of request requires a format change too...I guess
			stringBuilder.append("<ContentType>application/x-hme</ContentType>");
			stringBuilder.append("\n");
		}
		stringBuilder.append("      <Url>");
		stringBuilder.append(xmlEncode(url));
		stringBuilder.append("</Url>");
		stringBuilder.append("\n");
		stringBuilder.append("     </Content>");
		stringBuilder.append("\n");
		if(iconUrl != null) {
			stringBuilder.append("     <CustomIcon> \r\n");
			stringBuilder.append("      <Url>");
			stringBuilder.append(xmlEncode(iconUrl));
			stringBuilder.append("</Url> \r\n");
			stringBuilder.append("     </CustomIcon>\r\n");
		}
		stringBuilder.append("    </Links>");
		stringBuilder.append("\n");
		stringBuilder.append("   </Item>");
		stringBuilder.append("\n");
		
		return stringBuilder.toString();
		
	}
	
	public static String xmlEncode(String content) {
		return content.replace("&", "&amp;");
	}
}
